Package: editbl
Type: Package
Version: 1.3.2
Date: 2026-07-09
Title: 'DT' Extension for CRUD (Create, Read, Update, Delete)
        Applications in 'shiny'
Authors@R: c(person("Jasper", "Schelfhout", "", "jasper.schelfhout@openanalytics.eu",
           role = c("aut", "cre")),
           person("Maxim", "Nazarov", "", "maxim.nazarov@openanalytics.eu",
           role = c("rev")),
           person("Daan", "Seynaeve", "", "daan.seynaeve@openanalytics.eu",
           role = c("rev")),
           person("Lennart", "Tuijnder", "", "lennart.tuijnder@openanalytics.eu",
           role = c("rev")),
		   person("Saar", "Junius", "", "saar.junius@openanalytics.eu",
           role = c("aut")))
Maintainer: Jasper Schelfhout <jasper.schelfhout@openanalytics.eu>
Description: The core of this package is a function eDT() which enhances DT::datatable() such that it can be used to interactively modify data in 'shiny'. By the use of generic 'dplyr' methods it supports many types of data storage, with relational databases ('dbplyr') being the main use case.
License: GPL-3
Copyright: Open Analytics NV, 2023
Imports: shiny, shinyjs, DT, tibble, dplyr, rlang, uuid, fontawesome
        (>= 0.4.0)
Suggests: testthat, dtplyr, data.table, vctrs, RSQLite, dbplyr, glue,
        DBI, bit64, knitr, dm
URL: https://github.com/openanalytics/editbl
BugReports: https://github.com/openanalytics/editbl/issues
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-07-09 12:09:46 UTC; jschelfhout
Author: Jasper Schelfhout [aut, cre],
  Maxim Nazarov [rev],
  Daan Seynaeve [rev],
  Lennart Tuijnder [rev],
  Saar Junius [aut]
Repository: CRAN
Date/Publication: 2026-07-09 16:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(e_rows_delete,default)
S3method(e_rows_insert,default)
S3method(e_rows_insert,dtplyr_step)
S3method(e_rows_insert,tbl_dbi)
S3method(e_rows_update,data.frame)
S3method(e_rows_update,default)
S3method(e_rows_update,dtplyr_step)
S3method(e_rows_update,tbl_dbi)
S3method(inputServer,default)
S3method(inputUI,default)
S3method(rows_delete,dtplyr_step)
export(connectDB)
export(customButton)
export(eDT)
export(eDTOutput)
export(e_rows_delete)
export(e_rows_insert)
export(e_rows_update)
export(foreignTbl)
export(inputServer)
export(inputUI)
export(runDemoApp)
export(selectInputDT_Server)
export(selectInputDT_UI)
import(fontawesome)
importFrom(DT,DTOutput)
importFrom(DT,coerceValue)
importFrom(DT,dataTableProxy)
importFrom(DT,datatable)
importFrom(DT,formatStyle)
importFrom(DT,hideCols)
importFrom(DT,renderDT)
importFrom(DT,renderDataTable)
importFrom(DT,styleEqual)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,anti_join)
importFrom(dplyr,as_tibble)
importFrom(dplyr,collect)
importFrom(dplyr,count)
importFrom(dplyr,distinct)
importFrom(dplyr,everything)
importFrom(dplyr,filter)
importFrom(dplyr,if_any)
importFrom(dplyr,inner_join)
importFrom(dplyr,is.tbl)
importFrom(dplyr,left_join)
importFrom(dplyr,pull)
importFrom(dplyr,relocate)
importFrom(dplyr,rows_delete)
importFrom(dplyr,rows_insert)
importFrom(dplyr,rows_update)
importFrom(dplyr,rowwise)
importFrom(dplyr,select)
importFrom(dplyr,tbl)
importFrom(dplyr,tbl_vars)
importFrom(dplyr,tibble)
importFrom(dplyr,type_sum)
importFrom(dplyr,union)
importFrom(rlang,":=")
importFrom(shiny,HTML)
importFrom(shiny,NS)
importFrom(shiny,actionButton)
importFrom(shiny,dateInput)
importFrom(shiny,div)
importFrom(shiny,fluidPage)
importFrom(shiny,freezeReactiveValue)
importFrom(shiny,icon)
importFrom(shiny,is.reactive)
importFrom(shiny,isTruthy)
importFrom(shiny,isolate)
importFrom(shiny,modalButton)
importFrom(shiny,modalDialog)
importFrom(shiny,moduleServer)
importFrom(shiny,numericInput)
importFrom(shiny,observe)
importFrom(shiny,observeEvent)
importFrom(shiny,reactive)
importFrom(shiny,reactiveValues)
importFrom(shiny,renderPrint)
importFrom(shiny,renderUI)
importFrom(shiny,req)
importFrom(shiny,selectInput)
importFrom(shiny,shinyApp)
importFrom(shiny,showModal)
importFrom(shiny,showNotification)
importFrom(shiny,stopApp)
importFrom(shiny,tagList)
importFrom(shiny,tags)
importFrom(shiny,textInput)
importFrom(shiny,uiOutput)
importFrom(shiny,verbatimTextOutput)
importFrom(shinyjs,disable)
importFrom(shinyjs,disabled)
importFrom(shinyjs,enable)
importFrom(shinyjs,hidden)
importFrom(shinyjs,useShinyjs)
importFrom(tibble,as_tibble)
importFrom(utils,head)
importFrom(utils,packageName)
importFrom(utils,str)
importFrom(utils,tail)
importFrom(uuid,UUIDgenerate)

# {editbl}: DT extension for CRUD <img style="height: 139px;" src='https://github.com/openanalytics/editbl/blob/main/editbl_logo.svg?raw=true' align="right" />

[![CRAN status](https://www.r-pkg.org/badges/version/editbl)](https://cran.r-project.org/package=editbl)
[![R-CMD-check](https://github.com/openanalytics/editbl/actions/workflows/check-standard.yaml/badge.svg)](https://github.com/openanalytics/editbl/actions/workflows/check-standard.yaml)
[![codecov](https://codecov.io/gh/openanalytics/editbl/branch/main/graph/badge.svg)](https://app.codecov.io/gh/openanalytics/editbl)
![CRAN downloads](https://cranlogs.r-pkg.org/badges/editbl)

`editbl` ('*edit [tibble](https://cran.r-project.org/package=tibble)*') allows you to modify tables in a spreadsheet-like fashion. Not just in-memory `data.frame` objects, but also
data living in a database.

## Installation

* From CRAN:

```
install.packages('editbl')
```

* Latest development version:

```
remotes::install_github("https://github.com/openanalytics/editbl", ref = "main", subdir = "editbl")
```

## Get started

You can play around with a [live demo](https://openanalytics.github.io/editbl/).

Alternatively, choose your own dataset and use `eDT` to interactively explore and modify it:

```
modifiedData <- editbl::eDT(mtcars)
print(modifiedData)
```


The package itself also includes demo apps:

```
editbl::runDemoApp()
```
![](https://github.com/openanalytics/editbl/blob/main/editbl.gif?raw=true)

More introductory examples can be found [here](https://github.com/openanalytics/editbl/blob/main/editbl/R/demoApp.R).
Advanced examples can be found in the [vignettes](https://github.com/openanalytics/editbl/tree/main/editbl/vignettes).


## Features

### Overview of main features

* Supporting multiple backends and in-place editing
* Customizable (lightweight [DT](https://CRAN.R-project.org/package=DT) wrapper)
* Easy integration in [shiny](https://cran.r-project.org/package=shiny) apps
* Undo/redo button
* Copy rows
* Drag cells
* No need to have all data in-memory
* Tackles challenges such as enforcing foreign keys and hiding of surrogate keys
* Transactional commits (currently for `tbl_dbi` class and non in-place editing)
* Default values for new rows (UUID's, 'current date', 'inserted by', ...)
* Possible to set row level security
* Custom logic/checks before saving ([guidelines](https://github.com/openanalytics/editbl/blob/main/editbl/vignettes/howto_custom_logic.rmd))

### Constraints and normalized tables

Sometimes you want to restrict certain columns of your table to only contain specific values.
Many of these restrictions would be implemented at database level through the use of foreign keys to other tables.

`editbl` allows you to specify similar rules through the use of `foreignTbls` as an argument to `eDT()`.
Note that you can additionally hide surrogate keys by the use of `naturalKey` and `columnDefs` if you wish to.

```
a <- tibble::tibble(
    first_name = c("Albert","Donald","Mickey"),
    last_name_id = c(1,2,2)
  )

b <-  foreignTbl(
  a,
  tibble::tibble(
      last_name = c("Einstein", "Duck", "Mouse"),
      last_name_id = c(1,2,3)
    ),
  by = "last_name_id",
  naturalKey = "last_name"
)

eDT(a,
  foreignTbls = list(b),
  options = list(columnDefs = list(list(visible=FALSE, targets="last_name_id")))
)
```

### Support for different backends

`dplyr` code is used for all needed data manipulations and it is recommended to pass on your data as a `tbl`.
This allows editbl to support multiple backends through the usage of other packages like `dtplyr`, `dbplyr` etc.

In case you pass on other tabular objects like `data.frame` or `data.table` the function will internally automatically
cast back and forth to `tbl`. Small side effects may occur because of this (like loosing rownames), so it might be better
to cast yourself to `tbl` explicitly first.

```
# tibble support
modifiedData <- editbl::eDT(tibble::as_tibble(mtcars))

# data.frame support
modifiedData <- editbl::eDT(mtcars)

# data.table support
modifiedData <- editbl::eDT(data.table::data.table(mtcars))

# database support
tmpFile <- tempfile(fileext = ".sqlite")
file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile)
conn <- editbl::connectDB(dbname = tmpFile)
modifiedData <- editbl::eDT(dplyr::tbl(conn, "Artist"), in_place = TRUE)
DBI::dbDisconnect(conn)
unlink(tmpFile)

```

Note that there are some custom methods in the package itself for `rows_update` / `rows_delete` / `rows_insert`. The goal
would be to fully rely on `dplyr` once these functions are not experimental anymore and support all needed requirements.
These functions also explain the high amount of 'suggested' packages, while the core functionality of `editbl` has few
dependencies.

## Switching from DT

Let's say you already use `DT::datatable()` to display your data, but want to switch to `editbl::eDT()` to be able to edit it. Would this be a lot of effort? No!
In fact, `eDT()` accepts the exact same arguments. So it is almost as easy as replacing the functions and you are done.
Should you run into problems take a look [here](https://github.com/openanalytics/editbl/blob/main/editbl/vignettes/howto_switch_from_DT.rmd) for some pointers to look out for.

## Notes

* https://github.com/tidyverse/dtplyr/issues/260 might cause errors / warnings when using `eDT` with `dtplyr`. If possible convert to normal tibble first.
* `editbl` assumes that **all rows in your table are unique**. This assumption is the key (ba dum tss) to allow for only having the data partially in memory.
* `editbl` does not attempt to detect/give notifications on concurrent updates by other users to the same data, nor does it 'lock' the rows you are updating.
It just sends its updates to the backend by matching on the keys of a row. If other users have in the meantime made conflicting adjustments,
the changes you made might not be executed correctly or errors might be thrown.

## Roadmap

* Full `dplyr` compatibility so support for different backends is easily facilitated. Now there are 2 methods (`e_rows_update`, `e_rows_insert`) that need to be implemented to support a new backend.
* Full `DT` compatibility, including all extensions.
* Better editing / display options for time values. E.g. control over timezone and format of display / storage + nicer input forms.
* Any addition that supports the concept of editing data as flexible/easy as possible while respecting backend schema's and constraints.

## References

### Alternatives 

These are other popular CRUD packages in R.
Depending on your needs, they might be better alternatives.

[**DataEditR**](https://cran.r-project.org/package=DataEditR)

* Rstudio plugin
* Really flexible excel-like feeling
* Can only edit in-memory tables. Harder to support databases etc.

[**editData**](https://cran.r-project.org/package=editData)

* Rstudio plugin
* Nice features in terms of editing (pop-ups, more buttons,...)
* Can only edit in-memory tables. Harder to support databases etc.

[**Editor**](https://editor.datatables.net/)

*  Premium datatable extension allowing for editing data.

[**DT-Editor**](https://github.com/jienagu/DT-Editor)

* data.table focused

[**DTedit**](https://github.com/jbryer/DTedit)

* `DT` extension
* Very customizable (own callbacks)
* Few dependencies

### Additional links:
[CRAN `DT`](https://cran.r-project.org/package=DT)

[CRAN `tibble`](https://cran.r-project.org/package=tibble)

[Blogpost spreadsheets vs robust backends](https://www.openanalytics.eu/blog/2023/02/12/spreadsheets-backends-love/)

[Blogpost buttons in DT](https://thatdatatho.com/adding-action-buttons-in-rows-of-dt-data-table-in-r-shiny/)

[Blogpost shiny vs excel](https://appsilon.com/forget-about-excel-use-r-shiny-packages-instead/)

[Generic CRUD application](https://github.com/garrylachman/ElectroCRUD)

[Example SQLite database](https://www.sqlitetutorial.net/sqlite-sample-database/)

## ----echo = TRUE, results = 'hide'--------------------------------------------
library(dplyr)
library(shiny)
library(editbl)
library(dm)

## ----echo = TRUE, results = 'hide'--------------------------------------------
tmpFile <- tempfile(fileext = ".sqlite")
file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile) 
conn <-   DBI::dbConnect(
    dbname = tmpFile,
    drv = RSQLite::SQLite()
)

## ----echo = TRUE, results = 'hide'--------------------------------------------
dm <- dm::dm_from_con(conn, learn_keys = FALSE)

## ----echo = TRUE, results = 'hide'--------------------------------------------
dm <- dm %>%
    dm_add_pk(Artist, ArtistId) %>%
    dm_add_pk(Album, AlbumId) %>%
    dm_add_pk(Customer, CustomerId) %>%
    dm_add_pk(Employee, EmployeeId) %>%
    dm_add_pk(Genre, GenreId) %>%
    dm_add_pk(Invoice, InvoiceId) %>%
    dm_add_pk(InvoiceLine, InvoiceLineId) %>%
    dm_add_pk(MediaType, MediaTypeId) %>%
    dm_add_pk(Playlist, PlaylistId) %>%
    dm_add_pk(PlaylistTrack, c(PlaylistId, TrackId)) %>%
    dm_add_pk(Track, TrackId)
    

## ----echo = TRUE, results = 'hide'--------------------------------------------
dm <- dm %>%
    dm_add_fk(
      table = Album,
      columns = ArtistId,
      ref_table = Artist) %>%
    dm_add_fk(
      table = Invoice,
      columns = CustomerId,
      ref_table = Customer) %>%
    dm_add_fk(
      table = InvoiceLine,
      columns = InvoiceId,
      ref_table = Invoice) %>%
    dm_add_fk(
      table = InvoiceLine,
      columns = TrackId,
      ref_table = Track) %>%
    dm_add_fk(
      table = PlaylistTrack,
      columns = TrackId,
      ref_table = Track) %>%
    dm_add_fk(
      table = PlaylistTrack,
      columns = PlaylistId,
      ref_table = Playlist) %>%
    dm_add_fk(
      table = Track,
      columns = AlbumId,
      ref_table = Album) %>%
    dm_add_fk(
      table = Track,
      columns = MediaTypeId,
      ref_table = MediaType) %>%
    dm_add_fk(
      table = Track,
      columns = GenreId,
      ref_table = Genre)

## -----------------------------------------------------------------------------
dm::dm_flatten_to_tbl(dm, "Album", .recursive = TRUE)

## -----------------------------------------------------------------------------
getForeignTbls <- function(dm, table){
  dm_fks <- dm::dm_get_all_fks(dm)
  dm_fks <- dm_fks[dm_fks$child_table == table,]
  tbl_list <- dm::dm_get_tables(dm)
  
  foreignTbls <- lapply(seq_len(nrow(dm_fks)), function(i){
      r <- dm_fks[i,]   
      x <- tbl_list[r$child_table][[1]]
      y <- dm::dm_flatten_to_tbl(dm, !!(r$parent_table), .recursive = TRUE)
      
      child_fks <- unlist(r$child_fk_cols)
      parent_fks <- unlist(r$parent_key_cols)
      
      # Renaming of parent colums to avoid naming conflicts
      # Done a bit heuristically here for convenience.
      lookup <- parent_fks
      names(lookup) <- child_fks
      other_parent_cols <- setdiff(colnames(y), parent_fks)
      names(other_parent_cols) <- paste(r$parent_table, other_parent_cols, sep = '.')
      lookup <- c(lookup, other_parent_cols)
      y <- y %>% dplyr::rename(all_of(lookup))
      
      editbl::foreignTbl(
        x = x,
        y = y,
        by = child_fks,
        naturalKey = colnames(y)
    )
    
  })
  foreignTbls
  }
  

## -----------------------------------------------------------------------------
dbUI <- function(id) {
  ns <- NS(id)
  fluidPage(
      uiOutput(outputId = ns('selectUI')),
      eDTOutput(id = ns('DT'))
  )
}

dbServer <- function(id, dm) {
  moduleServer(
      id,
      function(input, output, session) {
        ns <- session$ns
        
        tables <- dm::dm_get_tables(dm)
        
        output$selectUI <- renderUI({
          selectInput(ns('table'), label = 'table', choices = names(tables))  
        })
        
        data <- reactive({
              req(input$table)
              tables[input$table][[1]]
              })
        
        foreignTbls <-  reactive({
              req(input$table)
              getForeignTbls(dm, input$table)
              })
        
        eDT(
            id = "DT",
            data = data,
            foreignTbls = foreignTbls,
            in_place = TRUE
        )
          
        invisible()
      }
  )
}


## ----screenshot.opts = list(vwidth = 700, vheight = 500, delay = 1), screenshot.alt = 'screenshots/howto_relational_db_dm_1.png'----
shiny::shinyApp( 
    ui = dbUI('id'),
    server =  function(input, output,session){
      dbServer('id', dm)
    })

## ----echo = TRUE, results = 'hide'--------------------------------------------
library(dplyr)
library(shiny)
library(editbl)

## ----echo = TRUE, results = 'hide'--------------------------------------------
tmpFile <- tempfile(fileext = ".sqlite")
file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile) 
conn <-   DBI::dbConnect(
    dbname = tmpFile,
    drv = RSQLite::SQLite()
)

## -----------------------------------------------------------------------------
DBI::dbListTables(conn)

## -----------------------------------------------------------------------------
dplyr::tbl(conn, 'Album')

## -----------------------------------------------------------------------------
dbUI <- function(id) {
  ns <- NS(id)
  fluidPage(
      eDTOutput(id = ns('Album'))
  )
}

dbServer <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {        
        Album <- eDT(
            id = "Album",
            key = "AlbumId",
            data = dplyr::tbl(conn, "Album"),
            in_place = TRUE 
        )
          
        invisible()
      }
  )
}

## ----screenshot.opts = list(vwidth = 700, vheight = 500), screenshot.alt = 'screenshots/howto_relational_db_1.png'----
shiny::shinyApp( 
    ui = dbUI('id'),
    server =  function(input, output,session){
      dbServer('id', conn)
    })

## -----------------------------------------------------------------------------
dbServer_hidden_keys <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {
        db_album <- dplyr::tbl(conn, "Album")
        db_artist <- dplyr::tbl(conn, "Artist")
        
        Album <- eDT(
            id = "Album",
            data = db_album,
            in_place = TRUE,
            foreignTbls = list(
              foreignTbl(
                x = db_album,
                y = db_artist,
                by = 'ArtistId',
                naturalKey = 'Name'
              )
             ),
             options = list(
               columnDefs = list(
                 list(visible=FALSE, targets=c("AlbumId","ArtistId"))
               )
             )
        )
        invisible()
      }
  )
}

## ----screenshot.opts = list(vwidth = 700, vheight = 500), screenshot.alt = 'screenshots/howto_relational_db_2.png'----
shiny::shinyApp( 
    ui = dbUI('id'),
    server =  function(input, output,session){
      dbServer_hidden_keys('id', conn)
    })

## -----------------------------------------------------------------------------

dbUI_advanced <- function(id) {
  ns <- NS(id)
  fluidPage(
      shiny::uiOutput(ns("artistSelector_UI")),
      eDTOutput(id = ns('Album'))
  )
}

dbServer_advanced <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {
        ns <- session$ns
                
                
        db_album <- dplyr::tbl(conn, "Album")
        db_artist <- dplyr::tbl(conn, "Artist")
        
        output$artistSelector_UI <- shiny::renderUI(
          shiny::selectInput(ns('artist'),
           label = 'artist',
           choices = db_artist %>% select(Name) %>% collect())
        )

        Album <- eDT(
            id = "Album",
            data = db_album,
            in_place = TRUE,
            foreignTbls = reactive(
            {
             req(input$artist)
             selected <- input$artist     
             list(
              foreignTbl(
                x = db_album,
                y = db_artist %>% filter(Name == selected),
                by = 'ArtistId',
                naturalKey = 'Name'
              )
             )}),
             options = list(
               columnDefs = list(
                 list(visible=FALSE, targets=c("AlbumId","ArtistId"))
               )
             )

        )
        invisible()
      }
  )
}

## ----screenshot.opts = list(vwidth = 700, vheight = 500), screenshot.alt = 'screenshots/howto_relational_db_3.png'----
shiny::shinyApp( 
    ui = dbUI_advanced('id'),
    server =  function(input, output,session){
      dbServer_advanced('id', conn)
    })

## ----echo = TRUE, results = 'hide'--------------------------------------------
library(shiny)
library(DT)
library(editbl)

## ----screenshot.opts = list(vwidth = 700), screenshot.alt = 'screenshots/howto_switch_from_DT_1.png'----
ui <- fluidPage(DTOutput("DT"))
server <- function(input, output, session){
  output$DT <- renderDataTable({
    datatable(mtcars)  %>%
     formatRound('disp', 1)
  })  
  observe({
      print(input[["DT_cell_clicked"]])
   })
}
shinyApp(ui, server)

## ----screenshot.opts = list(vwidth = 700), screenshot.alt = 'screenshots/howto_switch_from_DT_2.png'----
ui <- fluidPage(eDTOutput("DT"))
server <- function(input, output, session){
  editbl::eDT(
    id = "DT",
    data = mtcars,
    format = function(x){formatRound(x,'disp', 1)})
   
   observe({
      print(input[["DT-DT_cell_clicked"]])
   })
}
shinyApp(ui, server)

## ----screenshot.opts = list(vwidth = 700, vheight = 500), , screenshot.alt = 'screenshots/howto_row_level_access_1.png'----
library(editbl)
library(shiny)
conn <- DBI::dbConnect(RSQLite::SQLite(), "")
df <- data.frame(
    user = c("Albert","Donald","Mickey"),
    email = c('albert@einstein.com', 'donald@duck.com', 'mickey@mouse.com')
)
DBI::dbWriteTable(conn, "characters", df)
tibble <- dplyr::tbl(conn, 'characters')

CURRENT_USER = 'Mickey'

shiny::shinyApp( 
    ui = editbl::eDTOutput('id'),
    server =  function(input, output,session){
      result <- eDT(id='id',
        data = tibble %>% filter(user == CURRENT_USER),
        in_place = TRUE
      )
    })

print(tibble)

DBI::dbDisconnect(conn)

## ----screenshot.opts = list(vwidth = 700, vheight = 500), , screenshot.alt = 'screenshots/howto_row_level_access_2.png'----
library(editbl)
df <- tibble::tibble(
    user = c("Albert","Donald","Mickey"),
    email = c('albert@einstein.com', 'donald@duck.com', 'mickey@mouse.com')
)

CURRENT_USER = 'Mickey'

rowModificationLogic <- function(row){
        if (row[,'user'] == CURRENT_USER){
          TRUE
        } else {
          FALSE  
        }
    }

shiny::shinyApp( 
    ui = editbl::eDTOutput('id'),
    server =  function(input, output,session){
      eDT(id='id',
        data = df,
        canEditRow = rowModificationLogic,
        canDeleteRow = rowModificationLogic
        )
    })

## ----echo = TRUE, results = 'hide'--------------------------------------------
library(shiny)
library(editbl)

## -----------------------------------------------------------------------------
ui <- fluidPage(eDTOutput("DT"))

# Define a custom update function
e_rows_update.mtcars <- function(
    x,
    y,
    by = NULL,
    ...,
    match = match,
    unmatched = c("error", "ignore"),
    copy = FALSE,
    in_place = FALSE){		
		
		# Extra checks
		if(any(y[,"mpg"] < 0)){
			stop("mpg should be a positive value!")	
		}
		
		# Logging
		print("Updated rows:")
		print(y)
		
		# Drop extra class and execute default code
		class(x) <- class(x)[class(x) != 'mtcars']
		e_rows_update(
	      x = x,
	      y = y,
	      by = by,
	      ... ,
	      match = match,
	      unmatched = unmatched,
	      copy = copy,
	      in_place = in_place
	  )
	}

server <- function(input, output, session){
	
	# Add a new class to your data object
	df = tibble::tibble(mtcars)
	class(df) <- c("mtcars", class(df))

	  result = eDT(
	    id = "DT",
	    data = df
	)
	   
}
shinyApp(ui, server)

## -----------------------------------------------------------------------------
ui <- fluidPage(eDTOutput("DT"))

server <- function(input, output, session){
	  result = eDT(
	    id = "DT",
	    data = mtcars
	  )
	  
	  # Some logging
	  observe({
		print('These rows have been permanently deleted:')
		print(result$deleted())
	  })
	   
}
shinyApp(ui, server)

#' Run a demo app
#' @param app demoApp to run. Options: database / mtcars / custom
#' 
#' @details These apps are for illustrative purposes.
#' 
#' @param ... arguments passed onto the demoApp
#' @examples 
#' ## Only run this example in interactive R sessions
#' if(interactive()){
#'  
#'   # Database
#'   tmpFile <- tempfile(fileext = ".sqlite")
#'   file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile)
#'
#'   conn <- connectDB(dbname = tmpFile)
#'
#'   runDemoApp(app = "database", conn = conn)
#'   DBI::dbDisconnect(conn)
#'
#'   unlink(tmpFile)
#'
#'   # mtcars
#'   runDemoApp(app = "mtcars")
#'
#'   # Any tibble of your liking
#'   runDemoApp(app = "custom", dplyr::tibble(iris))
#' }
#' @inherit shiny::shinyApp return
#' @export 
runDemoApp <- function(app = "database", ...){
  fn <- switch(app,
      "database" = runDemoApp_DB,
      "mtcars" = runDemoApp_mtcars,
      "custom" = runDemoApp_custom,
      stop("demoApp not found")
      )
  do.call(fn, list(...))
}

#' Run a demo app
#' @importFrom shiny shinyApp
#' @inherit shiny::shinyApp return
#' @keywords internal
runDemoApp_DB <- function(){
  tmpFile <- tempfile(fileext = ".sqlite")
  file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile)
  conn <- editbl::connectDB(dbname = tmpFile)

  ui <- demoUI_DB(id = "app", conn = conn)
  server <- function(input, output, session) {
    demoServer_DB(id = "app", conn = conn)
  }
  shinyApp(ui, server)  
}

#' Run a demo app
#' @importFrom shiny shinyApp
#' @inherit shiny::shinyApp return
#' @keywords internal
runDemoApp_mtcars <- function(){
  ui <- demoUI_mtcars(id = "app")
  server <- function(input, output, session) {
    demoServer_mtcars(id = "app")
  }
  shinyApp(ui, server)  
}

#' Run a custom demo app
#' @param x `tbl`
#' @importFrom shiny shinyApp
#' @inherit shiny::shinyApp return
#' @keywords internal
runDemoApp_custom <- function(x){
  ui <- demoUI_custom(id = "app")
  server <- function(input, output, session) {
    demoServer_custom(id = "app", x= x)
  }
  shinyApp(ui, server)  
}

#' UI of the DB demo app
#' @param id `character(1)`
#' @param conn database connection object as given by \code{\link[DBI]{dbConnect}}.
#' @importFrom shiny NS tagList selectInput verbatimTextOutput
#' @return HTML
#' 
#' @author Jasper Schelfhout
#' @keywords internal
demoUI_DB <- function(id, conn) {
  ns <- NS(id)
  tagList(
      selectInput(
          inputId = ns("table"),
          label = "table",
          choices = DBI::dbListTables(conn)),
      eDTOutput(id = ns("editbl"))
  )
}

#' Server of the DB demo app
#' @param id `character(1)`
#' @param conn database connection object as given by \code{\link[DBI]{dbConnect}}.
#' @importFrom shiny reactive moduleServer renderPrint
#' @importFrom dplyr tbl
#' @return NULL, just executes the module server.
#' @author Jasper Schelfhout
#' @keywords internal
demoServer_DB <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {
        data <- reactive({
              dplyr::tbl(conn, input$table)
            })
        
        modifiedData <- eDT(
            id = "editbl",
            data = data,
            in_place = TRUE 
            )

        invisible()
      }
  )
}

#' UI of the demo mtcars app
#' @param id `character(1)`
#' @importFrom shiny NS tagList
#' @return HTML
#' 
#' @author Jasper Schelfhout
#' @keywords internal
demoUI_mtcars <- function(id) {
  demoUI_custom(id)
}

#' Server of the mtcars demo app
#' @param id `character(1)`
#' @importFrom dplyr tibble
#' @inherit demoServer_custom return
#' @author Jasper Schelfhout
#' @keywords internal
demoServer_mtcars <- function(id) {
  demoServer_custom(id, dplyr::tibble(datasets::mtcars))
}

#' UI of the demo mtcars app
#' @param id `character(1)`
#' @importFrom shiny NS tagList
#' @return HTML
#' 
#' @author Jasper Schelfhout
#' @keywords internal
demoUI_custom <- function(id) {
  ns <- NS(id)
  tagList(
      eDTOutput(id = ns("editbl")),
  )
}

#' Server of the mtcars demo app
#' @param id `character(1)`
#' @param x `tbl`
#' @return NULL, just executes the module server.
#' @author Jasper Schelfhout
#' @keywords internal
demoServer_custom <- function(id, x) {
  moduleServer(
      id,
      function(input, output, session) {        
        modifiedData <- eDT(
            id = "editbl",
            data = x,
            in_place = FALSE 
        )$result
        
        invisible()
      }
  )
}

test_that("Starting demo app db works", {
      runDemoApp (app = "database")
      expect_true(TRUE)
    })

test_that("Starting demo app mtcars works", {
      runDemoApp (app = "mtcars")
      expect_true(TRUE)
    })

test_that("Starting demo app custom works", {
      runDemoApp(app = "custom", dplyr::tibble(iris))
      expect_true(TRUE)
    })

#' Create a foreign tibble
#' 
#' @details This is a tibble that can be passed onto \code{\link{eDT}} as a referenced table.
#' 
#' It is the equivalent of a database table to which the `data` tbl of eDT has a foreign key.
#' 
#' It will be merged with the tbl passed onto the `data` argument allowing to provide restrictions
#' for certain columns.
#' 
#' Note that row uniqueness for the columns used in `by` and `naturalKey` is assumed.
#' This assumption will however not be checked since it is an expensive operation on big datasets.
#' However, if violated, it might give errors or unexpected results during usage of the eDT module.
#' 
#' @param x  `tbl`. The referencing table.
#' @param y `tbl`. The referenced table.
#' @param by `character`. Column names to match on.
#' Note that you should rename and/or typecast the columns in y should they not exactly match the columns in x.
#' @param naturalKey `character`. The columns that form the natural key in y.
#' These are the only ones that can actually get modified in `eDT` when changing cells in the table. 
#' Reasoning being that these columns should be sufficient to uniquely identify a row in the referenced table.
#' All other columns will be automatically fetched and filled in.
#' 
#' @param allowNew `logical`. Whether or not new values are allowed. If `TRUE`,
#' the rows in the foreignTbl will only be used as suggestions, not restrictions.
#' 
#' @examples 
#' a <- tibble::tibble(
#'    first_name = c("Albert","Donald","Mickey"),
#'    last_name_id = c(1,2,2)
#'  )
#'
#' b <-  foreignTbl(
#'  a,
#'  tibble::tibble(
#'      last_name = c("Einstein", "Duck", "Mouse"),
#'      last_name_id = c(1,2,3)
#'    ),
#'  by = "last_name_id",
#'  naturalKey = "last_name"
#')
#' 
#' ## Only run this in interactive R sessions
#' if(interactive()){
#'   eDT(a,
#'    foreignTbls = list(b),
#'    options = list(columnDefs = list(list(visible=FALSE, targets="last_name_id")))
#'   )
#'  }
#'
#' 
#' @return List with unmodified arguments. However, they have now been checked for validity.
#' - y, see argument `y`.
#' - by, see argument `by`.
#' - naturalKey, see argument `naturalKey`.
#' - allowNew, see argument `allowNew`
#' 
#' @importFrom dplyr tbl_vars all_of select
#' 
#' @author Jasper Schelfhout
#' @export
foreignTbl <- function(
    x,
    y,
    by = intersect(dplyr::tbl_vars(x), dplyr::tbl_vars(y)),
    naturalKey = dplyr::tbl_vars(y),
    allowNew = FALSE
){
  stopifnot(all(naturalKey %in% colnames(y)))
  stopifnot(is.logical(allowNew))
  if(is.null(names(by))){
    names(by) <- by
  }
  
  clashingNames <- intersect(
      setdiff(dplyr::tbl_vars(y), by),
      dplyr::tbl_vars(x)
  )
  if(length(clashingNames)){
    stop(sprintf("Name clashes on: %s. Please rename.", paste(clashingNames, collapse = ", ")))
  }
  
  x_types <- x %>% select(all_of(by)) %>% getColumnTypeSums %>% unlist()
  y_types <- y %>% select(all_of(by)) %>% getColumnTypeSums %>% unlist()
  typeMisMatches <- names(x_types[x_types != y_types])
  if(length(typeMisMatches)){
    stop(sprintf("%s is not of the same type: %s vs %s. Try casting with dplyr::mutate()",
            typeMisMatches,
            x_types[typeMisMatches],
            y_types[typeMisMatches]))
  }
  
  foreignTbl <- list(
      y = y,
      by = by,
      naturalKey = naturalKey,
      allowNew = allowNew
  )
  
  return(foreignTbl)
}

#' Merge a tbl with it a foreignTbl
#' 
#' @details see also `dplyr` join functions, for example `dplyr::left_join`.
#'  
#' @param tbl `tbl`
#' @param foreignTbl `list` as created by \code{\link{foreignTbl}}
#' @param keepNA `logical` keep rows from tbl with NA keys.
#' @param by named `character`, columns to join on.
#' @param copy `logical`, whether or not to copy the `foreignTbl` to the source of argument `tbl` for joining.
#' @param type `character(1)`, type of joint to perform. Can be 'inner' or 'left'.
#' @return `tbl`, containing both columns from argument `tbl` and argument `foreignTbl`.
#' 
#' @importFrom dplyr left_join inner_join filter union
#' 
#' @author Jasper Schelfhout
joinForeignTbl <- function(
    tbl,
    foreignTbl,
    keepNA = TRUE,
    by = foreignTbl$by,
    copy = TRUE,
    type = c("inner", "left")[1]){  
  .data <- NULL # for R CMD CHECK
  if(is.null(names(by))){
    names(by) <- by
  }
  
  if(type == "inner"){
    result <- dplyr::inner_join(
        x = tbl,
        y = foreignTbl$y,
        by = by,
        copy = copy)
  } else if(type == "left"){
    result <- dplyr::left_join(
        x = tbl,
        y = foreignTbl$y,
        by = by,
        copy = copy)
  } else {
    stop("only inner and left join supported")
  }
  
  
  if(keepNA && type == "inner"){
    naTbl <- tbl
    for(key in names(by)){      
      naTbl <- dplyr::filter(naTbl, is.na(.data[[key]]))
    }
    
    naTbl <- dplyr::left_join(
        x = naTbl,
        y = foreignTbl$y,
        by = by,
        copy = copy
    )
    result <- dplyr::union(result, naTbl)
  }
  
  return(result)
}

#' Fill data columns based on foreignTbls
#' 
#' @details
#' When a combination of columns is not found in the foreignTbl,
#' fill the deductedColumns with NA.
#' 
#' @details on foreignTbls suggesting conflicting data,
#' an arbitrary choice is made. It is best to afterwards check with
#' checkForeignTbls to see if a valid result is obtained.
#' 
#' @param tbl `tbl`
#' @param foreignTbls  list of foreign tbls as created by \code{\link{foreignTbl}}
#' @return tbl
#' 
#' @author Jasper Schelfhout
#' @keywords internal
fillDeductedColumns <- function(tbl, foreignTbls){
  columnOrder <- names(tbl)
  nrow <- nrow(tbl)
  
  # Clear columns that should be deducted
  for (foreignTbl in foreignTbls){
    autoFill <- setdiff(colnames(foreignTbl$y), foreignTbl$naturalKey)
    tbl[,autoFill] <- NULL
  }
  
  # Fill in columns
  for(foreignTbl in foreignTbls){
    tbl <- joinForeignTbl(tbl, foreignTbl, by = foreignTbl$naturalKey, type = "left")
  }
  
  newNrow <- nrow(tbl)
  
  if(nrow < newNrow){
    stop("One of the given foreign tbls has non unique keys.")
  }
  
  tbl[,columnOrder]
}

#' Check if all rows in tbl fufill `foreignTbl` constraints
#' @param tbl `tbl`
#' @param foreignTbls  list of foreign tbls as created by \code{\link{foreignTbl}}
#' @return `logical` stating if tbl fufills all constraints imposed by all foreign tbls.
#' @importFrom dplyr count pull anti_join filter if_any all_of
#' 
#' @author Jasper Schelfhout
#' @keywords internal
checkForeignTbls <- function(tbl, foreignTbls){
  n <- NULL # R CMD CHECK fix
  
  
  for(foreignTbl in foreignTbls){
    # match on combination on naturalKey and surrogateKey
    # E.g. check that the row actually exists in the foreignTbl
    if(foreignTbl$allowNew){
      next()
    }
    by <- unique(unname(c(foreignTbl$by, foreignTbl$naturalKey)))
    
    nonExisting <- dplyr::anti_join(tbl, foreignTbl$y,
            by = by,
            copy = TRUE)  %>% 
        dplyr::filter(if_any(all_of(by), ~ !is.na(.))) # Do not complain about empty rows, might be nice as a parameter.
    
    if(dplyr::pull(dplyr::count(nonExisting), n)){
      stop(sprintf("Invalid %s: %s",
              paste(foreignTbl$naturalKey, collapse = ", "),
              paste(as.character(nonExisting[1,foreignTbl$naturalKey]), collapse = ", ")
          )) 
    }
  }
  TRUE
}

#' Cast all columns that exist in a foreignTbl to factor
#' 
#' @details Can be used to fixate possible options when editing.
#' 
#' @param data `data.frame`
#' @param foreignTbls  list of foreign tbls as created by \code{\link{foreignTbl}}
#' @importFrom dplyr distinct select pull tbl_vars
#' @return data.frame
#' 
#' @author Jasper Schelfhout
#' @keywords internal
castToFactor <- function(data, foreignTbls){
  levels <- list()
  for(foreignTbl in foreignTbls){
    for(column in dplyr::tbl_vars(foreignTbl$y)){
      currentLevels <- levels[[column]]
      newRestrictions <- dplyr::pull(dplyr::distinct(dplyr::select(foreignTbl$y, column)))
      if(is.null(currentLevels)){
        levels[[column]] <- newRestrictions
      } else {
        levels[[column]] <- intersect(currentLevels, newRestrictions)
      }
    }
  }
  
  for(column in intersect(dplyr::tbl_vars(data), names(levels))){
    data[[column]] <- factor(data[[column]], levels = levels[[column]])
  }
  data
}

#' Get all columns that are not natural keys
#' @param foreignTbls  list of foreign tbls as created by \code{\link{foreignTbl}}
#' @return `character`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
getNonNaturalKeyCols <- function(foreignTbls){
  result <- c()
  for(foreignTbl in foreignTbls){
    cols <- as.character(dplyr::tbl_vars(foreignTbl$y))
    naturalKey <- foreignTbl$naturalKey
    result <- unique(c(result, setdiff(cols, naturalKey)))
  }
  result
}

test_that("passing on a foreign tibble works", {
      a <- tibble::tibble(
              a = c(1,2,3),
              key1 = c(NA,2,4),
              key2 = c(NA,2,4))
      
      b <-  foreignTbl(
          a,
          tibble::tibble(
                  b = c("a", "b", "c"),
                  key1 = c(1,2,3),
                  key2 = c(1,2,3)
              ),
          naturalKey = "b"
      )
      
      # Try running the app manually
      ui <- eDTOutput("app")
      server <- function(input,output,session){
        eDT(id = "app",
            data = reactive(a),
            foreignTbls = list(b),
            options = list(columnDefs = list(list(visible=FALSE, targets="key2"))))
      }
      shiny::shinyApp(ui, server)
      
      # Actual test
      shiny::testServer(
          app = eDTServer,
          args = list(data = reactive(a), foreignTbls = list(b)),
          expr = {
            expect_true(TRUE)
          })
})

test_that("Can use foreign tibbles to fill in non natural key values.",{
      tbl <- tibble::tibble(
              a = c(1,2),
              key1 = as.double(c(NA,NA)),
              key2 = as.double(c(NA,NA))
          )
      
      merged_tbl <- cbind(tbl, tibble(
              b = c("b1", "b2"), 
              c = c("c1", "c2")))

      
      b <-  foreignTbl(
          tbl,
          tibble::tibble(
                  b = c("b1", "b2", "b3"),
                  key1 = c(1,2,3)
              ),
          by = "key1",
          naturalKey = "b"
      )
      
      c <-  foreignTbl(
          tbl,
          tibble::tibble(
                  c = c("c1", "c2", "c3"),
                  key2 = c(1,2,3)
              ),
          by = "key2",
          naturalKey = "c"
      )
      
      foreignTbls = list(b, c)
      
      result <- fillDeductedColumns(merged_tbl, foreignTbls)
      
      expect_equal(result$key1, c(1,2))
      expect_equal(result$key2, c(1,2))
    })

test_that("Contradicting foreign tibbles give error when filling in data.",{
      tbl <- tibble::tibble(
              a = c(1,2),
              key = as.double(c(NA,NA))
          )
      
      merged_tbl <- tibble::tibble(cbind(tbl,
              tibble::tibble(color = c("blue", "orange"))
          ))
      
      b <-  foreignTbl(
          tbl,
          tibble::tibble(
                  color = c("blue", "orange"),
                  key = c(1,2)
              ),
          by = "key",
          naturalKey = "color"
      )
      
      c <-  foreignTbl(
          tbl,
          tibble::tibble(
                  color = c("blue", "purple"),
                  key = c(2,1)
              ),
          by = "key",
          naturalKey = "color"
      )
            
      expect_error({fillDeductedColumns(tbl, foreignTbls = list(b,c))})
      expect_error({fillDeductedColumns(tbl, foreignTbls = list(c,b))})
      
    })

test_that("Empty foreign tibbles list returns given tibble",{
      tbl <- tibble::tibble(
              a = c(1,2),
              key = as.double(c(NA,NA)),
              color = c("blue", "orange")
          )
          
      result <- fillDeductedColumns(tbl, foreignTbls = list())
      expect_equal(result, tbl)
    })

test_that("checkForeignTbls throws error",{
      
      tbl <-  tibble::tibble(
              a = c(1,2, 3),
              key = c(1,2, 3)
          )
      
      merged_tbl <- cbind(tbl,
              tibble::tibble(color = c("blue", "orange", "purple"))
          )
      
      b <-  foreignTbl(
          tbl,
          tibble::tibble(
                  color = c("blue", "orange"),
                  key = c(1,2)
              ),
          by = "key",
          naturalKey = "color"
      )
      
      expect_error(checkForeignTbls(tbl = merged_tbl,foreignTbls = list(b)))
    })

test_that("joinForeignTbl() type 'inner' works.",{
      a <- tibble::tibble(
           a = c(1,2,3),
           key1 = c(NA,2,4),
           key2 = c(NA,2,4))
       
       b <-  tibble::tibble(
           b = c("a", "b", "c"),
           key1 = c(1,2,3),
           key2 = c(1,2,3)
       )
       
       foreignTbl <- foreignTbl(a,b)
       
       result <- joinForeignTbl(a, foreignTbl, type = 'inner')
       
       # NA keys in 'a' are expected to stay.
       expect_equal(colnames(result), c("a", "key1", "key2", "b"))
       expect_true(nrow(result) == 2)
    })

test_that("joinForeignTbl() type 'left' works.",{
      a <- tibble::tibble(
          a = c(1,2,3),
          key1 = c(NA,2,4),
          key2 = c(NA,2,4))
      
      b <-  tibble::tibble(
          b = c("a", "b", "c"),
          key1 = c(1,2,3),
          key2 = c(1,2,3)
      )
      
      foreignTbl <- foreignTbl(a,b)
      
      result <- joinForeignTbl(a, foreignTbl, type = 'left')
      
      expect_equal(colnames(result), c("a", "key1", "key2", "b"))
      expect_true(nrow(result) == 3)
    })

#' rows_update implementation for data.frame backends.
#' @inheritParams e_rows_update
#' @inherit e_rows_update return details
#' @author Jasper Schelfhout
#' @export
e_rows_update.data.frame <- function(x, y, by = NULL, match = NULL,..., copy = FALSE, in_place = FALSE){
  if(in_place){
    stop("Can not edit in place")
  }
  
  if(is.null(by)){
    by <- colnames(x)[1]
  }
  
  if(is.null(match)){
    match <- list(x = y[by], y = y[by])
  }
  
  idx <- vctrs::vec_match(match$x, x[by])
  
  bad <- which(is.na(idx))
  if (length(bad)) {
    stop("Attempting to update missing rows.")
  }
  
  x[idx, names(y)] <- y
  
  x
}

#' rows_insert implementation for `data.table` backends.
#' 
#' @inherit e_rows_insert
#' 
#' @author Jasper Schelfhout
#' @export
e_rows_insert.dtplyr_step <- function(x, y, by = NULL, ..., copy = FALSE, in_place = FALSE){  
  x_dt <- data.table::copy(data.table::as.data.table(x))
  y_dt <- data.table::as.data.table(y)
  
  if(is.null(by)){
    by <- colnames(x)[1]
  }
  
  if(in_place){
    stop("Adding rows by reference to a data.table is not supported yet.")
  } else {
    result <- dtplyr::lazy_dt(rbind(x_dt,y_dt))
  }
  
  return(result)
}

#' rows_delete implementation for data.table backends.
#' @inheritParams dplyr::rows_delete
#' @inherit dplyr::rows_delete return details
#' @export
#' @author Jasper Schelfhout
rows_delete.dtplyr_step <- function(x, y, by = NULL, ..., unmatched, copy = FALSE, in_place = FALSE){  
  x_dt <- data.table::copy(data.table::as.data.table(x))
  y_dt <- data.table::as.data.table(y)
  
  if(!nrow(y_dt)){
    return(x)
  }
  
  if(is.null(by)){
    by <- colnames(x_dt)[1]
  }
  
  if(in_place){
    stop("In place deletes for data.tables are not supported yet.
 See issue: https://github.com/Rdatatable/data.table/issues/635")
  }
  
  matches <- unlist(lapply(seq_len(nrow(x_dt)), function(i){
        nrow(merge(x_dt[i,], y_dt, by = by)) > 0
      }))
  
  x_dt <- x_dt[!matches,]
  result <- dtplyr::lazy_dt(x_dt)
  
  result
}



#' rows_update implementation for data.table backends.
#' @inherit e_rows_update
#' @author Jasper Schelfhout
#' @export
e_rows_update.dtplyr_step <- function(x, y, by = NULL, match = NULL,..., copy = FALSE, in_place = FALSE){
  args <- c(as.list(environment()), list(...))
  
  x_dt <- data.table::as.data.table(x)
  y_dt <- data.table::as.data.table(y)
  
  if(!nrow(y_dt)){
    return(x)
  }
  
  if(is.null(by)){
    by <- colnames(x_dt)[1]
  }
  
  if(is.null(args$match)){ # Be explicit about argument since otherwhise base::match will used.
    match <- list(x = y_dt[,by, with = FALSE, drop = FALSE],
        y = y_dt[,by, with = FALSE, drop = FALSE])
  }
  
  for (i in seq_len(nrow(match$y))){
    yMatch <-  match$y[i,, drop = FALSE]
    values <- merge(yMatch, y, by = by)[,colnames(y_dt)]
    xMatch <-  match$x[i,, drop = FALSE]
    
    xRows <- which(unlist(lapply(seq_len(nrow(x_dt)), function(i){
                  nrow(merge(x_dt[i,], xMatch, by = by)) > 0
                })))
    
    x_dt[xRows,] <- values
  }
  
  result <- dtplyr::lazy_dt(x_dt)
  return(invisible(result))  
}

#' UI part of \code{\link{eDT}}
#' 
#' @details Works exactly like \code{\link[DT]{DTOutput}} apart from the fact that instead of the `outputId`
#' argument, `id` is requested. Reason being that this function is a UI to a shiny module.
#' This means that the datatable can be found under the id \code{'{namespace}-{id}-DT'} instead of \code{'{namespace}-{outputId}'}.
#' 
#' Also some minor CSS and javascript is executed for functional puposes.
#' 
#' @param id `character(1)`
#' @param ... arguments passed to \code{\link[DT]{DTOutput}} 
#' @importFrom DT DTOutput
#' @importFrom shiny actionButton tagList HTML tags fluidPage tags
#' @importFrom shinyjs disabled useShinyjs hidden
#' @inherit eDT examples
#' @return HTML
#' 
#' @author Jasper Schelfhout
#' @export
eDTOutput <- function(id,...) {
  ns <- NS(id)
  
  # input$current_id to detect clicked row.
  js <- HTML(
      sprintf("function get_id(clicked_id, ns) {
              Shiny.setInputValue(ns+\"current_id\", clicked_id, {priority: \"event\"});
              };
              document.addEventListener(\"keydown\", (event) => {
              if (event.ctrlKey && event.key === \"z\") {
              Shiny.setInputValue(\"%1$s\",  Math.random(), {priority: \"event\"});
              }});
              document.addEventListener(\"keydown\", (event) => {
              if (event.ctrlKey && event.key === \"y\") {
              Shiny.setInputValue(\"%2$s\",  Math.random(), {priority: \"event\"});
              }});",
          ns("undo"), ns("redo")
      )
  )
  
  fluidPage(
      shinyjs::useShinyjs(),
      tags$script(js),
      # Hack that ensures fontawesome is properly loaded
      shinyjs::hidden(actionButton(ns("activate_shiny_css"), label = "hidden", icon = icon("plus"))),
      
      tags$style(HTML(disableDoubleClickButtonCss(ns("DT")))),   
      DT::DTOutput(outputId = ns("DT"), ...)
  )
}

#' Create a modifieable datatable.
#' 
#' @details Works the same as \code{\link[DT]{datatable}}.
#' This function is however a shiny module and comes with additional arguments and different defaults.
#' Instead of having `output$id = renderDT(DT::datatable(iris))`, `eDT(id = 'id', data = iris)` should be used on the server side.
#' On the UI side \code{\link{eDTOutput}} should be used instead of \code{\link[DT]{DTOutput}}.
#' 
#' @details Can also be used as standalone app when not ran in reactive context.
#' @details All arguments except 'id' and 'env' can be normal objects or reactive objects.
#' 
#' @param id `character(1)` module id
#' @param data `tbl`. When using non-standard tibble objects, install the required package implementing specific methods.
#' (`data.table`: dtplyr, spark: sparklyr, generic database: dbplyr,...).
#' While other tabular objects like `data.frame` are supported, explicitly casting to `tbl` is recommended.
#' @inheritParams DT::datatable
#' @param keys `character`. Defaults to all columns under the assumption that at least every row is unique.
#' @param format function accepting and returning a \code{\link[DT]{datatable}}
#' @param in_place `logical`. Whether to modify the data object in place or to return a modified copy.
#' @param foreignTbls `list`. List of objects created by \code{\link{foreignTbl}}
#' @param columnOrder `vector`. Order of columns, original data and foreignTbls combined. Defaults to no order specified.
#' @param statusColor named `character`. Colors to indicate status of the row.
#' @param inputUI `function`. UI function of a shiny module with at least arguments `id` `data` and `...`.
#' #'   elements with inputIds identical to one of the column names are used to update the data.
#' @param defaults expression that evaluates to a `tibble` with (a subset of) columns of the data.
#'   It will be evaluated for each new row in the environment defined by 'env'.
#'   This allows for defaults like Sys.time() or uuid::UUIDgenerate() as well as dynamic inputs.
#' @param cloneDefaults expression that evaluates to a `tibble` with (a subset of) columns of the data.
#'   It will be evaluated for each cloned row in the environment defined by 'env'.
#' @param env `environment` in which the server function is running. Should normally not be modified.
#' @inheritParams canXXXRowTemplate
#' @param utilityColumns named character vector. Defines names for (hidden) utility columns
#'   used by `eDT` to keep track of modifications. Should normally only be adjusted in rare case of name clashes with data.
#'  ``` 
#'  c(
#'    status = '_editbl_status',
#'    buttons = '_editbl_buttons',
#'    identity = '_editbl_identity',
#'    deleted = '_editbl_deleted'
#'   )
#'  ```
#' @return list
#' - result `reactive` modified version of `data` (saved)
#' - state `reactive` current state of the `data` (unsaved)
#' - selected `reactive` selected rows of the `data` (unsaved)
#' - edited `reactive` edited rows (saved)
#' - inserted `reactive` inserted rows (saved)
#' - deleted `reactive` deleted rows (saved)
#' 
#' @examples 
#' ## Only run this example in interactive R sessions
#' if(interactive()){
#'   # tibble support
#'   modifiedData <- editbl::eDT(tibble::as_tibble(mtcars))
#' 
#'   # data.table support
#'   modifiedData <- editbl::eDT(dtplyr::lazy_dt(data.table::data.table(mtcars)))
#' 
#'   # database support
#'   tmpFile <- tempfile(fileext = ".sqlite")
#'   file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile)
#' 
#'   conn <- editbl::connectDB(dbname = tmpFile)
#'   modifiedData <- editbl::eDT(dplyr::tbl(conn, "Artist"), in_place = TRUE)
#'   DBI::dbDisconnect(conn)
#' 
#'   unlink(tmpFile)
#' 
#'   # Within shiny
#'   library(shiny)
#'   library(editbl)
#'   shinyApp(
#'     ui = fluidPage(fluidRow(column(12, eDTOutput('tbl')))),
#'     server = function(input, output) {
#'       eDT('tbl',iris,)
#'     }
#'   )
#' 
#'   # Custom inputUI
#'   editbl::eDT(mtcars, inputUI = function(id, data){
#'     ns <- NS(id)
#'     textInput(
#'     ns("mpg"),
#'     label = "mpg",
#'     value = data$mpg)})
#' 
#'   # Do not allow delete
#'   editbl::eDT(mtcars, canDeleteRow = FALSE)
#' }
#' 
#' @author Jasper Schelfhout
#' @export
eDT <- function(
    data,
    options = list(
        dom = 'Bfrtlip',
        keys = TRUE,
        ordering = FALSE,
        autoFill = list(update = FALSE, focus = 'focus'),
        buttons = list("add", "undo", "redo", "save")
    ),
    class = "display",
    callback = NULL,
    rownames = FALSE,
    colnames = NULL,
    container,
    caption = NULL,
    filter = c("none", "bottom", "top"),
    escape = TRUE,
    style = "auto",
    width = NULL,
    height = NULL,
    elementId = NULL,
    fillContainer = getOption("DT.fillContainer", NULL),
    autoHideNavigation = getOption("DT.autoHideNavigation", NULL),
    selection = "none",
    extensions = c('KeyTable', 'AutoFill', "Buttons"),
    plugins = NULL,
    editable = list(target = "cell"),
    id,
    keys = NULL,
    in_place = FALSE,
    format = function(x){x},
    foreignTbls = list(),
    columnOrder = c(),
    statusColor = c("insert"="#e6e6e6", "update"="#32a6d3", "delete"="#e52323"),
    inputUI = editbl::inputUI,
    defaults = tibble(),
    cloneDefaults = tibble(),
    env = environment(),
    canEditRow = TRUE,
    canDeleteRow = TRUE,
    canCloneRow = TRUE,
    utilityColumns = NULL
) {
  args <- as.list(environment())
  
  # if not in reactive context start standalone app
  if(is.null(shiny::getDefaultReactiveDomain())){
    if(missing(id)){
      args$id <- "nevergonnaletyoudown"
    }
    # FIXME probably a better way to deal with missing arguments
    if(missing(container)){
      args$container <- NULL
    }
    result <- do.call(eDT_app, args)
  } else {
    if(missing(id)){
      stop("Please specify an id")
    }
    result <- do.call(eDTServer, args)
  }
  result
}


#' @inheritParams eDT
#' @importFrom shiny moduleServer observe reactiveValues reactive 
#'  observeEvent actionButton icon renderPrint showNotification req
#'  isolate is.reactive modalDialog modalButton renderUI uiOutput showModal
#'  freezeReactiveValue isTruthy
#' @importFrom DT dataTableProxy renderDT formatStyle styleEqual hideCols
#' @importFrom dplyr collect %>% relocate is.tbl all_of tibble
#' @importFrom utils str tail
#' @importFrom uuid UUIDgenerate
#' @importFrom shinyjs disable enable
#' @author Jasper Schelfhout
eDTServer <- function(
    id,
    data,
    options = list(
        dom = 'Bfrtlip',
        keys = TRUE,
        ordering = FALSE,
        autoFill = list(update = FALSE, focus = 'focus'),
        buttons = list("add", "undo", "redo", "save")
    ),
    class = "display",
    callback = NULL,
    rownames = FALSE,
    colnames = NULL,
    container,
    caption = NULL,
    filter = c("none", "bottom", "top"),
    escape = TRUE,
    style = "auto",
    width = NULL,
    height = NULL,
    elementId = NULL,
    fillContainer = getOption("DT.fillContainer", NULL),
    autoHideNavigation = getOption("DT.autoHideNavigation", NULL),
    selection = "none",
    extensions = c('KeyTable', 'AutoFill', "Buttons"),
    plugins = NULL,
    editable = list(target = "cell"),
    keys = NULL,
    in_place = FALSE,
    format = function(x){x},
    foreignTbls = list(),
    columnOrder = c(),
    statusColor = c("insert"="#e6e6e6", "update"="#32a6d3", "delete"="#e52323"),
    inputUI = editbl::inputUI,
    defaults = tibble(),
    cloneDefaults = tibble(),
    env = environment(),
    canEditRow = TRUE,
    canDeleteRow = TRUE,
    canCloneRow = TRUE,
    utilityColumns = NULL
  ) {
  missingContainer <- missing(container)
  moduleServer(
      id,
      function(input, output, session) {
        ns <- session$ns
        
        rv <- reactiveValues(
            changelog = list(),
            changeLogTracker = 0,
            fullTableRefresh = 0,
            edits_react = 0, # to force refreshing even when reactive value stays the same
            changelog_react = 0 # # to force refreshing even when reactive value stays the same
        )
        
        # Utility columns
        defaultUtilityColumns <- c(
            status = '_editbl_status',
            buttons = '_editbl_buttons',
            identity = '_editbl_identity',
            deleted = '_editbl_deleted'
        )
        utilityColumns <- overwriteDefaults(defaultUtilityColumns, utilityColumns)
        statusCol <- utilityColumns[['status']]
        buttonCol <- utilityColumns[['buttons']]
        identityCol <- utilityColumns[['identity']]
        deleteCol <- utilityColumns[['deleted']]
        
        # Make arguments reactive / set defaults
        # This way users can pass on both reactive an non reactive arguments
        # Need to be explicit about environement. Otherwhise they overwrite themselves.
        argEnv <- parent.frame(3)
        
        if(!shiny::is.reactive(data)){
          data <- shiny::reactive(data, env = argEnv)
        }
        
        if(!shiny::is.reactive(options)){
          options <- shiny::reactive(options, env = argEnv)
        }
        
        if(!shiny::is.reactive(class)){
          class <- shiny::reactive(class, env = argEnv)
        }
        
        if(!shiny::is.reactive(callback)){
          callback <- shiny::reactive(callback, env = argEnv)
        }
        
        if(!shiny::is.reactive(rownames)){
          rownames <- shiny::reactive(rownames, env = argEnv)
        }
        
        if(!shiny::is.reactive(colnames)){
          colnames <- shiny::reactive(colnames, env = argEnv)
        }
        
        if(!missingContainer && !shiny::is.reactive(container)){
          container <- shiny::reactive(container, env = argEnv)
        }
        
        if(!shiny::is.reactive(caption)){
          caption <- shiny::reactive(caption, env = argEnv)
        }
        
        if(!shiny::is.reactive(filter)){
          filter <- shiny::reactive(filter, env = argEnv)
        }
        
        if(!shiny::is.reactive(escape)){
          escape <- shiny::reactive(escape, env = argEnv)
        }
        
        if(!shiny::is.reactive(style)){
          style <- shiny::reactive(style, env = argEnv)
        }
        
        if(!shiny::is.reactive(width)){
          width <- shiny::reactive(width, env = argEnv)
        }
        
        if(!shiny::is.reactive(height)){
          height <- shiny::reactive(height, env = argEnv)
        }
        
        if(!shiny::is.reactive(elementId)){
          elementId <- shiny::reactive(elementId, env = argEnv)
        }
        
        if(!shiny::is.reactive(fillContainer)){
          fillContainer <- shiny::reactive(fillContainer, env = argEnv)
        }
        
        if(!shiny::is.reactive(autoHideNavigation)){
          autoHideNavigation <- shiny::reactive(autoHideNavigation, env = argEnv)
        }
        
        if(!shiny::is.reactive(selection)){
          selection <- shiny::reactive(selection, env = argEnv)
        }
        
        if(!shiny::is.reactive(extensions)){
          extensions <- shiny::reactive(extensions, env = argEnv)
        }
        
        if(!shiny::is.reactive(plugins)){
          plugins <- shiny::reactive(plugins, env = argEnv)
        }
        
        if(!shiny::is.reactive(editable)){
          editable <- shiny::reactive(editable, env = argEnv)
        }
        
        if(is.null(keys)){
          keys <- reactive({
                as.character(dplyr::tbl_vars(data()))
              })
        } else if (!shiny::is.reactive(keys)){
          keys <- shiny::reactive(keys, env = argEnv)
        }
        
        if(!shiny::is.reactive(in_place)){
          in_place <- shiny::reactive(in_place, env = argEnv)
        }
        
        if(!shiny::is.reactive(format)){
          format <- shiny::reactive(format, env = argEnv)
        }
        
        if(!shiny::is.reactive(foreignTbls)){
          foreignTbls <- shiny::reactive(foreignTbls, env = argEnv)
        }
        
        if(!shiny::is.reactive(columnOrder)){
          columnOrder <- shiny::reactive(columnOrder, env = argEnv)
        }
        
        if(!shiny::is.reactive(statusColor)){
          statusColor <- shiny::reactive(statusColor, env = argEnv)
        }
        
        if(!shiny::is.reactive(inputUI)){
          inputUI <- shiny::reactive(inputUI, env = argEnv)
        }
        
        if(!shiny::is.reactive(defaults)){  
          defaults <- shiny::reactive({
                eval(substitute(defaults, env))
              },
              env = argEnv)
        }
        
        if(!shiny::is.reactive(cloneDefaults)){
          cloneDefaults <- shiny::reactive(cloneDefaults, env = argEnv)
        }
        
        if(!shiny::is.reactive(canEditRow)){
          canEditRow <- shiny::reactive(canEditRow, env = argEnv)
        }
        
        if(!shiny::is.reactive(canDeleteRow)){
          canDeleteRow <- shiny::reactive(canDeleteRow, env = argEnv)
        }
        
        if(!shiny::is.reactive(canCloneRow)){
          canCloneRow <- shiny::reactive(canCloneRow, env = argEnv)
        }
        
        # Force re-evaluting reactive after adding a new row for values like Sys.time(), uuid::UUIDgenerate()
        defaultsAddBound <- defaults %>% shiny::bindEvent(input$add)
        # Force re-evaluting reactive after cloning a row for values like Sys.time(), uuid::UUIDgenerate()
        defaultsCloneBound <- cloneDefaults %>% shiny::bindEvent(input$clone)
        
        # Some arguments can have various formats.
        # Standardize first to the most expressive format to make
        # it easier to work with in downstream code.
        colnames_std <- reactive({
              standardizeArgument_colnames(colnames(),  data())
            }
        )
        
        editable_std <- reactive({
              standardizeArgument_editable(editable(), data())
            })
        
        # Columns that are used as additional information but 
        # are not part of the key, nor the original table.
        deductedColnames <- reactive({
              getNonNaturalKeyCols(foreignTbls())
            })
        
        # When source data changes, reset module
        # rv$committedData equals all changes
        # rv$checkPointData equals the committed data with additional utility columns
        # rv$modifiedData keeps track of the current modified/displayed status.
        observe(priority = 2, label = "Reset module with new data",{
              rv$fullTableRefresh
              
              data <- data()
              
              if(!dplyr::is.tbl(data)){
                warning("Data is not of class `tbl`. Converting automatically.")
                data <- castToTbl(data)
              }
              
              rv$committedData <- data
              
              for(foreignTbl in foreignTbls()){
                data <- joinForeignTbl(data,foreignTbl)
              }
              
              # Order the columns of the full dataframe
              if (length(columnOrder()) > 0) {
                if (length( setdiff(columnOrder(), names(data)) ) > 0) {
                  warning(sprintf("The columnOrder variable contains columns that are not present in the data. Ignoring the following redundant columns: %s.",
                      paste(setdiff(columnOrder(), names(data)), collapse = ", ")))
                  colOrder <- columnOrder()[columnOrder() %in% names(data)]
                  data <- data %>% dplyr::select(colOrder)
                } else if (length( setdiff(names(data), columnOrder()) ) > 0) {
                  warning(sprintf("Not all columns are included in the columnOrder. Adding the following missing columns to the right of the table: %s.",
                      paste(setdiff(names(data), columnOrder()), collapse = ", ")))
                  colOrder <- c(columnOrder(), setdiff(names(data), columnOrder()))
                  data <- data %>% dplyr::select(colOrder)
                } else {
                  data <- data %>% dplyr::select(columnOrder())
                }
              }
              
              if(any(utilityColumns %in% dplyr::tbl_vars(data))){
                stop(sprintf("Adjust the utility columns such that they don't clash with the data column names.",
                        paste(utilityColumns, collapse = ", ")))
              }
              
              data <- dplyr::collect(data)
              data <- as.data.frame(data)
              
              data <- initData(
                  data,
                  ns = ns,
                  canEditRow = isolate(canEditRow()),
                  canDeleteRow = isolate(canDeleteRow()),
                  canCloneRow = isolate(canCloneRow()),
                  statusCol = statusCol,
                  buttonCol = buttonCol,
                  iCol = identityCol,
                  deleteCol = deleteCol)
              rv$checkPointData <- data
              rv$modifiedData <- data
              rv$changelog <- list()
              
              DT::selectRows(proxyDT, NA)
              freezeReactiveValue(input, "DT_rows_selected")
            })
        
        # Update server side and client side data
        # rv$newState gets assigned by various actions in the app.
        observe(label = "Replace front-end data",{
              rv$triggerNewState
              req(!is.null(rv$newState) && isTruthy(rv$newState))
              castCols <- base::colnames(isolate(data()))
              data <- rv$newState
              data <- relocate(data,  dplyr::all_of(buttonCol))     
              rv$modifiedData <- data
              
              data <- castForDisplay(data, cols = castCols)
              DT::replaceData(
                  proxy = proxyDT,
                  data = data,
                  resetPaging = FALSE,
                  rownames = rownames()
              )
            })
        
        output$DT <- DT::renderDT({           
              data()
              rv$fullTableRefresh
              
              # Reactive arguments that need slight modifications to work
              # with extra utitily columns
              options <- options()
              colnames <- colnames_std()
              rownames <- rownames()
              escape <- escape()
              editable <- editable_std()
              
              data <- isolate(rv$modifiedData)
              
              baseCols <- setdiff(base::colnames(data), utilityColumns)
              baseColsI <- which(base::colnames(data) %in% baseCols)
              buttonColI <- which(base::colnames(data) == buttonCol) -1 + rownames
              deductedCols <- which(base::colnames(data) %in% deductedColnames()) -1 + rownames
              
              data <- castForDisplay(data)
              
              bcol <- buttonCol
              names(bcol) <- ""
              colnames <- c(bcol,colnames)
              
              # Hide completely hidden utility columns
              options$columnDefs <- c(
                  options$columnDefs, # Not sure about datatable internals, but this gives priority to user specified columnDefs.
                  list(list(
                          visible = FALSE,
                          targets = which(base::colnames(data) %in% 
                                  c(statusCol, identityCol, deleteCol)) - !rownames)
                  ))
              
              if(escape == TRUE){
                escape <- -buttonColI
              }
              else if(escape == FALSE){
                escape <- FALSE
              }
              else if(is.numeric(escape)){
                escape <- c(escape, -buttonColI)
              } else if(is.character(escape)){
                escape <- which(base::colnames(data) %in% escape) + rownames
                escape <- c(escape, -buttonColI)
              }
              
              # Make sure utility columns are not editable
              options$autoFill$columns <- c(options$autoFill$columns, baseColsI - 1 + rownames)
              
              if(!inherits(editable, "logical")){
                if(!"disable" %in% names(editable)){
                  editable <- c(editable, list("disable" = list("columns" = c(buttonColI, deductedCols))))
                } else {
                  editable$disable <- list("columns" = unique(c(editable$disable$columns,
                              buttonColI,
                              deductedCols)))
                }
              }
              
              # For backwards compatibility
              # Maybe remove at some point? E.g. just require to be explicit about options?
              if(is.null(options$dom)){
                options$dom <- "Bfrtip"
              }
              if(is.null(options$buttons)){
                options$buttons <- list("add", "undo", "redo", "save")
              }
              
              options$buttons <- lapply(options$buttons, function(x){
                    if(is.character(x) && x %in% c("add", "undo", "redo", "save")){
                      icon = switch(x,
                          "add" = icon("plus"),
                          "undo" = icon("rotate-left"),
                          "redo" = icon("rotate-right"),
                          "save" = icon("floppy-disk"),
                          ""
                      )   
                      disabled = switch(x,
                          "add" = FALSE,
                          "undo" = TRUE,
                          "redo" =  TRUE,
                          "save" = TRUE,
                          TRUE
                      )       
                      
                      customButton(ns(x), label = x, icon, disabled = disabled)
                    } else {
                      x
                    }
                  })
              
              # Deal with the fact that 'container' can be a missing argument
              # Which is why put arguments in a list and use do.call instead of passing on directly.
              # FIXME: there should be a better approach              
              internalArgs <- list(
                  data = data,
                  options = options,
                  class = class(),
                  callback = DT::JS(c(keyTableJS,autoFillJs, callback())),
                  rownames = rownames,
                  colnames = colnames,
                  caption = caption(),
                  filter = filter(),
                  escape = escape,
                  style = style(),
                  width = width(),
                  height = height(),
                  elementId = elementId(),
                  fillContainer = fillContainer(),
                  autoHideNavigation = autoHideNavigation(),
                  selection = selection(),
                  extensions = extensions(),
                  plugins = plugins(),
                  editable = editable
              )
              if(!missingContainer){
                internalArgs <- c(internalArgs, list(container = container()))
              }
              
              do.call(DT::datatable, internalArgs) %>%
                  formatStyle(statusCol, target='row',
                      backgroundColor = styleEqual('inserted',statusColor()["insert"]))%>%
                  formatStyle(statusCol, target='row',
                      backgroundColor = styleEqual('edited',statusColor()["update"]))%>%
                  formatStyle(deleteCol, target='row',
                      backgroundColor = styleEqual(TRUE,statusColor()["delete"])) %>%
                  format()()
            })
        
        proxyDT <- DT::dataTableProxy("DT")
        
        notInModalColumns <- reactive({              
              # Columns that should not be edited through the modal
              deducted <- deductedColnames() # Therefore non-editable
              invisible <- unlist(lapply(options()$columnDefs, function(x){
                        if(!is.null(x$visible)){
                          if(!x$visible){
                            x$targets
                          }
                        }
                      }))
              notEditable <- names(rv$modifiedData)[editable_std()$disable$columns + 1]
              unique(c(
                      utilityColumns,
                      deducted,
                      invisible,
                      notEditable
                  ))
            })
        
        # Use different id each time to prevent conflicts / flashing re-rendering due to renderUI
        editModalId <- reactive({
              input$edit
              gsub("-", "_", uuid::UUIDgenerate())
            })
        
        observeEvent(input$edit, {
              rv$modalData <- inputServer(
                  editModalId(),
                  data = rv$modifiedData[clickedRow(),],
                  notEditable = notInModalColumns,
                  colnames = colnames_std,
                  foreignTbls = foreignTbls)
            })
        
        observeEvent(input$edit, {
              showModal(
                  modalDialog(
                      inputUI()(id = ns(editModalId()), data = rv$modifiedData[clickedRow(),]),
                      footer = tagList(
                          actionButton(ns("confirmEdit"), "Ok"),
                          modalButton("cancel")
                      ),
                      easyClose = TRUE
                  )
              )
            })
        
        clickedRow <- reactive({
              identity = sub("^.*_","",input$current_id)
              which(rv$modifiedData[,identityCol] == identity)
            })
        
        effectiveChanges <- reactive({
              data <- do.call(rbind, rv$changelog[seq_len(rv$changeLogTracker)])
              
              # get last state of the row per id
              data <- do.call(rbind,(lapply(unique(data[[identityCol]]), 
                            function(i){tail(data[data[[identityCol]] == i,],1)})))
              data
            })
        
        observe({
              req(!is.null(rv$changelog) && isTruthy(rv$changelog))
              rv$changelog_react
              
              rv$changeLogTracker <- length(rv$changelog)
            })
        
        observeEvent(input$undo,{
              i <- rv$changeLogTracker
              req(i > 0)
              data <- rv$modifiedData
              undoChanges <- rv$changelog[[rv$changeLogTracker]]
              
              for(row in seq_len(nrow(undoChanges))){
                undoChange <- undoChanges[row,]
                
                lastLogState <- do.call(rbind,rv$changelog[seq_len(max(0,i-1))])
                lastLogState <- tail(
                    lastLogState[lastLogState[[identityCol]] == undoChange[[identityCol]],],1
                )
                
                lastCheckPointState <- rv$checkPointData[
                    rv$checkPointData[[identityCol]] == undoChange[[identityCol]],]
                
                if(!is.null(lastLogState) && nrow(lastLogState)){
                  stateBeforeChange <- lastLogState
                } else {
                  stateBeforeChange <- lastCheckPointState
                }
                
                if(undoChange[[statusCol]] == "inserted" && nrow(stateBeforeChange) == 0){ # delete if row did not exist before
                  data <- data[data[[identityCol]] != undoChange[[identityCol]],]
                } else if (undoChange[[statusCol]] == "deleted" && !undoChange[[identityCol]] %in% data[[identityCol]]){ # re-insert if row is now deleted
                  data <- cbind(undoChange, data)
                } else { # set row to previous state
                  data[data[[identityCol]] == undoChange[[identityCol]],] <- stateBeforeChange
                }
              }
              
              rv$changeLogTracker <- max(0, rv$changeLogTracker - 1)
              rv$newState <- data
            })
        
        observeEvent(input$redo,{
              i <- rv$changeLogTracker
              nChanges <- length(rv$changelog)
              req(i < nChanges)
              
              data <- rv$modifiedData
              
              redoChanges <- rv$changelog[[i + 1]]
              
              for (iRedoChange in seq_len(nrow(redoChanges))){
                redoChange <- redoChanges[iRedoChange,]
                if(redoChange[[statusCol]] == "inserted" && !redoChange[[identityCol]] %in% data[[identityCol]]){
                  data <- rbind(redoChange,data)
                } else {
                  data[data[[identityCol]] == redoChange[[identityCol]],] <- redoChange
                }
              }
              
              rv$changeLogTracker <- i + 1
              rv$newState <- data
              
            })
        
        observeEvent(input$confirmEdit, {
              i <- clickedRow()
              data <- rv$modifiedData
              req(evalCanEditRow(
                      row=data[i,],
                      canEditRow=canEditRow(),
                      statusCol=statusCol))
              data[i,] <-  fillDeductedColumns(rv$modalData(), foreignTbls())
              
              currentStatus <- data[i,statusCol]
              if(currentStatus == "unmodified"){
                data[i,statusCol] <- "edited"
              }
              
              newChange <- data[i,]
              changelog <- rv$changelog[seq_len(rv$changeLogTracker)]
              changelog[[rv$changeLogTracker +1]] <- newChange
              rv$changelog <- changelog
              
              rv$newState <- data
              shiny::removeModal()
            })
        
        observeEvent(input$DT_cells_filled, {
              req(!is.null(input$DT_cells_filled) && isTruthy(input$DT_cells_filled))
              edits <- input$DT_cells_filled
              edits$row <- input$DT_rows_current[edits$row]           
              rv$edits <- edits
              rv$edits_react <-  rv$edits_react + 1
            })
        
        observeEvent(input$DT_cell_edit, {
              rv$edits <- input$DT_cell_edit
              rv$edits_react <- rv$edits_react + 1
            })
        
        observeEvent(rv$edits_react, {
              req(!is.null(rv$edits) && isTruthy(rv$edits))
              edits <- unique(rv$edits)
              rv$edits <- NULL
              
              data <- rv$modifiedData
              tryCatch({                    
                    newRows <- list()
                    
                    for(i in sort(unique(edits$row))){
                      changes <- edits[edits$row == i,]
                      currentRow <- data[i,]
                      if(!evalCanEditRow(
                          row = currentRow, 
                          canEditRow=canEditRow(), 
                          statusCol=statusCol
                          )
                      ){
                        next
                      }
                      
                      newRow <- currentRow
                      
                      hasChanged <- FALSE
                      for(iChange in seq_len(nrow(changes))){
                        change <- changes[iChange,]
                        j <- change$col + 1 + rownames()
                        currentValue <- currentRow[,j]
                        newValue <- coerceValue(
                            change$value,
                            data[i, j])
                        
                        if(!identical(currentValue, newValue)){
                          hasChanged <- TRUE
                          newRow[,j] <- newValue
                          currentStatus <- newRow[[statusCol]]
                          if(currentStatus == "unmodified"){
                            newRow[,statusCol] <- "edited"
                          }
                        }
                      }
                      
                      if(!hasChanged){
                        next()
                      }
                      
                      newRow <- fillDeductedColumns(newRow, foreignTbls())
                      newRows[[as.character(i)]] <- newRow
                      data[i,] <- newRow
                      
                    }    
                    newChanges <- do.call(rbind, newRows)
                    
                    changelog <- rv$changelog[seq_len(rv$changeLogTracker)]
                    changelog[[rv$changeLogTracker +1]] <- newChanges
                    rv$changelog <- changelog
                    rv$changelog_react <- rv$changelog_react + 1
                    
                    rv$newState <- data
                    rv$triggerNewState <- Sys.time() # Need to reupdate if edits were not allowed
                    
                  },
                  error = function(e){
                    rv$resultMessage <- "The change you just made is not allowed. Reverting."
                    rv$showResultMessage <- Sys.time()
                    rv$newState <- data
                    rv$triggerNewState <- Sys.time()
                  }
              )
            })
        
        observeEvent(input$delete,{
              rowNumber <- clickedRow()
              data <- rv$modifiedData
              row <- data[rowNumber,]
              req(evalCanDeleteRow(
                      row = row,
                      canDeleteRow=canDeleteRow(),
                      statusCol=statusCol
                  )
              )
              row[[deleteCol]] <- !row[[deleteCol]]
              
              newChange <- row
              changelog <- rv$changelog[seq_len(rv$changeLogTracker)]
              changelog[[rv$changeLogTracker +1]] <- newChange
              rv$changelog <- changelog
              
              data[rowNumber,] <- row
              rv$newState <- data
            })
          
          observeEvent(input$clone,{
              rowNumber <- clickedRow()
              data <- rv$modifiedData
              
              req(evalCanCloneRow(
                  row = row,
                  canCloneRow=canCloneRow(),
                  statusCol=statusCol
                )
              )
              
              newRow <- data[rowNumber,]
              cloneDefaults <- defaultsCloneBound()
              for(col in base::colnames(cloneDefaults)){
                currentClass <- base::class(data[[col]])
                defaultClass <- base::class(cloneDefaults[[col]])
                
                if(!col %in% base::colnames(newRow)){
                  stop(sprintf("Column %s not available. Not adding default.", col))
                } else if (! identical(currentClass, defaultClass)){
                  stop(sprintf("Default set for %s is of type %s instead of %s", col, defaultClass, currentClass))
                } else {
                  newRow[[col]] <- cloneDefaults[[col]]
                }
              }
              newRow[,statusCol] <- "inserted"
              newRow[,deleteCol] <- FALSE
              newRow[,identityCol] <- uuid::UUIDgenerate()
              newRow <- addButtons(
                df = newRow,
                columnName = buttonCol,
                iCol = identityCol,
                ns = ns,
                canDeleteRow = canDeleteRow(),
                canEditRow = canEditRow(),
                canCloneRow = canCloneRow(),
                statusCol = statusCol
              )
              
              # save to changelog
              newChange <- newRow
              changelog <- rv$changelog[seq_len(rv$changeLogTracker)]
              changelog[[rv$changeLogTracker +1]] <- newChange
              rv$changelog <- changelog
              
              data <- rbind(
                data[0:(rowNumber-1),],
                newRow,
                data[rowNumber:nrow(data),]
              )
              
              rv$newState <- data
            })
        
        observeEvent(input$add,{
              data <- rv$modifiedData
              # create new row
              newRow <- data %>%
                  dplyr::filter(FALSE)
              newRow <- newRow[1,]
              newRow <- fixInteger64(newRow) # https://github.com/Rdatatable/data.table/issues/4561
              
              defaults <- defaultsAddBound()
              for(col in base::colnames(defaults)){
                currentClass <- base::class(data[[col]])
                defaultClass <- base::class(defaults[[col]])
                
                if(!col %in% base::colnames(newRow)){
                  stop(sprintf("Column %s not available. Not adding default.", col))
                } else if (! identical(currentClass, defaultClass)){
                  stop(sprintf("Default set for %s is of type %s instead of %s", col, defaultClass, currentClass))
                } else {
                  newRow[[col]] <- defaults[[col]]
                }
              }
              newRow[,statusCol] <- "inserted"
              newRow[,deleteCol] <- FALSE
              newRow[,identityCol] <- uuid::UUIDgenerate()
              newRow <- addButtons(
                  df = newRow,
                  columnName = buttonCol,
                  iCol = identityCol,
                  ns = ns,
                  canDeleteRow = canDeleteRow(),
                  canEditRow = canEditRow(),
                  canCloneRow = canCloneRow(),
                  statusCol = statusCol)
              
              # save to changelog
              newChange <- newRow
              changelog <- rv$changelog[seq_len(rv$changeLogTracker)]
              changelog[[rv$changeLogTracker +1]] <- newChange
              rv$changelog <- changelog
              
              data <- rbind(
                  newRow,
                  data
              )
              
              rv$newState <- data
            })
        
        effectiveInserted <- reactive({
              modified <- effectiveChanges()         
              modified[
                  modified[[statusCol]] == "inserted" &
                      modified[[deleteCol]] == FALSE,]
            })
        
        effectiveEdited <- reactive({
              modified <- effectiveChanges()
              modified[
                  modified[[statusCol]] == "edited" &
                      modified[[deleteCol]] == FALSE,]
            })
        
        effectiveDeleted <- reactive({
              modified <- effectiveChanges()
              modified[
                  modified[[statusCol]] != "inserted" &
                      modified[[deleteCol]] == TRUE,]
            })
        
        observeEvent(input$save,{
              shiny::showModal(
                  modalDialog(
                      title = "Do you really want to make the changes?",
                      sprintf("inserted: %s
                              edited: %s
                              deleted: %s
                              ",
                          nrow(effectiveInserted()),
                          nrow(effectiveEdited()),
                          nrow(effectiveDeleted())
                      ),
                      footer = tagList(
                          actionButton(ns("confirmCommit"), label = "Ok"),
                          modalButton("Cancel")
                      ),
                      easyClose = TRUE
                  )
              
              )
            })
        
        observe({
              if(is.null(effectiveChanges())){
                shinyjs::disable("save")
              } else {
                shinyjs::enable("save")             
              }
            })
        
        observe({
              if(rv$changeLogTracker == 0){
                shinyjs::disable("undo")
              } else {
                shinyjs::enable("undo")             
              }
            })
        
        observe({
              if(length(rv$changelog) <= rv$changeLogTracker){
                shinyjs::disable("redo")
              } else {
                shinyjs::enable("redo")             
              }
            })
        
        observeEvent(input$confirmCommit, {     
              req(!is.null(effectiveChanges()) && isTruthy(effectiveChanges()))
              modified <- effectiveChanges()
              cols <- as.character(dplyr::tbl_vars(data()))
              checkPoint <- rv$checkPointData
              tryCatch({
                    # Keep order. Delete > update > insert.
                    result <- rv$committedData
                    beginTransaction(result)
                    
                    # deletes
                    deleted <- merge(
                        modified[modified[[deleteCol]] == TRUE,identityCol,drop = FALSE],
                        checkPoint,
                        by = identityCol)[,keys(),drop = FALSE]
                    
                    # edits
                    edited <- effectiveEdited()
                    if(!checkForeignTbls(edited, foreignTbls())){
                      stop("You made invalid edits to a row.")
                    }
                    edited <- edited[, c(cols, identityCol)]
                    match_x <-  merge(
                        edited[,identityCol, drop = FALSE],
                        checkPoint,
                        by = identityCol)
                    match_x <- match_x[order(match_x[[identityCol]]), keys(), drop = FALSE]
                    match_y <- edited[order(edited[[identityCol]]), keys(), drop = FALSE]
                    match <- list(
                        x = match_x,
                        y = match_y
                    )
                    edited <- edited[cols]
                    
                    # inserts
                    inserted <- effectiveInserted()
                    if(!checkForeignTbls(inserted, foreignTbls())){
                      stop("You made invalid inserts.")
                    }
                    inserted <- inserted[,cols,drop=FALSE]
                    
                    if(nrow(deleted)){
                      if(inherits(result, 'tbl_dbi') & in_place()){
                        # dbplyr:::rows_delete.tbl_dbi requires y to be in the same source and will start a transaction for this if not the case.
                        # Most backends do not support nested transactions. Therefore manually copy the data first.
                        # See also: https://github.com/tidyverse/dbplyr/issues/1298
                        temp_name = paste0("editbl_", gsub("-", "", UUIDgenerate()))
                        deleted <- dplyr::copy_to(
                            dest = dbplyr::remote_src(result),
                            df = deleted,
                            name = temp_name,
                            temporary = TRUE,
                            in_transaction = FALSE,
                            types = dbplyr::db_col_types(
                                dbplyr::remote_con(result),
                                dbplyr::remote_table(result))[base::colnames(deleted)])
                      }
                      
                      result <- e_rows_delete(
                          x = result,
                          y = deleted,
                          by = keys(),
                          in_place = in_place(),
                          unmatched = 'ignore')
                      
                      if(inherits(result, 'tbl_dbi')){
                        DBI::dbRemoveTable(dbplyr::remote_con(result), temp_name)
                      }
                    }
					
                    if(nrow(edited)){
                      result <- e_rows_update(
                          x = result,
                          y = edited,
                          match = match,
                          by = keys(),
                          in_place = in_place())				  
                    }					
                    
                    if(nrow(inserted)){
# Needed code should there be a switch to dbplyr::rows_insert
#                      if(inherits(result, 'tbl_dbi')){
#                        # https://github.com/openanalytics/editbl/issues/1
#                        # dbplyr:::rows_insert.tbl_dbi requires y to be in the same source and will start a transaction for this if not the case.
#                        # Most backends do not support nested transactions. Therefore manually copy the data first.
#                        temp_name = paste0("editbl_", gsub("-", "", UUIDgenerate()))
#                        inserted <- dplyr::copy_to(
#                            dest = result$src,
#                            df = inserted,
#                            name = temp_name,
#                            temporary = TRUE,
#                            in_transaction = FALSE)
#                      }
                      
                      result <- e_rows_insert(
                          x = result,
                          y = inserted,
                          by = keys(),
                          in_place = in_place(),
                          conflict = "ignore")
                      
# Needed code should there be a switch to dbplyr::rows_insert                     
#                      if(inherits(result, 'tbl_dbi')){
#                        DBI::dbRemoveTable(result$src$con, temp_name)
#                      }       
                    }
                    
					# Data to return to user
                    rv$committedData <- result
					rv$inserted <- inserted
					rv$edited <- edited
					rv$deleted <- deleted
					
					
                    # Set modified and rendered to comitted version
                    # re-read data in case of in_place modification
                    # This is because certain backends might modify the row further with defaults etc.
                    if(in_place()){
                      rv$fullTableRefresh <- UUIDgenerate()
                    } else {
                      checkPointState <- rv$modifiedData
                      
                      # Remove deleted rows
                      checkPointState <- checkPointState[checkPointState[[deleteCol]] != TRUE,]
                      
                      # Update inserted rows to being 'unmodified'
                      insertedRows <- which(checkPointState[,statusCol] == 'inserted')
                      for(i in insertedRows){
                        currentRow <- checkPointState[i,]
                        adjustedRow <- currentRow
                        adjustedRow[,statusCol] <- 'unmodified'
                        adjustedRow <- addButtons(
                            df = adjustedRow,
                            columnName = buttonCol,
                            iCol = identityCol,
                            ns = ns,
                            canEditRow = canEditRow(),
                            canDeleteRow = canDeleteRow(),
                            canCloneRow = canCloneRow(),
                            statusCol = statusCol
                            )
                        checkPointState[i,] <- adjustedRow
                      }
                      
                      # Update edited rows to being 'unmodified'
                      checkPointState[[statusCol]] <- 'unmodified'

                      rv$changelog <- list()
                      rv$checkPointData <- checkPointState
                      rv$newState <- checkPointState
                    }
                    
                    commitTransaction(result)
                    rv$resultMessage <- "success."
                  }, error = function(cond){                    
                    rollbackTransaction(result)
                    rv$resultMessage <- sprintf(
                        "Failure: %s",
                        conditionMessage(cond)
                    )
                  })
              rv$showResultMessage <- Sys.time()
              shiny::removeModal()
            })
        
        observeEvent(rv$showResultMessage,{
              shiny::showNotification(
                  rv$resultMessage
              )   
            })
        
        result <- reactive({
              inputData <- isolate(data())
              result <- rv$committedData
              tryCatch({
                    result <- castFromTbl(tbl = result, template = inputData)
                  }, error = function(e){
                    warning(sprintf("%s. Returning 'tbl'.", e$message))
                  })
              result
            })
        
        # Ensure selection holds while deleting / adding rows
        observe(priority = 1,{
              req(!is.null(rv$modifiedData) && isTruthy(rv$modifiedData))
              req(!is.null(isolate(rv$selected)) && isTruthy(isolate(rv$selected)))
              
              currentSelection <- isolate(rv$selected)[[identityCol]]  
              newIndexes <- which(rv$modifiedData[[identityCol]] %in% currentSelection)
              if(length(newIndexes)){
                DT::selectRows(proxyDT, newIndexes)
              }
            })
        
        observe({
              rv$selected <- selected() # To force evaluation
            })
        
        selected <- reactive({
              rows <-input$DT_rows_selected
              data <- isolate(rv$modifiedData)[rows,]
              data
            })
        
        dataVars <- reactive({
              dplyr::tbl_vars(data())
            })
        
        return(list(
                result = result,
                state = reactive({castToTemplate(rv$modifiedData[!rv$modifiedData[[deleteCol]],dataVars()], data())}),
                selected = reactive({castToTemplate(selected()[,dataVars()], data())}),
				inserted = reactive({castToTemplate(rv$inserted[,dataVars()], data())}),
				edited = reactive({castToTemplate(rv$edited[,dataVars()], data())}),
				deleted = reactive({castToTemplate(rv$deleted[,dataVars()], data())})
            ))
      }
  )
}

#' Determine if a row can be deleted
#' 
#' @details calling this around the user passed on function ensures
#'  that newly inserted rows are being excempt from the logic.
#'  Moreover, the output of the function can be checked.
#' 
#' @param row `tibble`, single row
#' @param canDeleteRow `function` with argument 'row' defining logic on wether or
#'    not the row can be modified. Can also be `logical` TRUE or FALSE.
#' @param statusCol `character(1)` name of column with general status (e.g. modified or not).
#' @return `boolean`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
evalCanDeleteRow <- function(
    row,
    canDeleteRow = TRUE,
    statusCol='_editbl_status'
  ){
  # Prevent evaluating logic and speed up for most common use-case
  if(is.logical(canDeleteRow) && canDeleteRow){
    return(TRUE)
  }
  
  if (!is.null(statusCol) && row[[statusCol]] == 'inserted'){
    return(TRUE)
  }
  
  if(is.function(canDeleteRow)){
    result <- canDeleteRow(row=row)
    if(!is.logical(result)){
      stop('canDeleteRow should return a logical value.')
    }
  } else if (is.logical(canDeleteRow)) {
    result <- canDeleteRow
  }
  
  return(result)
}

#' Determine if a row can be edited
#' 
#' @details calling this around the user passed on function ensures
#'  that newly inserted rows are being excempt from the logic.
#'  Moreover, the output of the function can be checked.
#' 
#' @param row `tibble`, single row.
#' @param canEditRow `function` with argument 'row' defining logic on wether or
#'    not the row can be modified. Can also be `logical` TRUE or FALSE.
#' @param statusCol `character(1)` name of column with general status (e.g. modified or not).
#' @return `boolean`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
evalCanEditRow <- function(row, canEditRow = TRUE, statusCol='_editbl_status'){
  
  # Prevent evaluating logic and speed up for most common use-case
  if(is.logical(canEditRow) && canEditRow){
    return(TRUE)
  }
  
  if (!is.null(statusCol) && row[[statusCol]] == 'inserted'){
    return(TRUE)
  }
  
  if(is.function(canEditRow)){
    result <- canEditRow(row=row)
    if(!is.logical(result)){
      stop('canEditRow should return a logical value.')
    }
  } else if (is.logical(canEditRow)) {
    result <- canEditRow
  }
  
  return(result)
}


#' Determine if a row can be cloned
#' 
#' @details calling this around the user passed on function ensures
#'  that newly inserted rows are being excempt from the logic.
#'  Moreover, the output of the function can be checked.
#' 
#' @param row `tibble`, single row.
#' @param canCloneRow `function` with argument 'row' defining logic on wether or
#'    not the row can be cloned. Can also be `logical` TRUE or FALSE.
#' @param statusCol `character(1)` name of column with general status (e.g. modified or not).
#' @return `boolean`
#' 
#' @author Saar Junius
#' @keywords internal
evalCanCloneRow <- function(row, canCloneRow = TRUE, statusCol='_editbl_status'){
  
  # Prevent evaluating logic and speed up for most common use-case
  if(is.logical(canCloneRow) && canCloneRow){
    return(TRUE)
  }
  
  if (!is.null(statusCol) && row[[statusCol]] == 'inserted'){
    return(TRUE)
  }
  
  if(is.function(canCloneRow)){
    result <- canCloneRow(row=row)
    if(!is.logical(result)){
      stop('canCloneRow should return a logical value.')
    }
  } else if (is.logical(canCloneRow)) {
    result <- canCloneRow
  }
  
  return(result)
}


#' Add some extra columns to data to allow for / keep track of modifications
#' @param data `data.frame`
#' @param ns namespace function
#' @param buttonCol `character(1)` name of column with buttons
#' @param statusCol `character(1)` name of column with general status (e.g. modified or not).
#' @param deleteCol `character(1)` name of the column with deletion status.
#' @param iCol `character(1)` name of column containing a unique identifier.
#' @inheritParams canXXXRowTemplate
#' @return data with extra columns buttons, status, i.
#' @importFrom dplyr relocate all_of
#' @importFrom uuid UUIDgenerate
#' @author Jasper Schelfhout
#' @keywords internal
initData <- function(
    data,
    ns,
    buttonCol = "buttons",
    statusCol = "_editbl_status",
    deleteCol = "_editbl_deleted",
    iCol = "i",
    canDeleteRow = TRUE,
    canEditRow = TRUE,
    canCloneRow = TRUE
){
  data[statusCol] <- rep("unmodified", nrow(data))
  data[deleteCol] <- rep(FALSE, nrow(data))
  
  if(nrow(data) > 0){
    data[iCol] <- unlist(lapply(seq_len(nrow(data)), uuid::UUIDgenerate))
  } else {
    data[iCol] <- character(0)
  }
  data <- addButtons(
      df = data,
      columnName = buttonCol,
      ns = ns,
      iCol = iCol,
      canEditRow = canEditRow,
      canDeleteRow = canDeleteRow,
      canCloneRow = canCloneRow,
      statusCol=NULL) # Not to trigger unneeded logic since all values are 'unmodified'
  data <- relocate(data, all_of(buttonCol))
  data
}

#' Add modification buttons as a column
#' @param df `data.frame`
#' @param columnName `character(1)`
#' @param ns namespace function
#' @param iCol `character(1)` name of column containing a unique identifier.
#' @param statusCol `character(1)` name of column with general status (e.g. modified or not).
#'    if `NULL`, the data is interpreted as 'unmodified'.
#' @inheritParams canXXXRowTemplate
#' @return df with extra column containing buttons
#' @importFrom dplyr across everything rowwise
#' @importFrom rlang :=
#' 
#' @author Jasper Schelfhout
#' @keywords internal
addButtons <- function(
    df,
    columnName,
    ns,
    iCol = 'i',
    canEditRow = TRUE,
    canDeleteRow = TRUE,
    canCloneRow = TRUE,
    statusCol = '_editbl_status'
){
  ns_char = ns("")
  
  if(!nrow(df)){
    df[columnName] <- character(0)
    return(df)
  } else {
    df <- df %>%
        dplyr::rowwise() %>%
        dplyr::mutate(!!columnName := createButtons(
                row = dplyr::across(dplyr::everything(), ~ .),
                suffix = get(iCol),
                ns = !!ns_char,
                canEditRow = !!canEditRow,
                canDeleteRow = !!canDeleteRow,
                canCloneRow = !!canCloneRow,
                statusCol = !!statusCol
            )) %>%
        dplyr::ungroup() %>%
        as.data.frame()
  }
  
  df
}


#' Helper function to write HTML
#' @inheritParams createDeleteButtonHTML
#' @details only to be used interactively. sprintf() implementation
#'   is faster.
#' @importFrom shiny div actionButton icon
#' @seealso createEditButtonHTML
#' @keywords internal
createDeleteButtonHTML_shiny <- function(
    ns = "%1$s",
    suffix = "%2$s",
    disabled = FALSE){
  as.character(
      actionButton(
          inputId = sprintf("%1$sdelete_row_%2$s", ns,suffix),
          label = "",
          icon = icon("trash"),
          style = "color: red;background-color: white",
          onclick =  HTML(sprintf("get_id(this.id, '%1$s');
                      Shiny.setInputValue(\"%1$sdelete\", Math.random(), {priority: \"event\"});",
                  ns
              )
          )
      )
  )
}


#' Helper function to write HTML
#' @inheritParams createEditButtonHTML
#' @details only to be used interactively. sprintf() implementation
#'   is faster.
#' @seealso createEditButtonHTML
#' @importFrom shiny div actionButton icon
#' @keywords internal
createEditButtonHTML_shiny <- function(
    ns = "%1$s",
    suffix = "%2$s",
    disabled = FALSE){
  as.character(
      actionButton(
          inputId = sprintf("%1$sedit_row_%2$s", ns,suffix),
          label = "",
          disabled = disabled,
          icon = icon("pen-to-square"),
          style = "background-color: white",
          onclick = HTML(sprintf("get_id(this.id, '%1$s');
                      Shiny.setInputValue(\"%1$sedit\", Math.random(), {priority: \"event\"});",
                  ns
              ))
      )  
  )
}

#' Helper function to write HTML
#' @inheritParams createCloneButtonHTML
#' @details only to be used interactively. sprintf() implementation
#'   is faster.
#' @seealso createCloneButtonHTML
#' @importFrom shiny div actionButton icon
#' @keywords internal
createCloneButtonHTML_shiny <- function(
  ns = "%1$s",
  suffix = "%2$s",
  disabled = FALSE){
  as.character(
    actionButton(
      inputId = sprintf("%1$sclone_row_%2$s", ns,suffix),
      label = "",
      disabled = disabled,
      icon = icon("clone"),
      style = "background-color: white",
      onclick = HTML(sprintf("get_id(this.id, '%1$s');
            Shiny.setInputValue(\"%1$sclone\", Math.random(), {priority: \"event\"});",
          ns
        ))
    )  
  )
}

#' Generate HTML for an in-row edit button
#' @param suffix `character(1)` id of the row
#' @param ns `character(1)` namespace
#' @param disabled `logical(1)` wether or not the button has to be disabled
#' @return `character(1)` HTML
#' @keywords internal
createEditButtonHTML <- function(
    ns,
    suffix,
    disabled = FALSE
){
  if(disabled){
    disabled_str = 'disabled'
  } else {
    disabled_str = ''
  }
  sprintf(r"(<button id="%1$sedit_row_%2$s" type="button" class="btn btn-default action-button" %3$s style="background-color: white" onclick="get_id(this.id, &#39;%1$s&#39;);&#10;                      Shiny.setInputValue(&quot;%1$sedit&quot;, Math.random(), {priority: &quot;event&quot;});">
      <i class="far fa-pen-to-square" role="presentation" aria-label="pen-to-square icon"></i>   
      </button>)", ns, suffix, disabled_str)
}

#' Generate HTML for an in-row delete button
#' @param suffix `character(1)` id of the row
#' @param ns `character(1)` namespace
#' @param disabled `logical(1)` wether or not the button has to be disabled
#' @return `character(1)` HTML
#' @keywords internal
createDeleteButtonHTML <- function(
    ns = "%1$s",
    suffix = "%2$s",
    disabled=FALSE){
  if(disabled){
    disabled_str = 'disabled'
  } else {
    disabled_str = ''
  }
  sprintf(r"(<button id="%1$sdelete_row_%2$s" type="button" class="btn btn-default action-button"  %3$s style="color: red;background-color: white" onclick="get_id(this.id, &#39;%1$s&#39;);&#10;                      Shiny.setInputValue(&quot;%1$sdelete&quot;, Math.random(), {priority: &quot;event&quot;});">
      <i class="fas fa-trash" role="presentation" aria-label="trash icon"></i>
      </button>)", ns, suffix, disabled_str)
}

#' Generate HTML for an in-row clone button
#' @param suffix `character(1)` id of the row
#' @param ns `character(1)` namespace
#' @param disabled `logical(1)` wether or not the button has to be disabled
#' @return `character(1)` HTML
#' @keywords internal
createCloneButtonHTML <- function(
  ns = "%1$s",
  suffix = "%2$s",
  disabled=FALSE){
  if(disabled){
    disabled_str = 'disabled'
  } else {
    disabled_str = ''
  }
  sprintf(r"(<button id="%1$sclone_row_%2$s" type="button" class="btn btn-default action-button"  %3$s style="background-color: white" onclick="get_id(this.id, &#39;%1$s&#39;);&#10;                      Shiny.setInputValue(&quot;%1$sclone&quot;, Math.random(), {priority: &quot;event&quot;});">
      <i class="fa-solid fa-clone" role="presentation" aria-label="clone icon"></i>
      </button>)", ns, suffix, disabled_str)
}


#' Re-usable documentation
#' @param canEditRow can be either of the following:
#'    - `logical`, e.g. TRUE or FALSE. In case of `FALSE`, buttons will not be visible except for new rows.
#'    - `function`. Needs as input an argument `row` which accepts a single row `tibble` and as output TRUE/FALSE.
#' @param canCloneRow can be either of the following:
#'    - `logical`, e.g. TRUE or FALSE. In case of `FALSE`, buttons will not be visible except for new rows.
#'    - `function`. Needs as input an argument `row` which accepts a single row `tibble` and as output TRUE/FALSE.
#' @param canDeleteRow can be either of the following:
#'    - `logical`, e.g. TRUE or FALSE. In case of `FALSE`, buttons will not be visible except for new rows.
#'    - `function`. Needs as input an argument `row` which accepts a single row `tibble` and as output TRUE/FALSE.
#' @keywords internal
canXXXRowTemplate <- function(canEditRow, canCloneRow, canDeleteRow){
  NULL
}

#' Create buttons to modify the row.
#' @details buttons used per row in the app.
#' @param row `tibble` with single row
#' @param suffix `character(1)`
#' @param ns `character(1)` namespace
#' @param statusCol `character(1)` name of column with general status (e.g. modified or not).
#'    if `NULL`, the data is interpreted as 'unmodified'.
#' @inheritParams canXXXRowTemplate
#' @return `character(1)` HTML
#' @keywords internal
createButtons <- function(
    row,
    suffix,
    ns, 
    canEditRow = TRUE,
    canDeleteRow = TRUE,
    canCloneRow = TRUE,
    statusCol = '_editbl_status'
){
  # Hide delete completely if canDeleteRow = FALSE. With the exception of newly created rows.
  if(is.logical(canDeleteRow) 
	&& !canDeleteRow 
	&& (is.null(statusCol) || row[[statusCol]] != 'inserted')
  ){
    deleteButton <- ""
  } else {
    deleteButton <- createDeleteButtonHTML(
			  ns=ns,
			  suffix=suffix,
			  disabled = !evalCanDeleteRow(row=row, canDeleteRow=canDeleteRow, statusCol=statusCol))
  }
  
  # Hide edit completely if canEditRow = FALSE. With the exception of newly created rows.
  if(is.logical(canEditRow) 
		  && !canEditRow 
		  && (is.null(statusCol) || row[[statusCol]] != 'inserted')
		  ){
	  editButton <- ""
  } else {
	  editButton <- createEditButtonHTML(
			  ns=ns,
			  suffix=suffix,
			  disabled = !evalCanEditRow(row=row, canEditRow=canEditRow, statusCol=statusCol))
  }
  
  # Hide clone button completely if canCloneRow = FALSE. With the exception of newly created rows.
  if(is.logical(canCloneRow) 
		  && !canCloneRow 
		  && (is.null(statusCol) || row[[statusCol]] != 'inserted')
		  ){
	  cloneButton <- ""
  } else {
	  cloneButton <- createCloneButtonHTML(
			  ns = ns,
			  suffix = suffix,
			  disabled = !evalCanCloneRow(row = row, canCloneRow = canCloneRow, statusCol = statusCol))
  }
  
  result <- sprintf('<div class="btn-group">%1$s%2$s%3$s</div>',
      deleteButton,
      editButton,
      cloneButton
  )
  
  result
}

#' Function to generate CSS to disable clicking events on a column
#' @param id `character(1)` namespaced id of the datatable
#' @details <https://stackoverflow.com/questions/60406027/how-to-disable-double-click-reactivity-for-specific-columns-in-r-datatable>
#' @details <https://stackoverflow.com/questions/75406546/apply-css-styling-to-a-single-dt-datatable>
#' @return `character` CSS
#' @keywords internal
disableDoubleClickButtonCss <- function(id){
  sprintf("
          #%1$s > .dataTables_wrapper > table tbody td:nth-child(1) {pointer-events: none;}
          #%1$s > .dataTables_wrapper > table tbody td:nth-child(1)>div {pointer-events: auto;}
          ",id)
}

keyTableJS <- c(
    # Trigger doubleclick by enter
    "table.on('key', function(e, datatable, key, cell, originalEvent){",
    "  var targetName = originalEvent.target.localName;",
    "  var col = cell.index().column;",
    "  if(key == 13 && targetName == 'body' && col){", # do not modify first column
    "    $(cell.node()).trigger('dblclick.dt');",
    "  }",
    "});",
    # Blur old cell when moving
    "table.on('keydown', function(e){",
    "  var keys = [9,13,37,38,39,40];",
    "  if(e.target.localName == 'input' && keys.indexOf(e.keyCode) > -1){",
    "    $(e.target).trigger('blur');",
    "  }",
    "});",
    # Click new cell when moving
    "table.on('key-focus', function(e, datatable, cell, originalEvent){",
    "  var targetName = originalEvent.target.localName;",
    "  var type = originalEvent.type;",
    "  var col = cell.index().column;", 
    "  if(type == 'keydown' && targetName == 'input' && col){",  # do not modify first column
    "    if([9,37,38,39,40].indexOf(originalEvent.keyCode) > -1){",
    "      $(cell.node()).trigger('dblclick.dt');",
    "    }",
    "  }",
    "});"
)

# get autoFill edits https://laustep.github.io/stlahblog/posts/DTcallbacks.html
autoFillJs <- c(
    "var tbl = $(table.table().node());",
    "var id = tbl.closest('.datatables').attr('id');",
    "table.on('preAutoFill', function(e, datatable, cells){",
    "  var out = [];",
    "  for(var i = 0; i < cells.length; ++i){",
    "    var cells_i = cells[i];",
    "    for(var j = 0; j < cells_i.length; ++j){",
    "      var c = cells_i[j];",
    "      var value = (c.set === null || (typeof c.set === 'number' && isNaN(c.set)))? '' : c.set;", # null => problem in R
    "      out.push({",
    "        row: c.index.row + 1,",
    "        col: c.index.column,",
    "        value: value",
    "      });",
    "    }",
    "  }",
    "  Shiny.setInputValue(id + '_cells_filled:DT.cellInfo', out,  {priority: \"event\"});",
    "  table.rows().invalidate();", # this updates the column type
    "});"
)

test_that("passing on an empty tibble works", {
      data <- dplyr::as_tibble(sleep)
      data <- data[0,]
      
      # Try running the app manually
      ui <- eDTOutput("app")
      server <- function(input,output,session){
        eDTServer(id = "app", data = reactive(data))
      }
      shiny::shinyApp(ui, server)
      
      # Actual test
      shiny::testServer(
          app = eDTServer,
          args = list(data = data),
          expr = {
            session$setInputs(add = 1)       
            session$setInputs(save = 1)
            session$setInputs(confirmCommit = 1)
            expect_true(nrow(rv$committedData) == 1)
          })
})

test_that("Editing a single table column works", {
      data <- dplyr::tibble(one_column = c(1,2))
      
      # Try running the app manually
      ui <- eDTOutput("app")
      server <- function(input,output,session){
        eDTServer(id = "app", data = reactive(data))
      }
      shiny::shinyApp(ui, server)
      
      # Actual test
      shiny::testServer(
          app = eDTServer,
          args = list(data = data),
          expr = {
			# Make a cell edit
			session$setInputs(DT_cell_edit = data.frame( 
					row = 1,
					col = 1,
					value = 3
					))
			session$flushReact() 
            
			# Save changes
			session$setInputs(confirmCommit = 1)
			session$flushReact() 
						
			expect_true(rv$modifiedData[1,'one_column'] == 3)
			expect_true(rv$resultMessage == 'success.')		
          })
    })

test_that("Deletion of a row works", {
      data <- dplyr::tibble(id = 1:2, name = letters[1:2])
      
      shiny::testServer(
          app = eDTServer,
          args = list(data = data),
          expr = {
            session$flushReact()
            test_id = rv$modifiedData[1,"_editbl_identity"] # generated uuid
            session$setInputs(current_id = paste0('delete_row_', test_id))
            session$setInputs(delete = 1)
            session$flushReact()
            session$setInputs(confirmCommit = 1)
            expect_equal(nrow(result()),1)
          }
      )
    })

test_that("Can not delete row when canDeleteRow blocks it", {
      data <- dplyr::tibble(id = 1:2, name = letters[1:2])
      
      shiny::testServer(
          app = eDTServer,
          args = list(data = data, canDeleteRow = function(...){FALSE}),
          expr = {
            session$flushReact()
            test_id = rv$modifiedData[1,"_editbl_identity"] # generated uuid
            session$setInputs(current_id = paste0('delete_row_', test_id))
            session$setInputs(delete = 1)
            session$flushReact()
            session$setInputs(confirmCommit = 1)
            expect_equal(nrow(result()),2)
          }
      )
    })

test_that("working with selectInputDT works.", {
      songs <- tibble::tibble(
          song = c("Never gonna give you up", "Self Esteem"),
          artist_id = c(1,2)
          )
          
      artists <- dplyr::tibble(
         artist_id = c(1,2),
         first_name = c("Rick", "Dexter"),
         last_name = c('Astley', "Holland")
      )
      
      # Try running the app manually
      ui <- eDTOutput("app")
      server <- function(input,output,session){
        eDTServer(id = "app",
          data = songs,
          foreignTbls = list(
            foreignTbl(songs, artists, by = "artist_id", naturalKey = c("first_name", "last_name"))
          ),
          columnOrder = c("artist_id", "last_name", "first_name", "song")    
        )
      }
      
      # Test if using edit a second time still works
      # Reactivity problem.
      # TODO: write proper tests
      
      expect_true(TRUE)
      
      shiny::shinyApp(ui, server)
    })

test_that("Row dragging works when filter is on", {
      ui <- eDTOutput("app")
      server <- function(input,output,session){
        eDTServer(id = "app",
            data = mtcars,
            filter = "bottom")
      }

      # Test
      # 1. add on a filter for a column
      # 2. drag a cell using making use of the autofill extension
      # 3. remove the filter
      # 4. check if the same cell is still modified 
      # TODO: try to encode this
      expect_true(TRUE)
      
      shiny::shinyApp(ui, server)
    })

test_that("Can support all data types",{
      df = tibble(
          integer = 1L,
          double = 0.5,
          time = as.POSIXct('2020-01-01 01:02:03'),
          date = as.Date('2020-01-01')
          )
          ui <- eDTOutput("app")
          server <- function(input,output,session){
            eDTServer(id = "app",data = df)
          }
          
          # This just ensures the app doesn't crash on these different
          # data types
          expect_true(TRUE)
          
          shiny::shinyApp(ui, server)
    })

#' rows_insert implementation for DBI backends.
#' 
#' @examples
#' library(dplyr)
#' 
#' # Set up a test table
#' conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' artists_df <- data.frame(
#'          ArtistId = c(1,2),
#'          Name = c("AC/DC", "The Offspring")
#' )
#' DBI::dbWriteTable(conn, "Artist", artists_df)     
#'  
#' # Insert new row
#' artists <- tbl(conn, "Artist")
#' DBI::dbBegin(conn)
#' e_rows_insert(artists,
#'  data.frame(ArtistId = 999, Name = "testArtist"),
#'  in_place = TRUE)
#' 
#' DBI::dbRollback(conn)
#' DBI::dbDisconnect(conn)
#' 
#' @inheritParams e_rows_insert
#' @inherit e_rows_insert return details
#' 
#' @author Jasper Schelfhout
#' @export
e_rows_insert.tbl_dbi <- function(x, y, by = NULL, ..., copy = FALSE, in_place = FALSE){
  if(!in_place){
    stop("Can only edit in place")
  }
  
  if(copy){
    stop("copy TRUE not supported yet.")
  }
  
  if(is.null(by)){
    by <- colnames(x)[1]
  }
  
  if(in_place){
    lapply(seq_len(nrow(y)), function(i){
          rowInsert(
              conn = x$src$con,
              table =  get_db_table_name(x),
              values = as.list(y[i,])
          )
        })
    return(invisible(x))
  }
  return(x)
}

#' rows_update implementation for DBI backends.
#' @inheritParams e_rows_update
#' @inherit e_rows_update return details
#' @examples
#' library(dplyr)
#' 
#' # Set up a test table
#' conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' artists_df <- data.frame(
#'          ArtistId = c(1,2),
#'          Name = c("AC/DC", "The Offspring")
#' )
#' DBI::dbWriteTable(conn, "Artist", artists_df)     
#' 
#' # Update rows without changing the key.
#' artists <- tbl(conn, "Artist")
#' DBI::dbBegin(conn)
#' y <- data.frame(ArtistId = 1, Name = "DC/AC")
#' e_rows_update(
#'      x = artists,
#'      y = y,
#'      by = "ArtistId",
#'      in_place = TRUE)
#' DBI::dbRollback(conn)
#' 
#' # Update key values of rows.
#' DBI::dbBegin(conn)
#' y <- data.frame(ArtistId = 999, Name = "DC/AC")
#' match <- list(
#'    x = data.frame("ArtistId" = 1),
#'    y = data.frame("ArtistId" = 999)
#' )
#' e_rows_update(
#'     x = artists,
#'     y = y,
#'     match = match,
#'     by = "ArtistId",
#'     in_place = TRUE)
#' DBI::dbRollback(conn)
#' DBI::dbDisconnect(conn)
#' 
#' @author Jasper Schelfhout
#' @export
e_rows_update.tbl_dbi <- function(x, y, by = NULL, match = NULL,..., copy = FALSE, in_place = FALSE){
  if(!in_place){
    stop("Can only edit in place")
  }
  
  if(copy){
    stop("copy TRUE not supported yet.")
  }
  
  if(is.null(by)){
    by <- colnames(x)[1]
  }
  
  if(is.null(match)){
    match <- list(x = y[by], y = y[by])
  }
  
  lapply(seq_len(nrow(match$y)), function(i){
        yMatch <-  match$y[i,, drop = FALSE]
        values <- merge(yMatch, y, by = by)   
        xMatch <-  match$x[i,, drop = FALSE]
        
        rowUpdate(
            conn = x$src$con,
            table =  get_db_table_name(x),
            values = as.list(values),
            where = as.list(xMatch)
        )
      })
  return(invisible(x))  
}

#' Get name of the tbl in the database
#' @param x `tbl_dbi`
#' @return SQL, the table name as used in the database
get_db_table_name <- function(x){
  name <- coalesce(
      tryCatch({dbplyr::remote_table(x)}, error = function(e){NULL}), # Since dbplyr 2.4.0
      tryCatch({x$lazy_query$x$x}, error = function(e){NULL}), # tbl can apparently get more nested
      tryCatch({x$lazy_query$x}, error = function(e){NULL}), # normal place
      tryCatch({x$ops$x}, error = function(e){NULL}) # old tbl compatibility
  )
        
  if(is.null(name)){
    stop("Can not find in-database table name based on the tibble object.")
  }
  
  if(inherits(name,'dbplyr_table_ident')){
    attributes <- unclass(name)
    id <- list()
    if(!is.na(attributes$table)){
      id$table <- attributes$table
    }
    if(!is.na(attributes$schema)){
      id$schema <- attributes$schema
    }
    if(!is.na(attributes$catalog)){
      id$catalog <- attributes$catalog
    }
    name <- DBI::dbQuoteIdentifier(
        conn = x$src$con,
        x = do.call(DBI::Id, id)
    )
  } else {
    name <- DBI::SQL(name)
  }
    
  return(name)
}

#' Add a row to a table in the database.
#' 
#' @param conn database connection object as given by \code{\link[DBI]{dbConnect}}.
#' @param table character
#' @param values named list, row to add. Names are database column names. Unspecified columns will get database defaults.
#' @return integer number of affected rows.
#' @keywords internal
rowInsert <- function (
    conn,
    table,
    values){

  for(i in seq_along(values)){
      values[[i]] <- castToSQLSupportedType(values[[i]])
  }
  
  # Treat missing values as NULL and use database default.
  # This might need to be a flag in the function.
  # E.g. there is a difference between wanting to insert NA or let a database generate a default.
  values <- values[!is.na(values)]

  query <- glue::glue_sql(
      .con = conn,
      "INSERT INTO {`table`} ({`columns`*}) VALUES ({values*})",
      table = table,
      columns = names(values),
      values = values
  )
  
  DBI::dbExecute(
      conn,
      query)
}


#' Update rows in the database.
#' 
#' @param conn database connection object as given by \code{\link[DBI]{dbConnect}}.
#' @param table character
#' @param values named list, values to be set. Names are database column names.
#' @param where named list, values to filter on. Names are database column names. If NULL no filter is applied.
#' @return integer number of affected rows.
#' @keywords internal
rowUpdate <- function (
    conn,
    table,
    values,
    where){
  values <- lapply(values, function(x){
        castToSQLSupportedType(x)
      })
  
  query <- glue::glue_sql(
      .con =conn,
      "UPDATE {`table`} SET ({`columns`*}) = ({values*})",
      table = table,
      columns = names(values),
      values = values
  )
  
  if(length(where)){
    whereSQL  <- lapply(seq_along(where), function(i){
          whereSQL(
              conn = conn,
              table =  table,
              column = names(where)[i],
              operator = "in",
              values = where[[i]]
          )   
        })
    whereSQL <- paste(whereSQL, collapse = " AND ")
    query <- paste(query, "WHERE", whereSQL)
  }
  
  DBI::dbExecute(
      conn,
      query)
}

#' Cast the data type to something supported by SQL.
#' @param x single value or vector of values
#' @return x, possibly cast to different type
#' 
#' @author Jasper Schelfhout
#' @keywords internal
castToSQLSupportedType <- function(x){
  if(!inherits(x, c("numeric", "character"))){
    x <- as.character(x)
  }
  x
}

#' Generate where sql
#' 
#' @param conn database connection object as given by \code{\link[DBI]{dbConnect}}.
#' @param table character table name (or alias used in query)
#' @param column character column of table
#' @param values character vector of values
#' @param operator character
#' @return character sql
#' 
#' @author Jasper Schelfhout
#' @keywords internal
whereSQL <- function(
    conn,
    table,
    column,
    operator = 'in',
    values = NULL){
  
   values <- castToSQLSupportedType(values)
 
  if(length(values) > 0){  
    
    singleOperators <- c("like", "not like", "=", "!=", ">", "<", ">=", "<=", "<>")
    multiOperators <- c("in", "not in")
    if(length(values) == 1){
      operators <- c(singleOperators, multiOperators)
    } else {
      operators <- multiOperators
    } 
    
    if(!operator %in% operators){
      stop(sprintf("Should use on of following operators: %s",
              paste(operators, collapse = ", ")))
    }
    
    sql <- sprintf("{`table`}.{`column`} %s ({values*})",
        operator)
    sql <-  glue::glue_sql(
        .con = conn,
        sql,
        table = table,
        column = column,
        operator = operator,
        values = values
    )
  } else{
    sql = "TRUE"
  }
  sql
}

conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
test_df <- data.frame(
    id = 1:3,
    name = letters[1:3]
    )
DBI::dbCreateTable(conn, "test", test_df)
DBI::dbAppendTable(conn, "test", test_df)

test_tbl <- dplyr::tbl(conn, "test")

test_that("e_rows_insert.tbl_dbi works",{
      e_rows_insert.tbl_dbi(test_tbl, data.frame(id = 4, name = "insert"), in_place = TRUE)
      result <- DBI::dbGetQuery(conn, "select * from test where id = 4")$name
      expect_equal(result,'insert')
    })

test_that("e_rows_update.tbl_dbi works",{
      e_rows_update(test_tbl, data.frame(id = 1, name = "update"), by = "id", in_place = TRUE)
      result <- DBI::dbGetQuery(conn, "select * from test where id = 1")$name
      expect_equal(result,'update')
    })

DBI::dbDisconnect(conn)

#' Open interactive app to explore and modify data
#' 
#' @details When \code{\link{eDT}} is not used within the server of a shiny app, it will
#' call this function to start up a shiny app itself. Just as `DT::datatable()` displays a table
#' in the browser when called upon interactively.
#' 
#' @param ... arguments past to \code{\link{eDT}}
#' @importFrom shiny shinyApp
#' @return data (or a modified version thereof) once you click 'close'
#' @keywords internal
eDT_app <- function(...){
  args <- list(...)
  ui <- eDT_app_ui(eDTId = args$id)
  server <- function(input, output, session) {
    do.call(eDT_app_server, args)
  }
  shiny::runApp(list(ui = ui, server = server))  
}

#' UI of eDT_app
#' @param moduleId `character(1)` id to connect with eDT_app_server
#' @param eDTId `character(1)` id to connect \code{\link{eDTOutput}} to \code{\link{eDT}} within the module.
#' @importFrom shiny NS tagList actionButton
#' @return HTML
#' 
#' @author Jasper Schelfhout
#' @keywords internal
eDT_app_ui <- function(moduleId = "nevergonnagiveyouup", eDTId = "nevergonnaletyoudown") {
  ns <- NS(moduleId)
  tagList(
      actionButton(inputId = ns("close"), "label" = "close"),
      eDTOutput(id = ns(eDTId)),
  )
}

#' Server of eDT_app
#' @param moduleId `character(1)` id to connect with eDT_app_server
#' @param ... arguments passed to \link{eDT}
#' @importFrom shiny reactive moduleServer observeEvent stopApp reactiveValues
#' @importFrom dplyr tbl
#' @return moduleServer which on application stop returns version of x with made changes
#' 
#' @author Jasper Schelfhout
#' @keywords internal
eDT_app_server <- function(moduleId =  "nevergonnagiveyouup",...) {
  args <- list(...)
  moduleServer(
      moduleId,
      function(input, output, session) {  
        modifiedData <- do.call(eDTServer, args)$result
        
        observeEvent(input$close, {
              shiny::stopApp(modifiedData())
            })
      }
  )
}

#' UI part of a DT select input
#' @param id `character(1)` same one as used in \code{\link{selectInputDT_Server}}
#' @importFrom shiny NS uiOutput
#' @return HTML
#' @inherit selectInputDT_Server examples
#' @export
#' @author Jasper Schelfhout
selectInputDT_UI <- function(id){
  ns <- NS(id)
  uiOutput(ns("DT_UI"))
}

#' Server part to use a \code{\link[DT]{datatable}} as select input
#' 
#' @seealso `shiny::selectInput`. This function can be more convenient for selecting rows
#' with multiple columns.
#' 
#' @param id `character(1)` same one as used in \code{\link{selectInputDT_UI}}
#' @param choices `data.frame`
#' @param label `character(1)`
#' @param selected `data.frame` with rows available in `choices`.
#' @param multiple `logical`. Whether or not multiple row selection is allowed
#' 
#' @examples
#' ## Only run this example in interactive R sessions
#' if(interactive()){
#'   ui <- selectInputDT_UI('id')
#'   data <- data.frame(id = 1:3, name = letters[1:3])
#'   server <- function(input,output, session){
#'     selected = selectInputDT_Server('id', choices = data, selected = data[1,] )
#'     observe({print(selected())})
#'   }
#'   shiny::shinyApp(ui, server)
#' 
#' }
#' @importFrom shiny is.reactive reactive renderUI moduleServer
#' @importFrom DT renderDataTable datatable
#' @return A selection of rows from the `data.frame` provided under choices.
#' @export 
#' @author Jasper Schelfhout
selectInputDT_Server <- function(id,
    label = "",
    choices,
    selected = NULL,
    multiple = FALSE){
  moduleServer(
      id,
      function(input,output,session){
        # Make arguments reactive
        # Need to be explicit about environement. Otherwhise they overwrite themselves.
        # This way users can pass on both reactive an non reactive arguments
        argEnv <- parent.frame(3)
        
        if(!shiny::is.reactive(label)){
          label <- shiny::reactive(label, env = argEnv)
        }
        
        if(!shiny::is.reactive(choices)){
          choices <- shiny::reactive(choices, env = argEnv)
        }
        
        if(!shiny::is.reactive(selected)){
          selected <- shiny::reactive(selected, env = argEnv)
        }
        
        if(!shiny::is.reactive(multiple)){
          multiple <- shiny::reactive(multiple, env = argEnv)
        }
        
        observe({
              if(!multiple() && nrow(selected()) > 1){
                stop("Can not have more than 1 row selected.")
              }
            })
        
        ns <- session$ns
        
        hasSelection <- reactive({
              if(is.null(selected())){
                FALSE
              } else if ({ all(is.na(selected()))}){
                FALSE
              } else {
                TRUE
              }
            })
        
        data_selection_first <- reactive({
              if(hasSelection()){
                dt <- unique(rbind(selected(), choices()))
              } else {
                dt <- choices()
              }
              dt
            })
        
        output$DT_UI <- renderUI({
              tagList(
                  label(),
                  DT::DTOutput(ns("DT"))
              )
            })
        
        mode <- reactive({
              if(multiple()){
                mode = "multiple"
              } else {
                mode = "single"
              }
              mode
            })
        
        rowNrs <- reactive({
              if (hasSelection()){
                rowNrs = seq_len(nrow(selected()))
              } else {
                c()
              }
              rowNrs
            })
        
        output$DT <- DT::renderDataTable({
             data <- data_selection_first()             
              DT::datatable(
                  data,
                  rownames = TRUE,
                  options = list(
                      scrollX = TRUE,
                      columnDefs = list(
                          list(
                              visible = FALSE,
                              targets = c(0)) # hide row names https://github.com/rstudio/DT/issues/945
                      )
                  ),
                  filter = "top",
                  selection = list(
                      mode = mode(),
                      selected = rowNrs(),
                      target = 'row'))
            })
        
        reactive({
              data_selection_first()[input$DT_rows_selected,]
            })
      })
}

test_that("selectInputDT works", {  
      # Try running the app manually
      df <- data.frame(
          id = 1:10,
          name = letters[1:10]
          )
      
      ui <- selectInputDT_UI("app")
      
      server <- function(input,output,session){
        selectInputDT_Server(id = "app", label = "test", choices = df, selected = df[4,])
      }
      shiny::shinyApp(ui, server)
      
      # Actual test
      shiny::testServer(
          app = selectInputDT_Server,
          args = list(id = "app", label = "test", choices = df, selected = df[4,]),
          expr = {
            session$setInputs(DT_rows_selected = 2)                   
            expect_equal(data_selection_first()[input$DT_rows_selected,]$name, "a") # Not 'b' because selection always comes first
          })
    })

#' @importFrom tibble as_tibble
#' @import fontawesome
#' @keywords internal
testImports <- NULL

#' Cast columns to the type of the template
#' 
#' @param template `data.frame`
#' @param x `data.frame`
#' 
#' @details only affects columns in both the template and x
#' @keywords internal
coerceColumns <- function(template, x){
  for(col in intersect(names(x), names(template))){
    x[[col]] <-   coerceValue(x[[col]], template[[col]])
  }
  x
}

#' `DT::coerceValue` with better `POSIXct` support
#' 
#' @details Will assume UTC in case no timezone is specified.
#' 
#' @inheritParams DT::coerceValue
#' @importFrom DT coerceValue
#' 
#' @keywords internal
#' 
#' @author Jasper Schelfhout
coerceValue <- function(val,old){
  
  if (inherits(old, c('POSIXlt', 'POSIXct'))) {  
    
    # Try a bunch of formats supported by ISO
    newVal <- strptime(val, '%Y-%m-%dT%H:%M:%SZ', tz = 'UTC')
    if(is.na(newVal)){
      newVal <- strptime(val, '%Y-%m-%d %H:%M:%SZ', tz = 'UTC')
    }
    if(is.na(newVal)){
      newVal <- strptime(val, '%Y-%m-%d %H:%M:%S', tz = 'UTC')
    }
    if(is.na(newVal)){
      newVal <- strptime(val, '%Y-%m-%dT%H:%M:%S', tz = 'UTC')
    }
    
    if (inherits(old, 'POSIXlt')) return(newVal)
    return(as.POSIXct(newVal))
  } else if(inherits(old, "logical")){
    newVal <- as.logical(val)
  } else {
    return(DT::coerceValue(val,old))
  }
}

#' Connect to a database.
#'  
#' @details Connects by default to a test SQLite database originally obtained here:
#' [chinook_git](https://github.com/lerocha/chinook-database/blob/master/ChinookDatabase/DataSources/Chinook_Sqlite.sqlite)
#' 
#' @param dbname `character(0)`
#' @param drv database driver
#' @param ... arguments passed to `DBI::dbConnect`
#' @importFrom utils packageName
#' @examples 
#' 
#' conn <- connectDB()
#' DBI::dbDisconnect(conn)
#' 
#' @return database connection
#' @export
connectDB <- function(
    dbname = system.file("extdata", "chinook.sqlite", package = utils::packageName()),
    drv = RSQLite::SQLite(),
    ...
    ){
  DBI::dbConnect(
      dbname = dbname,
      drv = drv,
      ...
      )
}

#' Return first non `NULL` argument
#' @param ... set of arguments
#' 
#' @author Jasper Schelfhout
#' @keywords internal
coalesce <- function(...){
  args <- list(...)
  result <- NULL
  for(arg in args){
    if(!is.null(arg)){
      return(arg)
    }
  }
  return(result)
}

#' Cast columns in `data.frame` to editable types in datatable
#' @param data `data.frame`
#' @param cols `character` columns to perform casting on.
#' @return `data.frame` with some columns cast to another type
#' 
#' @author Jasper Schelfhout
#' @keywords internal
castForDisplay <- function(data, cols = colnames(data)){
  stopifnot(cols %in% colnames(data))
  
  for(col in cols){
    if(!inherits(data[[col]], c("integer", "character"))){
      data[[col]] <- as.character(data[[col]])
    }
  }
  data
}

#' Cast data to tbl
#' @param data object
#' @return tbl
#' 
#' @importFrom dplyr as_tibble is.tbl
#' 
#' @author Jasper Schelfhout
#' @keywords internal
castToTbl <- function(data){
  if(dplyr::is.tbl(data)){
    result <- data
  } else if(inherits(data, 'data.table')){
	requireNamespace('dtplyr')
    result <- dtplyr::lazy_dt(data)
  } else {
    result <- dplyr::as_tibble(data)
  }
  result
}

#' Cast `tbl` or `data.frame` x to the types of the template
#' 
#' @details If template is a `tbl` with database source, convert to an in-memory tibble with same data types instead.
#' @details Rownames might differ or get lost.
#' 
#' @param x `data.frame`, `tbl` or `data.table`
#' @param template `data.frame`, `tbl` or `data.table`
#' @return object containing data of x in the class and structure of the template.
#' 
#' @author Jasper Schelfhout
#' @keywords internal
castToTemplate <- function(x, template){
  if(!all(base::colnames(x) == base::colnames(template)))
    stop("Template and casted tbl should have exactly the same colums")
  
  rowNames <- attr(x, 'row.names')
  
  templateRow <- dplyr::collect(dplyr::filter(template, dplyr::row_number()==1))
  # In case the template is empty, templateRow will only contain the format (without content)
  # Add an empty row to make sure that the format will be correctly inherited
  if (nrow(templateRow) == 0) {
    templateRow[1,] <- NA
  }
  result <- rbind(templateRow,x)[-1,]
  
  # Tbl doesn't properly support row names
  if(!inherits(template, 'tbl')){
    try({rownames(result) <- rowNames}, silent = TRUE)
  }
 
  result
}

#' Cast tbl to class of template
#' @param tbl `tbl`
#' @param template tabular object like `data.frame` or `data.table` or `tbl`.
#' @return tbl cast to the type of template
#' @importFrom dplyr is.tbl
#' 
#' @author Jasper Schelfhout
#' @keywords internal
castFromTbl <- function(tbl, template){
  if(dplyr::is.tbl(template)){
    result <- tbl
  } else {
    result <- castToTemplate(tbl,template)
  }
  result
}

#' Standardize colnames argument to the format of named character vector
#' @inheritParams eDT
#' @importFrom dplyr tbl_vars
#' @return named character vector
#' 
#' @author Jasper Schelfhout
#' @keywords internal
standardizeArgument_colnames <- function(colnames, data){
  if(is.null(colnames)){
    result <- as.character(dplyr::tbl_vars(data))
    names(result) <- result
  } else if (is.numeric(colnames)) {
    result <- dplyr::tbl_vars(data)[colnames]
    names(result) = names(colnames)
  } else if (is.character(colnames)){
    if(!is.null(names(colnames))){
     result <- colnames
    } else {
     result <- dplyr::tbl_vars(data)[seq_len(length(colnames))] 
     names(result) <- colnames
    }
  }
  result
}

#' Standardized editable argument to be in the form of a list
#' @inheritParams eDT
#' @return list of the form `list(target = foo, ...)`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
standardizeArgument_editable <- function(
    editable,
    data
    ){
  if(is.logical(editable)){
    if(editable){
      return(list(target = "cell"))
    } else {
      return(list(target = "cell", disable = list(columns = seq_len(ncol(data))))) 
    }
  }

  if(is.character(editable)){
      return(list(target = editable))
  }
  
  if(is.list(editable)){
    return(editable)
  }
  
  stop("editable is not in a standard format.")
}

#' Replace instances of integer64 with actual NA values instead of weird default 9218868437227407266
#' 
#' @details [github issue](https://github.com/Rdatatable/data.table/issues/4561)
#' 
#' @param x `data.frame`
#' @return x with `integer64` columns set to `bit64::as.integer64(NA)`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
fixInteger64 <- function(x){
  for(column in dplyr::tbl_vars(x)){
    if(inherits(x[[column]], "integer64")){
      x[[column]] <- rep(bit64::as.integer64(NA), nrow(x))
    }
  }
  x
}

#' Get types of columns in a tbl
#' @param tbl `tbl`
#' @return named list with types of the colums
#' 
#' @importFrom dplyr type_sum collect
#' @importFrom utils head
#' 
#' @author Jasper Schelfhout
#' @keywords internal
getColumnTypeSums <- function(tbl){
  tbl %>% 
      head %>% 
      collect %>% 
      lapply(dplyr::type_sum)
}

#' Generate a custom button for \code{\link{eDT}}
#' 
#' @details
#' Combines elements of `shiny::actionButton` and [datatable options](https://datatables.net/reference/option/)
#' 
#' @param id `character(1)`, namespaced id
#' @param label `character(1)`
#' @param icon `shiny::icon`
#' @param disabled `logical`. Whether or not the button should start in a disabled state.
#' @return list to be used in `eDT(options = list(buttons = xxx))`
#' 
#' @examples
#' if(interactive()){
#' 
#'   ui <- eDTOutput("data")
#'   server <- function(input,output,session){
#'     b <- customButton('print', label = 'print')
#'     eDT_result <- eDT(id = "data", mtcars, options = list(buttons = list("save", b)))
#'     observeEvent(input$print,{
#'           print(eDT_result$state())
#'     })
#'   }
#'   shinyApp(ui,server)
#' }
#' 
#' @author Jasper Schelfhout
#' @export
customButton <- function(id, label, icon = "", disabled = FALSE){
      list(
          attr = list(
              id = id,
              class = "btn btn-default action-button shiny-bound-input",
              disabled = disabled
              ),
          extend = "",
          text = paste(as.character(icon), label, sep = " "),
          action = DT::JS(sprintf("function (e, dt, node, config ) {
                      Shiny.setInputValue('%s', true, {priority: 'event'});
                      }", id))
      )
}


#' Overwrite default settings with provided settings
#' @param defaults named character vector
#' @param settings named character vector
#' @return named character vector
#' 
#' @author Jasper Schelfhout
#' @keywords internal
overwriteDefaults <- function(defaults,settings){
  
  if(is.null(settings)){
    return(defaults)
  }
  
  result <- defaults
  for(entry in names(settings)){
    result[entry] <- settings[entry]
  }
  result
}

# Data.frame with multiple data types
df_typed <- data.frame(
    char = "x",
    num = 1,
    int = 1L,
    date = as.Date('2023-01-01'),
    time = as.POSIXct('2023-01-01 00:00:01', tz = "UTC")              
)

df_char <- data.frame(
    char = "x",
    num = "1",
    int = "1",
    date = "2023-01-01",
    time = '2023-01-01 00:00:01'
)

df_template <- df_typed[0,]



# Actual tests
test_that("coerceColumns works",{   
       result <- coerceColumns(df_template, df_char)     
       expect_equal(
           result,
           df_typed
           )
    })

test_that('coalesce returns first value',{
      result <- coalesce(x = 2, y = 3)
      expect_equal(result,2)
    })

test_that('coalesce skips null',{
      result <- coalesce(x = NULL, y = 3, z = 4)
      expect_equal(result,3)
    })

test_that('coalesce gives NULL when only NULL available',{
      result <- coalesce(x = NULL, y = NULL)
      expect_equal(result,NULL)
    })

test_that('coalesce gives when no parameters available',{
      result <- coalesce()
      expect_equal(result,NULL)
    })

test_that('castForDisplay gives when no parameters available',{
      result <- castForDisplay(df_typed)
      int <- c("int")
      notInt <- setdiff(colnames(df_typed), "int")
      expect_equal(result[,notInt],df_char[,notInt])
      expect_equal(result[,int],df_typed[,int])
    })

test_that("castToTbl works for data.frame",{
      result <- castToTbl(df_typed)
      expect_equal(result, tibble::tibble(
              char = "x",
              num = 1,
              int = 1L,
              date = as.Date("2023-01-01"),
              time = as.POSIXct("2023-01-01 00:00:01", tz = "UTC")
              ))
    })

test_that("castFromTbl can cast to data.frame",{
      result <- castFromTbl(tibble::as_tibble(mtcars), mtcars)
      expect_true(is.data.frame(result))
      expect_equal(colnames(result), colnames(mtcars))
    })

test_that("castFromTbl can cast to data.table",{
      result <- castFromTbl(tibble::as_tibble(mtcars), data.table::as.data.table(mtcars))
      expect_true(data.table::is.data.table(result))
      expect_equal(colnames(result), colnames(mtcars))
    })

test_that("castToTemplate can cast from data.frame to data.table",{
      result <- castToTemplate(x = iris, template = data.table::data.table(iris))
      expect_equal(result, data.table::as.data.table(iris))
    })
  
test_that("castToTemplate can cast from data.frame to data.table with empty template",{
      result <- castToTemplate(x = iris, template = data.table::data.table(iris[0,]))
      expect_equal(result, data.table::as.data.table(iris))
    })

test_that("castToTemplate can cast from data.frame to tbl",{
      result <- castToTemplate(x = iris,  template = dplyr::as_tibble(iris))
      expect_equal(result, dplyr::as_tibble(iris))
    })
  
test_that("castToTemplate can cast from data.frame to tbl with empty template",{
      result <- castToTemplate(x = iris,  template = dplyr::as_tibble(iris[0,]))
      expect_equal(result, dplyr::as_tibble(iris))
    })

test_that("castToTemplate can cast from tbl to data.frame",{
      result <- castToTemplate(x = dplyr::as_tibble(iris), template = iris)
      expect_equal(result, iris)
    })
  
test_that("castToTemplate can cast from tbl to data.frame with empty template",{
      result <- castToTemplate(x = dplyr::as_tibble(iris), template = iris[0,])
      expect_equal(result, iris)
    })

test_that("castToTemplate can cast from data.table to data.frame",{
      result <- castToTemplate(x = data.table::as.data.table(iris), template = iris)
      expect_equal(result, iris)
    })
  
test_that("castToTemplate can cast from data.table to data.frame with empty template",{
      result <- castToTemplate(x = data.table::as.data.table(iris), template = iris[0,])
      expect_equal(result, iris)
    })

test_that("castToTemplate can cast from data.table to tbl",{
      result <- castToTemplate(x = data.table::as.data.table(iris), template = dplyr::as_tibble(iris))
      expect_equal(result, dplyr::as_tibble(iris))
    })
  
test_that("castToTemplate can cast from data.table to tbl with empty template",{
      result <- castToTemplate(x = data.table::as.data.table(iris), template = dplyr::as_tibble(iris[0,]))
      expect_equal(result, dplyr::as_tibble(iris))
    })

test_that("castToTemplate can cast from tbl to data.table",{
      result <- castToTemplate(x = dplyr::as_tibble(iris), template = data.table::as.data.table(iris))
      expect_equal(result, data.table::as.data.table(iris))
    })
  
test_that("castToTemplate can cast from tbl to data.table with empty template",{
      result <- castToTemplate(x = iris, template = data.table::as.data.table(iris[0,]))
      expect_equal(result, data.table::as.data.table(iris))
    })

test_that("standardizeArgument_colnames returns named character for named integer",{
      result <- standardizeArgument_colnames(c("test" = 1), dplyr::as_tibble(mtcars))
      expect_equal(result, c("test" = "mpg"))
    })

test_that("standardizeArgument_colnames returns named character for unnamed character",{
      result <- standardizeArgument_colnames(c("test"), dplyr::as_tibble(mtcars))
      expect_equal(result, c("test" = "mpg"))
    })

test_that("standardizeArgument_colnames returns named character for named character",{
      result <- standardizeArgument_colnames(c("test" = "mpg"), dplyr::as_tibble(mtcars))
      expect_equal(result, c("test" = "mpg"))
    })

test_that("standardizeArgument_editable TRUE",{
      result <- standardizeArgument_editable(TRUE)
      expect_equal(result, list(target = "cell"))
    })

test_that("standardizeArgument_editable FALSE",{
      result <- standardizeArgument_editable(FALSE, mtcars)
      expect_equal(result, list(target = "cell", disable = list(columns = 1:11)))
    })

test_that("standardizeArgument_editable 'row'",{
      result <- standardizeArgument_editable('row')
      expect_equal(result, list(target = "row"))
    })

test_that("standardizeArgument_editable list",{
      result <- standardizeArgument_editable(list(target = "cell", numeric = 2))
      expect_equal(result, list(target = "cell", numeric = 2))
    })

test_that("standardizeArgument_editable Error",{
      expect_error(standardizeArgument_editable(1))
    })

test_that('overwriteDefaults works',{
      x = c(a = 'a', b = 'b')
      y = c(a='c')
      result <- overwriteDefaults(x,y)
      expect_equal(result, c(a = 'c', b = 'b'))
    })

#' An input UI for a `data.frame`
#' 
#' @details A new method for this can be added if you wish to alter the default behavior of the pop-up modals in \code{\link{eDT}}.
#' 
#' @param id `character(1)` module id
#' @param  ... arguments passed onto methods
#' @inherit inputServer examples
#' 
#' @return HTML. A set of input fields corresponding to the given row.
#' 
#' @author Jasper Schelfhout
#' @export
inputUI <- function(id, ...){
  UseMethod("inputUI")
}

#' An input server for a `data.frame`
#' 
#' @details A new method for this can be added if you wish to alter the default behavior of the pop-up modals in \code{\link{eDT}}.
#' 
#' @param id `character(1)` module id
#' @param data single row `data.frame`
#' @param ... further arguments for methods
#' @return modified version of data
#' 
#' @examples
#' if(interactive()){
#'   library(shiny)
#'   ui <- inputUI('id')
#'   server <- function(input,output,session){
#'     input <- inputServer("id", mtcars[1,])
#'     observe({print(input())})
#'   }
#' shinyApp(ui, server)
#' }
#' 
#' @author Jasper Schelfhout
#' @export
inputServer <- function(id, data, ...){
  UseMethod("inputServer")
}

#' UI part for modal with input fields for editing
#' 
#' @details The UI elements that have an id identical to a column name are used for updating the data.
#' 
#' @param id character module id
#' @param ... for compatibility with method
#' @return HTML. A set of input fields corresponding to the given row.
#' 
#' @author Jasper Schelfhout
#' @export
inputUI.default <- function(id, ...){
  ns <- NS(id)
  uiOutput(ns("inputUI"))
}

#' An input server for a `data.frame`
#' 
#' @details Reads all inputs ids that are identical to column names of the data
#' and updates the data.
#' 
#' @param id `character(1)` module id
#' @param data single row `data.frame`
#' @param notEditable `character` columns that should not be edited
#' @param colnames named `character`
#' @param foreignTbls list of foreignTbls. See \code{\link{foreignTbl}}
#' @param ... for compatibility with other methods
#' @return reactive modified version of data
#' 
#' @author Jasper Schelfhout
#' @export
inputServer.default <- function(id, data, colnames, notEditable, foreignTbls, ...){  
  missingColnames <- missing(colnames)
  missingForeignTbls <- missing(foreignTbls)
  missingNotEditable <- missing(notEditable)
  
  moduleServer(
      id,
      function(input, output, session) {
        ns <- session$ns
        rv <- reactiveValues(inputDTData = list())
        
        # Make arguments reactive
        # Need to be explicit about environement. Otherwhise they overwrite themselves.
        # This way users can pass on both reactive an non reactive arguments
        argEnv <- parent.frame(3)
        
        if(!shiny::is.reactive(data)){
          data <- reactive(data, env = argEnv)
        }
        
        if(missingColnames){
          colnames <- reactive({
                colnames <- base::colnames(data())
                names(colnames) <- colnames
                colnames
              })
        }
        else if(!shiny::is.reactive(colnames)){
          colnames <- reactive(colnames, env = argEnv)
        }
        
        if(missingNotEditable){
          notEditable <- reactive({
                character(0)
              })
        }
        else if(!shiny::is.reactive(notEditable)){
          notEditable <- reactive(notEditable, env = argEnv)
        }
        
        if(missingForeignTbls){
          foreignTbls <- reactive({
                list()
              })
        }
        else if(!shiny::is.reactive(foreignTbls)){
          foreignTbls <- reactive(foreignTbls, env = argEnv)
        }
        
        inputDTs <- reactive({
              inputDTs <- lapply(foreignTbls(), function(x){
                    if(length(x$naturalKey) > 1){
                      id <- gsub("-", "_", force(uuid::UUIDgenerate()))
                      list(
                          id = id,
                          choices = dplyr::collect(
                              dplyr::select(
                                  x$y,
                                  dplyr::all_of(as.character(x$naturalKey))
                              )
                          ),
                          selected = data()[,x$naturalKey]
                      )
                    }
                  })
              if(length(inputDTs)){
                inputDTs[sapply(inputDTs, is.null)] <- NULL
              }
              inputDTs
            })
        
        observe({
              inputDTData <- list()
              for(inputDT in inputDTs()){
                inputDTData[[inputDT$id]] <- do.call(selectInputDT_Server, inputDT)
              }
              rv$inputDTData <- inputDTData
            })
        
        inputDTCols <- reactive({
              unique(unlist(lapply(inputDTs(), function(x){
                            names(x$choices)
                          })))
            })
        
        normalInputs <- reactive({
              uiData <- data()[,setdiff(base::colnames(data()), notEditable()), drop = FALSE]
              uiData <- castToFactor(uiData, foreignTbls())        
              inputNormalCols <- setdiff(base::colnames(uiData), inputDTCols())
              
              inputs <- lapply(inputNormalCols, function(x){
                    if(x %in% colnames()){
                      label = names(colnames())[which(colnames() == x)]
                    } else {
                      label = x
                    }
                    
                    shinyInput(
                        x = uiData[[x]],
                        inputId = ns(x),
                        label = label, 
                        selected = uiData[,x]
                    )
                  })
              inputs
            })
        
        DTInputs <- reactive({
              lapply(inputDTs(), function(inputDT){
                    selectInputDT_UI(id = ns(inputDT$id))                 
                  })
            })
        
        output$inputUI <- renderUI({
              do.call(tagList,c(normalInputs(), DTInputs()))
            })

        newData <- reactive({
              data <- data()
              
              # Table inputs
              for(DTinput in rv$inputDTData){
                newData <- DTinput()
                data[,base::colnames(newData)] <- newData
              }
              
              # Normal inputs
              for(col in intersect(base::colnames(data), names(input))){
                data[,col] = coerceValue(input[[col]], data[,col])
              }
              data
            })
        newData
      }
  )
}

#' Get a shiny input for a column of a `tbl`
#' @param x column
#' @param inputId shiny input Id
#' @param label `character(1)`
#' @param selected object of class of x
#' @importFrom shiny numericInput dateInput selectInput textInput
#' @return shiny input
#' 
#' @author Jasper Schelfhout
#' @keywords internal
shinyInput <- function(x, inputId, label, selected){
  if(inherits(x, "logical")){
    selectInput(inputId = inputId, label = label, choices = c(" " = NA, "TRUE" = TRUE, "FALSE" = FALSE), selected = selected)
  }
  else if(inherits(x, "numeric")){
    numericInput(inputId = inputId, label = label, value = selected)
  }
  else if(inherits(x, "Date")){
    dateInput(inputId = inputId, label = label, value = selected)
  }
  else if(inherits(x, "factor")){
    selectInput(inputId = inputId, label = label, choices = sort(levels(x)), selected = selected)
  } else {
    textInput(inputId = inputId, label = label, value = as.character(selected))
  }
}

test_that("Input generation for single column data.frame works", {
      data <- data.frame(x = as.numeric(1))
      
      # Try running the app manually
      ui <- inputUI.default(id = "app")
      server <- function(input,output,session){
        inputServer.default(id = "app", data = reactive(data))
      }
      shiny::shinyApp(ui, server)
      
      # Actual test
      shiny::testServer(
          app = inputServer.default,
          args = list(data = data),
          expr = {
            session$flushReact()
            session$setInputs(x = 2)
            expect_equal(newData(), data.frame(x = 2))
          })
    })

#' Delete rows from a tibble
#' @details Mainly a wrapper around \code{\link[dplyr]{rows_delete}}.
#' Allows for specific implementations should the behavior differ from what's needed by `editbl`.
#' Reason for separate method is to avoid conflicts on package loading.
#' 
#' @inheritParams dplyr::rows_delete
#' @inherit dplyr::rows_delete return details
#' @export
e_rows_delete <- function(
		x,
		y,
		by = NULL,
		...,
		unmatched = c("error", "ignore"),
		copy = FALSE,
		in_place = FALSE){
	UseMethod("e_rows_delete")
}

#' @inherit e_rows_delete
#' @export
#' @importFrom dplyr rows_delete
e_rows_delete.default <- function(
		x,
		y,
		by = NULL,
		...,
		unmatched = c("error", "ignore"),
		copy = FALSE,
		in_place = FALSE){
	dplyr::rows_delete(
			x = x,
			y = y,
			by = by,
			... ,
			unmatched = unmatched,
			copy = copy,
			in_place = in_place)
}


#' Insert rows into a tibble
#' @details Mainly a wrapper around \code{\link[dplyr]{rows_insert}}.
#' Allows for specific implementations should the behavior differ from what's needed by `editbl`.
#' Reason for separate method is to avoid conflicts on package loading.
#' 
#' @inheritParams dplyr::rows_insert
#' @inherit dplyr::rows_insert return details
#' @export
e_rows_insert <- function(
    x,
    y,
    by = NULL,
    ...,
    conflict = c("error", "ignore"),
    copy = FALSE,
    in_place = FALSE){
  UseMethod("e_rows_insert")
}

#' @inherit e_rows_insert
#' @export
#' @importFrom dplyr rows_insert
e_rows_insert.default <- function(
    x,    
    y,
    by = NULL,
    ...,
    conflict = c("error", "ignore"),
    copy = FALSE,
    in_place = FALSE){
  dplyr::rows_insert(
      x = x,
      y = y,
      by = by,
      ... ,
      conflict = conflict,
      copy = copy,
      in_place = in_place)
}

#' Update rows of a tibble
#' @details Mainly a wrapper around \code{\link[dplyr]{rows_update}}.
#' Allows for specific implementations should the behavior differ from what's needed by `editbl`.
#' Reason for separate method is to avoid conflicts on package loading.
#' @param match named `list` consisting out of two equal length `data.frame`'s with columns defined in `by`.
#' This allows for updates of columns defined in `by`.
#' @inheritParams dplyr::rows_update
#' @inherit dplyr::rows_update return details
#' @export
e_rows_update <- function(
    x,
    y,
    by = NULL,
    ...,
    match,
    unmatched = c("error", "ignore"),
    copy = FALSE,
    in_place = FALSE){
  UseMethod("e_rows_update")
}

#' @inherit e_rows_update
#' @importFrom dplyr rows_update
#' @export
e_rows_update.default <- function(
    x,
    y,
    by = NULL,
    ...,
    match = match,
    unmatched = c("error", "ignore"),
    copy = FALSE,
    in_place = FALSE){
  dplyr::rows_update(
      x = x,
      y = y,
      by = by,
      ... ,
      match = match,
      unmatched = unmatched,
      copy = copy,
      in_place = in_place
  )
}

#' Start a transaction for a tibble
#' @param tbl `tbl`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
beginTransaction <- function(tbl){
  if(inherits(tbl, "tbl_dbi")){
    DBI::dbBegin(tbl$src$con)
  }
}

#' Start a transaction for a tibble
#' @param tbl `tbl`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
commitTransaction <- function(tbl){
  if(inherits(tbl, "tbl_dbi")){
    DBI::dbCommit(tbl$src$con)
  }
}

#' Start a transaction for a tibble
#' @param tbl `tbl`
#' 
#' @author Jasper Schelfhout
#' @keywords internal
rollbackTransaction <- function(tbl){
  if(inherits(tbl, "tbl_dbi")){
    DBI::dbRollback(tbl$src$con)
  }
}

conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
test_df <- data.frame(
    id = 1:3,
    name = letters[1:3]
)
DBI::dbCreateTable(conn, "test", test_df)
DBI::dbAppendTable(conn, "test", test_df)

test_tbl <- dplyr::tbl(conn, "test")

test_that("beginTransaction and commitTransaction works for tbl_dbi",{
      beginTransaction(test_tbl)
      
      # rows_delete will start nested transaction if 'y' is not of same source as 'x'
      dbplyr::db_copy_to(conn, table = 'y1_temp', values = data.frame(id = 2), in_transaction = FALSE)
      y <- tbl(conn, 'y1_temp')
      
      rows_delete(x = test_tbl, y = y, in_place = TRUE, copy = TRUE, unmatched = "ignore")
      commitTransaction(test_tbl)
      result <- DBI::dbGetQuery(conn, "select * from test")
      expect_true(nrow(result) == 2)
    })

test_that("beginTransaction and rollbackTransaction works for tbl_dbi",{
      beginTransaction(test_tbl)
      
      # rows_delete will start nested transaction if 'y' is not of same source as 'x'
      dbplyr::db_copy_to(conn, table = 'y2_temp', values = data.frame(id = 1), in_transaction = FALSE)
      y <- tbl(conn, 'y2_temp')
      rows_delete(x = test_tbl, y = y, in_place = TRUE, copy = TRUE, unmatched = "ignore")
      
      rollbackTransaction(test_tbl)
      result <- DBI::dbGetQuery(conn, "select * from test")
      expect_true(nrow(result) == 2)
    })

DBI::dbDisconnect(conn)

test_that("e_rows_update can update a row in a data.frame", {
      x <- mtcars
      x$id <- seq_len(nrow(x))
      y <- x[1,]
      y$mpg <- 1000
      
      result <- e_rows_update(x = x, y = y, by = "id")
      expect_equal(result[1,"mpg"][[1]], 1000)
    })

test_that("e_rows_insert add row to data.table", {
      x <- dtplyr::lazy_dt(data.table::data.table(mtcars))
      y <- mtcars[1,]
      result <- e_rows_insert(x,y)
      expect_equal(nrow(x) + 1, nrow(result))
    })

test_that("e_rows_update can update a row in a data.table", {
      dt <- data.table::data.table(mtcars)
      dt$id <- seq_len(nrow(dt))
      x <- dtplyr::lazy_dt(dt)
      y <- dt[1,]
      y$mpg <- 1000
      
      result <- e_rows_update(x = x, y = y, by = "id")
      expect_equal(data.table::as.data.table(result)[1,"mpg"][[1]], 1000)
    })

test_that("rows_delete can delete a row in a data.table", {
      dt <- data.table::data.table(mtcars)
      dt$id <- seq_len(nrow(dt))
      x <- dtplyr::lazy_dt(dt)
      y <- dt[1,]
      y$mpg <- 1000
      
      result <- rows_delete(x = x, y = y, by = "id")
      expect_equal(nrow(result), nrow(dt) - 1)
    })

#' Run a development app
#' @details This app prints some of the server objects and has a button to interactively browse the code.
#' This is useful for debugging and experimenting with new features.
#' 
#' @importFrom shiny shinyApp
#' @inherit shiny::shinyApp return
#' @keywords internal
runDevApp <- function(){
  conn <- connectDB()
  ui <- devUI(id = "app", conn = conn)
  server <- function(input, output, session) {
    devServer(id = "app", conn = conn)
  }
  shinyApp(ui, server)  
}

#' UI of the development app
#' @param id `character(1)`
#' @param conn database connection object as given by \code{\link[DBI]{dbConnect}}.
#' @importFrom shiny NS tagList selectInput verbatimTextOutput actionButton
#' @return HTML
#' 
#' @author Jasper Schelfhout
#' @keywords internal
devUI <- function(id, conn) {
  ns <- NS(id)
  tagList(
      actionButton(ns("browse"), label = "browse parent"),
      selectInput(
          inputId = ns("table"),
          label = "table",
          choices = DBI::dbListTables(conn)),
      eDTOutput(id = ns("editbl")),
      verbatimTextOutput(ns("modifiedData"))
  )
}

#' Server of the development app
#' @param id `character(1)`
#' @param conn database connection object as given by \code{\link[DBI]{dbConnect}}.
#' @importFrom shiny reactive moduleServer renderPrint
#' @importFrom dplyr tbl
#' @return NULL, just executes the module server.
#' @author Jasper Schelfhout
#' @keywords internal
devServer <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {
        data <- reactive({
              dplyr::tbl(conn, input$table)
            })
        
        modifiedData <- eDT(
            id = "editbl",
            data = data,
            options = list(
                columnDefs = list(list(
                        visible = TRUE,
                        targets = "_all"))),
            in_place = TRUE,
            filter = "top"
        )$result
        
        
        observeEvent(input$browse,{
              browser()
            })
        
        output$modifiedData <- renderPrint({
              str(modifiedData())
            })
        invisible()
      }
  )
}

---
title: "Edit tables in a relational database"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: modify tables in a relational database.}
  %\VignetteEngine{knitr::rmarkdown}
---

This document will learn you how to connect to a database and leverage `editbl`
to explore and edit your data.

The package `editbl` is actually developed with the main goal of making it as
easy as possible to work with a relational database. It
therefore also uses terminology of the [relational
model](https://en.wikipedia.org/wiki/Relational_model). It is
recommended to understand the concepts [foreign
key](https://en.wikipedia.org/wiki/Foreign_key), [natural
key](https://en.wikipedia.org/wiki/Natural_key) and [surrogate
key](https://en.wikipedia.org/wiki/Surrogate_key) before proceeding.

That being said, let's load some packages first.
```{r echo = TRUE, results = 'hide'}
library(dplyr)
library(shiny)
library(editbl)
```

## Basics of handling a database in R

The first thing you need is a database connection. Here we connect to an
[sqlite](https://en.wikipedia.org/wiki/SQLite) file, which is a portable
database format.

```{r echo = TRUE, results = 'hide'}
tmpFile <- tempfile(fileext = ".sqlite")
file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile) 
conn <-   DBI::dbConnect(
    dbname = tmpFile,
    drv = RSQLite::SQLite()
)
```

With the [{DBI} package](https://CRAN.R-project.org/package=DBI) you can list
all tables and schemas that are available in the database for exploratory
purposes.

```{r}
DBI::dbListTables(conn)
```

The [{dbplyr} package](https://CRAN.R-project.org/package=dbplyr) allows us to
create `tbl` objects that reference the database tables without actually pulling
all the data in memory. These objects can be manipulated using
[{dplyr}](https://CRAN.R-project.org/package=dplyr) syntax in a lazy-evaluated
way. Meaning you can work with big data that doesn't even fit on your computer!
You can read more about it [here](https://dbplyr.tidyverse.org).


```{r}
dplyr::tbl(conn, 'Album')
```


## A simple table editor

Let's give a shot at building our first shiny app. This one will modify the
`Albums` table in the database.

First we specify a [shiny
module](https://mastering-shiny.org/scaling-modules.html). If you are
unfamiliar with shiny modules, all you have to know is that they act as normal
shiny apps, with `ns()` wrapped around in/output id's. Though I do recommend
reading more about them, since it will enhance what you can do with
[shiny](https://shiny.posit.co/).


```{r}
dbUI <- function(id) {
  ns <- NS(id)
  fluidPage(
      eDTOutput(id = ns('Album'))
  )
}

dbServer <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {        
        Album <- eDT(
            id = "Album",
            key = "AlbumId",
            data = dplyr::tbl(conn, "Album"),
            in_place = TRUE 
        )
          
        invisible()
      }
  )
}
```

We need to put `in_place = TRUE`. Reason being that we want to actually
modify tables within the database. If this argument is set to `FALSE`, `editbl`
will return a edited copy of the original data. This is useful for
`data.frame`'s, but not for databases.

We also specify which column(s) are the
[key](https://en.wikipedia.org/wiki/Primary_key) of the table. This is the
minimal set of columns to uniquely identify a row. You don't have to specify
this argument, but it makes `eDT()` more efficient. E.g. this way it only has to
match on the key columns instead of all columns when doing updates or deletes.


**IMPORTANT**: Make sure the key you specify actually is unique across the
table!

`editbl` will not itself check for uniqueness of rows because it's
computationally expensive. Ensuring uniqueness is usually the responsibility of
the database.

Let's run the app:
```{r, screenshot.opts = list(vwidth = 700, vheight = 500), screenshot.alt = 'screenshots/howto_relational_db_1.png'}
shiny::shinyApp( 
    ui = dbUI('id'),
    server =  function(input, output,session){
      dbServer('id', conn)
    })
```

Great, you can now modify the `Album` table in the database!

Tip: if you do not fill in the `AlbumId` for a new row, the
database will automatically do it for you. If you prefer however to handle
setting defaults within your application, take a look at the `defaults` argument
of `eDT()`.

## Advanced table editor (flat, joined table).

Let's take in one step further. You don't really care about `AlbumId` and
`ArtistId` do you? These are just meaningless [surrogate
keys](https://en.wikipedia.org/wiki/Surrogate_key) to uniquely identify rows. So
why not hide them?

This we can do by joining the `Artists` on `ArtistId` ([foreign
key](https://en.wikipedia.org/wiki/Foreign_key)). 

```{r}
dbServer_hidden_keys <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {
        db_album <- dplyr::tbl(conn, "Album")
        db_artist <- dplyr::tbl(conn, "Artist")
        
        Album <- eDT(
            id = "Album",
            data = db_album,
            in_place = TRUE,
            foreignTbls = list(
              foreignTbl(
                x = db_album,
                y = db_artist,
                by = 'ArtistId',
                naturalKey = 'Name'
              )
             ),
             options = list(
               columnDefs = list(
                 list(visible=FALSE, targets=c("AlbumId","ArtistId"))
               )
             )
        )
        invisible()
      }
  )
}
```

**IMPORTANT** Make sure the `naturalKey` you specify actually is unique across
the table! (same reason as above)

```{r, screenshot.opts = list(vwidth = 700, vheight = 500), screenshot.alt = 'screenshots/howto_relational_db_2.png'}
shiny::shinyApp( 
    ui = dbUI('id'),
    server =  function(input, output,session){
      dbServer_hidden_keys('id', conn)
    })
```

Way more convenient isn't it? Make sure to click the 'edit' button. Here you can
see artists has now become a dropdown. You've actually also created your first
constraint, ensuring that only artists that exists within the 'Artist' table
can be filled in. So it is now easier to fill in this information and you ensure
data correctness, double win.

## Advanced table editor (partial table).

You can use `dplyr::filter()` to only show a subset of relevant rows to the user.
This can be used for improvement of performance/navigation or enforcing
row-level security.

```{r}

dbUI_advanced <- function(id) {
  ns <- NS(id)
  fluidPage(
      shiny::uiOutput(ns("artistSelector_UI")),
      eDTOutput(id = ns('Album'))
  )
}

dbServer_advanced <- function(id, conn) {
  moduleServer(
      id,
      function(input, output, session) {
        ns <- session$ns
                
                
        db_album <- dplyr::tbl(conn, "Album")
        db_artist <- dplyr::tbl(conn, "Artist")
        
        output$artistSelector_UI <- shiny::renderUI(
          shiny::selectInput(ns('artist'),
           label = 'artist',
           choices = db_artist %>% select(Name) %>% collect())
        )

        Album <- eDT(
            id = "Album",
            data = db_album,
            in_place = TRUE,
            foreignTbls = reactive(
            {
             req(input$artist)
             selected <- input$artist     
             list(
              foreignTbl(
                x = db_album,
                y = db_artist %>% filter(Name == selected),
                by = 'ArtistId',
                naturalKey = 'Name'
              )
             )}),
             options = list(
               columnDefs = list(
                 list(visible=FALSE, targets=c("AlbumId","ArtistId"))
               )
             )

        )
        invisible()
      }
  )
}
```
```{r, screenshot.opts = list(vwidth = 700, vheight = 500), screenshot.alt = 'screenshots/howto_relational_db_3.png'}
shiny::shinyApp( 
    ui = dbUI_advanced('id'),
    server =  function(input, output,session){
      dbServer_advanced('id', conn)
    })
```

## Above and beyond

* You can use [{dplyr}](https://CRAN.R-project.org/package=dplyr) to
  select, rename, reorder, filter... Cut and slice through your data as you seem
  fit.

* You can use all arguments and extensions of
  [{DT}](https://CRAN.R-project.org/package=DT) to customize your display as
  much as you like.

* `eDT()` can take any argument as a reactive. This allows for advanced shiny
  integration. The example above uses this feature to filter based
  on a user-selected artist.

* You can use [{dm}](https://CRAN.R-project.org/package=dm) to
  automatically extract the database model based on a connection.
  (`vignette("howto_relational_db_dm")`)

Combining all the above opens up a whole lot of possibilities. Get creative and
enjoy!

---
title: "Edit tables in a relational database - dm package"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: dm package integration for database editor.}
  %\VignetteEngine{knitr::rmarkdown}
---


If it is your first time using
`editbl`, make sure to first read the introductory vignette
on how to work with relational databases (`vignette("howto_relational_db")`).

This document describes how make use of the
[{dm} package](https://CRAN.R-project.org/package=dm). This package is
useful to extract relational data models out of a database into R. Which can
help in setting up a correct configuration for `editbl`.

```{r echo = TRUE, results = 'hide'}
library(dplyr)
library(shiny)
library(editbl)
library(dm)
```

## Creating a database connection

The first thing you need is a database connection. Here we connect to an
[sqlite](https://en.wikipedia.org/wiki/SQLite) file, which is a portable
database format.

```{r echo = TRUE, results = 'hide'}
tmpFile <- tempfile(fileext = ".sqlite")
file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile) 
conn <-   DBI::dbConnect(
    dbname = tmpFile,
    drv = RSQLite::SQLite()
)
```

## Setting up the data model

```{r echo = TRUE, results = 'hide'}
dm <- dm::dm_from_con(conn, learn_keys = FALSE)
```

For some databases this is all you need to do.
Currently `dm` can not (yet) learn the keys from SQLite.
Thus we still manually have to specify the primary and foreign keys.
```{r echo = TRUE, results = 'hide'}
dm <- dm %>%
    dm_add_pk(Artist, ArtistId) %>%
    dm_add_pk(Album, AlbumId) %>%
    dm_add_pk(Customer, CustomerId) %>%
    dm_add_pk(Employee, EmployeeId) %>%
    dm_add_pk(Genre, GenreId) %>%
    dm_add_pk(Invoice, InvoiceId) %>%
    dm_add_pk(InvoiceLine, InvoiceLineId) %>%
    dm_add_pk(MediaType, MediaTypeId) %>%
    dm_add_pk(Playlist, PlaylistId) %>%
    dm_add_pk(PlaylistTrack, c(PlaylistId, TrackId)) %>%
    dm_add_pk(Track, TrackId)
    
```

```{r echo = TRUE, results = 'hide'}
dm <- dm %>%
    dm_add_fk(
      table = Album,
      columns = ArtistId,
      ref_table = Artist) %>%
    dm_add_fk(
      table = Invoice,
      columns = CustomerId,
      ref_table = Customer) %>%
    dm_add_fk(
      table = InvoiceLine,
      columns = InvoiceId,
      ref_table = Invoice) %>%
    dm_add_fk(
      table = InvoiceLine,
      columns = TrackId,
      ref_table = Track) %>%
    dm_add_fk(
      table = PlaylistTrack,
      columns = TrackId,
      ref_table = Track) %>%
    dm_add_fk(
      table = PlaylistTrack,
      columns = PlaylistId,
      ref_table = Playlist) %>%
    dm_add_fk(
      table = Track,
      columns = AlbumId,
      ref_table = Album) %>%
    dm_add_fk(
      table = Track,
      columns = MediaTypeId,
      ref_table = MediaType) %>%
    dm_add_fk(
      table = Track,
      columns = GenreId,
      ref_table = Genre)
```

## Fully fletched table editor

A relational database consists out of many normalized tables. This is a perfect
model for storing data, since it avoids duplicate information. However, it
often leads to rather incomprehensible tables with a lot of 'id' columns. The
goal of this editor is therefore to give people the opportunity to edit a table
in its 'flat' form. Meaning you join all tables with additional information
based on these 'id' keys. See also this function of `dm`:

```{r}
dm::dm_flatten_to_tbl(dm, "Album", .recursive = TRUE)
```
As you can see, providing the `ArtistName` to a user is way more convenient than
just the `ArtistId`. 

`editbl::eDT` can handle similar joins by its
`foreignTbls` argument. Let us define a function that
extracts the needed information from a `dm` object.

```{r}
getForeignTbls <- function(dm, table){
  dm_fks <- dm::dm_get_all_fks(dm)
  dm_fks <- dm_fks[dm_fks$child_table == table,]
  tbl_list <- dm::dm_get_tables(dm)
  
  foreignTbls <- lapply(seq_len(nrow(dm_fks)), function(i){
      r <- dm_fks[i,]   
      x <- tbl_list[r$child_table][[1]]
      y <- dm::dm_flatten_to_tbl(dm, !!(r$parent_table), .recursive = TRUE)
      
      child_fks <- unlist(r$child_fk_cols)
      parent_fks <- unlist(r$parent_key_cols)
      
      # Renaming of parent colums to avoid naming conflicts
      # Done a bit heuristically here for convenience.
      lookup <- parent_fks
      names(lookup) <- child_fks
      other_parent_cols <- setdiff(colnames(y), parent_fks)
      names(other_parent_cols) <- paste(r$parent_table, other_parent_cols, sep = '.')
      lookup <- c(lookup, other_parent_cols)
      y <- y %>% dplyr::rename(all_of(lookup))
      
      editbl::foreignTbl(
        x = x,
        y = y,
        by = child_fks,
        naturalKey = colnames(y)
    )
    
  })
  foreignTbls
  }
  
```

Next, let's use this to build a shiny app.

```{r}
dbUI <- function(id) {
  ns <- NS(id)
  fluidPage(
      uiOutput(outputId = ns('selectUI')),
      eDTOutput(id = ns('DT'))
  )
}

dbServer <- function(id, dm) {
  moduleServer(
      id,
      function(input, output, session) {
        ns <- session$ns
        
        tables <- dm::dm_get_tables(dm)
        
        output$selectUI <- renderUI({
          selectInput(ns('table'), label = 'table', choices = names(tables))  
        })
        
        data <- reactive({
              req(input$table)
              tables[input$table][[1]]
              })
        
        foreignTbls <-  reactive({
              req(input$table)
              getForeignTbls(dm, input$table)
              })
        
        eDT(
            id = "DT",
            data = data,
            foreignTbls = foreignTbls,
            in_place = TRUE
        )
          
        invisible()
      }
  )
}

```


```{r, screenshot.opts = list(vwidth = 700, vheight = 500, delay = 1), screenshot.alt = 'screenshots/howto_relational_db_dm_1.png'}
shiny::shinyApp( 
    ui = dbUI('id'),
    server =  function(input, output,session){
      dbServer('id', dm)
    })
```

As you click the 'edit' button, you will notice you can now select rows from the
referenced tables. This makes it easier to navigate compared to just having
id's to work with.

---
title: "Switching from DT"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: adjust your `DT` code to use `eDT` instead.}
  %\VignetteEngine{knitr::rmarkdown}
---

Let's say you already use `DT::datatable()` to display your data, but want to
switch to `editbl::eDT()` to be able to edit it. What should you look out for?

* `eDTOutput()` uses an `id` argument instead of `outputId` since it's actually a
  [shiny module](https://mastering-shiny.org/scaling-modules.html).
* `eDT()` adds extra (hidden) columns to your `datatable`. Try to format using
  column names instead of indexes.
* Your `datatable` now exists within a module (e.g. child namespace). This means
  your own chosen `outputId` is now `moduleId-DT`. This influences for example
  the values accessible under `input`. Example: switch from
  `input$outputId_cell_clicked` to `input[["moduleId-DT_cell_clicked"]]`.
* `eDT()` accepts all arguments of `DT::datatable()`, but has some different
  defaults for convenience.
* Any additional formatting should be done by passing a function to  the
  `format` argument of `eDT()`.
* As always be careful when using
  [extensions](https://datatables.net/extensions/index) or custom javascript,
  not everything works well together. The
  [KeyTable](https://datatables.net/extensions/keytable/) and
  [AutoFill](https://datatables.net/extensions/autofill/) extensions of
  datatable are used by default and should be well integrated. 

Here is an example covering the above:

```{r echo = TRUE, results = 'hide'}
library(shiny)
library(DT)
library(editbl)
```

```{r, screenshot.opts = list(vwidth = 700), screenshot.alt = 'screenshots/howto_switch_from_DT_1.png'}
ui <- fluidPage(DTOutput("DT"))
server <- function(input, output, session){
  output$DT <- renderDataTable({
    datatable(mtcars)  %>%
     formatRound('disp', 1)
  })  
  observe({
      print(input[["DT_cell_clicked"]])
   })
}
shinyApp(ui, server)
```

Reworked into `eDT()`:
```{r, screenshot.opts = list(vwidth = 700), screenshot.alt = 'screenshots/howto_switch_from_DT_2.png'}
ui <- fluidPage(eDTOutput("DT"))
server <- function(input, output, session){
  editbl::eDT(
    id = "DT",
    data = mtcars,
    format = function(x){formatRound(x,'disp', 1)})
   
   observe({
      print(input[["DT-DT_cell_clicked"]])
   })
}
shinyApp(ui, server)
```

---
title: "Row level access"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: implement row level access.}
  %\VignetteEngine{knitr::rmarkdown}
---

Sometimes you do not want to give a user access to the entire dataset.
You can either hide rows or give read-only access.

## Hide rows completely

In this example we only allow user Mickey to see his own row.

We simply can use `dplyr::filter()` on the table.
Note that this is most useful if you combine this with backends that support 
`in_place` editing. E.g. you can retrieve only a subset
of rows from a database and specifically modify those.
Take a look at the 'relational database' vignettes for more information
on how to work with a database.

```{r, screenshot.opts = list(vwidth = 700, vheight = 500), , screenshot.alt = 'screenshots/howto_row_level_access_1.png'}
library(editbl)
library(shiny)
conn <- DBI::dbConnect(RSQLite::SQLite(), "")
df <- data.frame(
    user = c("Albert","Donald","Mickey"),
    email = c('albert@einstein.com', 'donald@duck.com', 'mickey@mouse.com')
)
DBI::dbWriteTable(conn, "characters", df)
tibble <- dplyr::tbl(conn, 'characters')

CURRENT_USER = 'Mickey'

shiny::shinyApp( 
    ui = editbl::eDTOutput('id'),
    server =  function(input, output,session){
      result <- eDT(id='id',
        data = tibble %>% filter(user == CURRENT_USER),
        in_place = TRUE
      )
    })

print(tibble)

DBI::dbDisconnect(conn)
```

## Read-only access

In this example we only allow user Mickey to modify his own row. In contrast to
the previous example, he can still read data from others.

The arguments `canEditRow` and `canDeleteRow` can be used
to specify logic describing if modfications are allowed.
The passed on logic shoud be a function with the argument `row`. This is a
single row of the displayed table in datatype `tibble`.

```{r, screenshot.opts = list(vwidth = 700, vheight = 500), , screenshot.alt = 'screenshots/howto_row_level_access_2.png'}
library(editbl)
df <- tibble::tibble(
    user = c("Albert","Donald","Mickey"),
    email = c('albert@einstein.com', 'donald@duck.com', 'mickey@mouse.com')
)

CURRENT_USER = 'Mickey'

rowModificationLogic <- function(row){
        if (row[,'user'] == CURRENT_USER){
          TRUE
        } else {
          FALSE  
        }
    }

shiny::shinyApp( 
    ui = editbl::eDTOutput('id'),
    server =  function(input, output,session){
      eDT(id='id',
        data = df,
        canEditRow = rowModificationLogic,
        canDeleteRow = rowModificationLogic
        )
    })

```

---
title: "Custom logic"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{How to: implement custom logic.}
  %\VignetteEngine{knitr::rmarkdown}
---

When saving your modifications, eDT will just update the data hereby only
checking the data types and foreign tables. Sometimes however, you might want
some specific custom behavior or extra checks that `editbl` doesn't support out
of the box. There are two main solutions to write custom logic for the
modifications that were made.

```{r echo = TRUE, results = 'hide'}
library(shiny)
library(editbl)
```

## Option 1

Add your own implementations for `e_rows_update()`, `e_rows_insert()` and/or
`e_rows_delete`. This is arguably the most flexible approach since it nicely
integrates with the behavior of eDT.

```{r, }
ui <- fluidPage(eDTOutput("DT"))

# Define a custom update function
e_rows_update.mtcars <- function(
    x,
    y,
    by = NULL,
    ...,
    match = match,
    unmatched = c("error", "ignore"),
    copy = FALSE,
    in_place = FALSE){		
		
		# Extra checks
		if(any(y[,"mpg"] < 0)){
			stop("mpg should be a positive value!")	
		}
		
		# Logging
		print("Updated rows:")
		print(y)
		
		# Drop extra class and execute default code
		class(x) <- class(x)[class(x) != 'mtcars']
		e_rows_update(
	      x = x,
	      y = y,
	      by = by,
	      ... ,
	      match = match,
	      unmatched = unmatched,
	      copy = copy,
	      in_place = in_place
	  )
	}

server <- function(input, output, session){
	
	# Add a new class to your data object
	df = tibble::tibble(mtcars)
	class(df) <- c("mtcars", class(df))

	  result = eDT(
	    id = "DT",
	    data = df
	)
	   
}
shinyApp(ui, server)
```


## Option 2

You can capture the modifications made by editbl and handle them downstream in
the app. This is more straightforward, but doesn't support adding any logic
before saving.


```{r, }
ui <- fluidPage(eDTOutput("DT"))

server <- function(input, output, session){
	  result = eDT(
	    id = "DT",
	    data = mtcars
	  )
	  
	  # Some logging
	  observe({
		print('These rows have been permanently deleted:')
		print(result$deleted())
	  })
	   
}
shinyApp(ui, server)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreignTbl.R
\name{getNonNaturalKeyCols}
\alias{getNonNaturalKeyCols}
\title{Get all columns that are not natural keys}
\usage{
getNonNaturalKeyCols(foreignTbls)
}
\arguments{
\item{foreignTbls}{list of foreign tbls as created by \code{\link{foreignTbl}}}
}
\value{
\code{character}
}
\description{
Get all columns that are not natural keys
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{coerceColumns}
\alias{coerceColumns}
\title{Cast columns to the type of the template}
\usage{
coerceColumns(template, x)
}
\arguments{
\item{template}{\code{data.frame}}

\item{x}{\code{data.frame}}
}
\description{
Cast columns to the type of the template
}
\details{
only affects columns in both the template and x
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT_app.R
\name{eDT_app_ui}
\alias{eDT_app_ui}
\title{UI of eDT_app}
\usage{
eDT_app_ui(moduleId = "nevergonnagiveyouup", eDTId = "nevergonnaletyoudown")
}
\arguments{
\item{moduleId}{\code{character(1)} id to connect with eDT_app_server}

\item{eDTId}{\code{character(1)} id to connect \code{\link{eDTOutput}} to \code{\link{eDT}} within the module.}
}
\value{
HTML
}
\description{
UI of eDT_app
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selectInputDT.R
\name{selectInputDT_Server}
\alias{selectInputDT_Server}
\title{Server part to use a \code{\link[DT]{datatable}} as select input}
\usage{
selectInputDT_Server(
  id,
  label = "",
  choices,
  selected = NULL,
  multiple = FALSE
)
}
\arguments{
\item{id}{\code{character(1)} same one as used in \code{\link{selectInputDT_UI}}}

\item{label}{\code{character(1)}}

\item{choices}{\code{data.frame}}

\item{selected}{\code{data.frame} with rows available in \code{choices}.}

\item{multiple}{\code{logical}. Whether or not multiple row selection is allowed}
}
\value{
A selection of rows from the \code{data.frame} provided under choices.
}
\description{
Server part to use a \code{\link[DT]{datatable}} as select input
}
\examples{
## Only run this example in interactive R sessions
if(interactive()){
  ui <- selectInputDT_UI('id')
  data <- data.frame(id = 1:3, name = letters[1:3])
  server <- function(input,output, session){
    selected = selectInputDT_Server('id', choices = data, selected = data[1,] )
    observe({print(selected())})
  }
  shiny::shinyApp(ui, server)

}
}
\seealso{
\code{shiny::selectInput}. This function can be more convenient for selecting rows
with multiple columns.
}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{e_rows_update.default}
\alias{e_rows_update.default}
\title{Update rows of a tibble}
\usage{
\method{e_rows_update}{default}(
  x,
  y,
  by = NULL,
  ...,
  match = match,
  unmatched = c("error", "ignore"),
  copy = FALSE,
  in_place = FALSE
)
}
\arguments{
\item{x, y}{A pair of data frames or data frame extensions (e.g. a tibble).
\code{y} must have the same columns of \code{x} or a subset.}

\item{by}{An unnamed character vector giving the key columns. The key columns
must exist in both \code{x} and \code{y}. Keys typically uniquely identify each row,
but this is only enforced for the key values of \code{y} when \code{rows_update()},
\code{rows_patch()}, or \code{rows_upsert()} are used.

By default, we use the first column in \code{y}, since the first column is
a reasonable place to put an identifier variable.}

\item{...}{Other parameters passed onto methods.}

\item{match}{named \code{list} consisting out of two equal length \code{data.frame}'s with columns defined in \code{by}.
This allows for updates of columns defined in \code{by}.}

\item{unmatched}{For \code{rows_update()}, \code{rows_patch()}, and \code{rows_delete()},
how should keys in \code{y} that are unmatched by the keys in \code{x} be handled?

One of:
\itemize{
\item \code{"error"}, the default, will error if there are any keys in \code{y} that
are unmatched by the keys in \code{x}.
\item \code{"ignore"} will ignore rows in \code{y} with keys that are unmatched by the
keys in \code{x}.
}}

\item{copy}{If \code{x} and \code{y} are not from the same data source,
and \code{copy} is \code{TRUE}, then \code{y} will be copied into the
same src as \code{x}.  This allows you to join tables across srcs, but
it is a potentially expensive operation so you must opt into it.}

\item{in_place}{Should \code{x} be modified in place? This argument is only
relevant for mutable backends (e.g. databases, data.tables).

When \code{TRUE}, a modified version of \code{x} is returned invisibly;
when \code{FALSE}, a new object representing the resulting changes is returned.}
}
\value{
An object of the same type as \code{x}. The order of the rows and columns of \code{x}
is preserved as much as possible. The output has the following properties:
\itemize{
\item \code{rows_update()} and \code{rows_patch()} preserve the number of rows;
\code{rows_insert()}, \code{rows_append()}, and \code{rows_upsert()} return all existing
rows and potentially new rows; \code{rows_delete()} returns a subset of the
rows.
\item Columns are not added, removed, or relocated, though the data may be
updated.
\item Groups are taken from \code{x}.
\item Data frame attributes are taken from \code{x}.
}

If \code{in_place = TRUE}, the result will be returned invisibly.
}
\description{
Update rows of a tibble
}
\details{
Mainly a wrapper around \code{\link[dplyr]{rows_update}}.
Allows for specific implementations should the behavior differ from what's needed by \code{editbl}.
Reason for separate method is to avoid conflicts on package loading.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{overwriteDefaults}
\alias{overwriteDefaults}
\title{Overwrite default settings with provided settings}
\usage{
overwriteDefaults(defaults, settings)
}
\arguments{
\item{defaults}{named character vector}

\item{settings}{named character vector}
}
\value{
named character vector
}
\description{
Overwrite default settings with provided settings
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreignTbl.R
\name{checkForeignTbls}
\alias{checkForeignTbls}
\title{Check if all rows in tbl fufill \code{foreignTbl} constraints}
\usage{
checkForeignTbls(tbl, foreignTbls)
}
\arguments{
\item{tbl}{\code{tbl}}

\item{foreignTbls}{list of foreign tbls as created by \code{\link{foreignTbl}}}
}
\value{
\code{logical} stating if tbl fufills all constraints imposed by all foreign tbls.
}
\description{
Check if all rows in tbl fufill \code{foreignTbl} constraints
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{eDTOutput}
\alias{eDTOutput}
\title{UI part of \code{\link{eDT}}}
\usage{
eDTOutput(id, ...)
}
\arguments{
\item{id}{\code{character(1)}}

\item{...}{arguments passed to \code{\link[DT]{DTOutput}}}
}
\value{
HTML
}
\description{
UI part of \code{\link{eDT}}
}
\details{
Works exactly like \code{\link[DT]{DTOutput}} apart from the fact that instead of the \code{outputId}
argument, \code{id} is requested. Reason being that this function is a UI to a shiny module.
This means that the datatable can be found under the id \code{'{namespace}-{id}-DT'} instead of \code{'{namespace}-{outputId}'}.

Also some minor CSS and javascript is executed for functional puposes.
}
\examples{
## Only run this example in interactive R sessions
if(interactive()){
  # tibble support
  modifiedData <- editbl::eDT(tibble::as_tibble(mtcars))

  # data.table support
  modifiedData <- editbl::eDT(dtplyr::lazy_dt(data.table::data.table(mtcars)))

  # database support
  tmpFile <- tempfile(fileext = ".sqlite")
  file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile)

  conn <- editbl::connectDB(dbname = tmpFile)
  modifiedData <- editbl::eDT(dplyr::tbl(conn, "Artist"), in_place = TRUE)
  DBI::dbDisconnect(conn)

  unlink(tmpFile)

  # Within shiny
  library(shiny)
  library(editbl)
  shinyApp(
    ui = fluidPage(fluidRow(column(12, eDTOutput('tbl')))),
    server = function(input, output) {
      eDT('tbl',iris,)
    }
  )

  # Custom inputUI
  editbl::eDT(mtcars, inputUI = function(id, data){
    ns <- NS(id)
    textInput(
    ns("mpg"),
    label = "mpg",
    value = data$mpg)})

  # Do not allow delete
  editbl::eDT(mtcars, canDeleteRow = FALSE)
}

}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demoApp.R
\name{runDemoApp}
\alias{runDemoApp}
\title{Run a demo app}
\usage{
runDemoApp(app = "database", ...)
}
\arguments{
\item{app}{demoApp to run. Options: database / mtcars / custom}

\item{...}{arguments passed onto the demoApp}
}
\value{
An object that represents the app. Printing the object or passing it
to \code{\link[shiny:runApp]{runApp()}} will run the app.
}
\description{
Run a demo app
}
\details{
These apps are for illustrative purposes.
}
\examples{
## Only run this example in interactive R sessions
if(interactive()){
 
  # Database
  tmpFile <- tempfile(fileext = ".sqlite")
  file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile)

  conn <- connectDB(dbname = tmpFile)

  runDemoApp(app = "database", conn = conn)
  DBI::dbDisconnect(conn)

  unlink(tmpFile)

  # mtcars
  runDemoApp(app = "mtcars")

  # Any tibble of your liking
  runDemoApp(app = "custom", dplyr::tibble(iris))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devApp.R
\name{devServer}
\alias{devServer}
\title{Server of the development app}
\usage{
devServer(id, conn)
}
\arguments{
\item{id}{\code{character(1)}}

\item{conn}{database connection object as given by \code{\link[DBI]{dbConnect}}.}
}
\value{
NULL, just executes the module server.
}
\description{
Server of the development app
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{canXXXRowTemplate}
\alias{canXXXRowTemplate}
\title{Re-usable documentation}
\usage{
canXXXRowTemplate(canEditRow, canCloneRow, canDeleteRow)
}
\arguments{
\item{canEditRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}

\item{canCloneRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}

\item{canDeleteRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}
}
\description{
Re-usable documentation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{coerceValue}
\alias{coerceValue}
\title{\code{DT::coerceValue} with better \code{POSIXct} support}
\usage{
coerceValue(val, old)
}
\arguments{
\item{val}{A character string.}

\item{old}{An old value, whose type is the target type of \code{val}.}
}
\description{
\code{DT::coerceValue} with better \code{POSIXct} support
}
\details{
Will assume UTC in case no timezone is specified.
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{castToTemplate}
\alias{castToTemplate}
\title{Cast \code{tbl} or \code{data.frame} x to the types of the template}
\usage{
castToTemplate(x, template)
}
\arguments{
\item{x}{\code{data.frame}, \code{tbl} or \code{data.table}}

\item{template}{\code{data.frame}, \code{tbl} or \code{data.table}}
}
\value{
object containing data of x in the class and structure of the template.
}
\description{
Cast \code{tbl} or \code{data.frame} x to the types of the template
}
\details{
If template is a \code{tbl} with database source, convert to an in-memory tibble with same data types instead.

Rownames might differ or get lost.
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{evalCanEditRow}
\alias{evalCanEditRow}
\title{Determine if a row can be edited}
\usage{
evalCanEditRow(row, canEditRow = TRUE, statusCol = "_editbl_status")
}
\arguments{
\item{row}{\code{tibble}, single row.}

\item{canEditRow}{\code{function} with argument 'row' defining logic on wether or
not the row can be modified. Can also be \code{logical} TRUE or FALSE.}

\item{statusCol}{\code{character(1)} name of column with general status (e.g. modified or not).}
}
\value{
\code{boolean}
}
\description{
Determine if a row can be edited
}
\details{
calling this around the user passed on function ensures
that newly inserted rows are being excempt from the logic.
Moreover, the output of the function can be checked.
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinyInput.R
\name{inputUI}
\alias{inputUI}
\title{An input UI for a \code{data.frame}}
\usage{
inputUI(id, ...)
}
\arguments{
\item{id}{\code{character(1)} module id}

\item{...}{arguments passed onto methods}
}
\value{
HTML. A set of input fields corresponding to the given row.
}
\description{
An input UI for a \code{data.frame}
}
\details{
A new method for this can be added if you wish to alter the default behavior of the pop-up modals in \code{\link{eDT}}.
}
\examples{
if(interactive()){
  library(shiny)
  ui <- inputUI('id')
  server <- function(input,output,session){
    input <- inputServer("id", mtcars[1,])
    observe({print(input())})
  }
shinyApp(ui, server)
}

}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{customButton}
\alias{customButton}
\title{Generate a custom button for \code{\link{eDT}}}
\usage{
customButton(id, label, icon = "", disabled = FALSE)
}
\arguments{
\item{id}{\code{character(1)}, namespaced id}

\item{label}{\code{character(1)}}

\item{icon}{\code{shiny::icon}}

\item{disabled}{\code{logical}. Whether or not the button should start in a disabled state.}
}
\value{
list to be used in \code{eDT(options = list(buttons = xxx))}
}
\description{
Generate a custom button for \code{\link{eDT}}
}
\details{
Combines elements of \code{shiny::actionButton} and \href{https://datatables.net/reference/option/}{datatable options}
}
\examples{
if(interactive()){

  ui <- eDTOutput("data")
  server <- function(input,output,session){
    b <- customButton('print', label = 'print')
    eDT_result <- eDT(id = "data", mtcars, options = list(buttons = list("save", b)))
    observeEvent(input$print,{
          print(eDT_result$state())
    })
  }
  shinyApp(ui,server)
}

}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreignTbl.R
\name{foreignTbl}
\alias{foreignTbl}
\title{Create a foreign tibble}
\usage{
foreignTbl(
  x,
  y,
  by = intersect(dplyr::tbl_vars(x), dplyr::tbl_vars(y)),
  naturalKey = dplyr::tbl_vars(y),
  allowNew = FALSE
)
}
\arguments{
\item{x}{\code{tbl}. The referencing table.}

\item{y}{\code{tbl}. The referenced table.}

\item{by}{\code{character}. Column names to match on.
Note that you should rename and/or typecast the columns in y should they not exactly match the columns in x.}

\item{naturalKey}{\code{character}. The columns that form the natural key in y.
These are the only ones that can actually get modified in \code{eDT} when changing cells in the table.
Reasoning being that these columns should be sufficient to uniquely identify a row in the referenced table.
All other columns will be automatically fetched and filled in.}

\item{allowNew}{\code{logical}. Whether or not new values are allowed. If \code{TRUE},
the rows in the foreignTbl will only be used as suggestions, not restrictions.}
}
\value{
List with unmodified arguments. However, they have now been checked for validity.
\itemize{
\item y, see argument \code{y}.
\item by, see argument \code{by}.
\item naturalKey, see argument \code{naturalKey}.
\item allowNew, see argument \code{allowNew}
}
}
\description{
Create a foreign tibble
}
\details{
This is a tibble that can be passed onto \code{\link{eDT}} as a referenced table.

It is the equivalent of a database table to which the \code{data} tbl of eDT has a foreign key.

It will be merged with the tbl passed onto the \code{data} argument allowing to provide restrictions
for certain columns.

Note that row uniqueness for the columns used in \code{by} and \code{naturalKey} is assumed.
This assumption will however not be checked since it is an expensive operation on big datasets.
However, if violated, it might give errors or unexpected results during usage of the eDT module.
}
\examples{
a <- tibble::tibble(
   first_name = c("Albert","Donald","Mickey"),
   last_name_id = c(1,2,2)
 )

b <-  foreignTbl(
 a,
 tibble::tibble(
     last_name = c("Einstein", "Duck", "Mouse"),
     last_name_id = c(1,2,3)
   ),
 by = "last_name_id",
 naturalKey = "last_name"
)

## Only run this in interactive R sessions
if(interactive()){
  eDT(a,
   foreignTbls = list(b),
   options = list(columnDefs = list(list(visible=FALSE, targets="last_name_id")))
  )
 }


}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demoApp.R
\name{demoUI_custom}
\alias{demoUI_custom}
\title{UI of the demo mtcars app}
\usage{
demoUI_custom(id)
}
\arguments{
\item{id}{\code{character(1)}}
}
\value{
HTML
}
\description{
UI of the demo mtcars app
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{createEditButtonHTML}
\alias{createEditButtonHTML}
\title{Generate HTML for an in-row edit button}
\usage{
createEditButtonHTML(ns, suffix, disabled = FALSE)
}
\arguments{
\item{ns}{\code{character(1)} namespace}

\item{suffix}{\code{character(1)} id of the row}

\item{disabled}{\code{logical(1)} wether or not the button has to be disabled}
}
\value{
\code{character(1)} HTML
}
\description{
Generate HTML for an in-row edit button
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{initData}
\alias{initData}
\title{Add some extra columns to data to allow for / keep track of modifications}
\usage{
initData(
  data,
  ns,
  buttonCol = "buttons",
  statusCol = "_editbl_status",
  deleteCol = "_editbl_deleted",
  iCol = "i",
  canDeleteRow = TRUE,
  canEditRow = TRUE,
  canCloneRow = TRUE
)
}
\arguments{
\item{data}{\code{data.frame}}

\item{ns}{namespace function}

\item{buttonCol}{\code{character(1)} name of column with buttons}

\item{statusCol}{\code{character(1)} name of column with general status (e.g. modified or not).}

\item{deleteCol}{\code{character(1)} name of the column with deletion status.}

\item{iCol}{\code{character(1)} name of column containing a unique identifier.}

\item{canDeleteRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}

\item{canEditRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}

\item{canCloneRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}
}
\value{
data with extra columns buttons, status, i.
}
\description{
Add some extra columns to data to allow for / keep track of modifications
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{coalesce}
\alias{coalesce}
\title{Return first non \code{NULL} argument}
\usage{
coalesce(...)
}
\arguments{
\item{...}{set of arguments}
}
\description{
Return first non \code{NULL} argument
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT_app.R
\name{eDT_app}
\alias{eDT_app}
\title{Open interactive app to explore and modify data}
\usage{
eDT_app(...)
}
\arguments{
\item{...}{arguments past to \code{\link{eDT}}}
}
\value{
data (or a modified version thereof) once you click 'close'
}
\description{
Open interactive app to explore and modify data
}
\details{
When \code{\link{eDT}} is not used within the server of a shiny app, it will
call this function to start up a shiny app itself. Just as \code{DT::datatable()} displays a table
in the browser when called upon interactively.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl_dbi.R
\name{rowUpdate}
\alias{rowUpdate}
\title{Update rows in the database.}
\usage{
rowUpdate(conn, table, values, where)
}
\arguments{
\item{conn}{database connection object as given by \code{\link[DBI]{dbConnect}}.}

\item{table}{character}

\item{values}{named list, values to be set. Names are database column names.}

\item{where}{named list, values to filter on. Names are database column names. If NULL no filter is applied.}
}
\value{
integer number of affected rows.
}
\description{
Update rows in the database.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{castToTbl}
\alias{castToTbl}
\title{Cast data to tbl}
\usage{
castToTbl(data)
}
\arguments{
\item{data}{object}
}
\value{
tbl
}
\description{
Cast data to tbl
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selectInputDT.R
\name{selectInputDT_UI}
\alias{selectInputDT_UI}
\title{UI part of a DT select input}
\usage{
selectInputDT_UI(id)
}
\arguments{
\item{id}{\code{character(1)} same one as used in \code{\link{selectInputDT_Server}}}
}
\value{
HTML
}
\description{
UI part of a DT select input
}
\examples{
## Only run this example in interactive R sessions
if(interactive()){
  ui <- selectInputDT_UI('id')
  data <- data.frame(id = 1:3, name = letters[1:3])
  server <- function(input,output, session){
    selected = selectInputDT_Server('id', choices = data, selected = data[1,] )
    observe({print(selected())})
  }
  shiny::shinyApp(ui, server)

}
}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{disableDoubleClickButtonCss}
\alias{disableDoubleClickButtonCss}
\title{Function to generate CSS to disable clicking events on a column}
\usage{
disableDoubleClickButtonCss(id)
}
\arguments{
\item{id}{\code{character(1)} namespaced id of the datatable}
}
\value{
\code{character} CSS
}
\description{
Function to generate CSS to disable clicking events on a column
}
\details{
\url{https://stackoverflow.com/questions/60406027/how-to-disable-double-click-reactivity-for-specific-columns-in-r-datatable}

\url{https://stackoverflow.com/questions/75406546/apply-css-styling-to-a-single-dt-datatable}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{commitTransaction}
\alias{commitTransaction}
\title{Start a transaction for a tibble}
\usage{
commitTransaction(tbl)
}
\arguments{
\item{tbl}{\code{tbl}}
}
\description{
Start a transaction for a tibble
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demoApp.R
\name{runDemoApp_DB}
\alias{runDemoApp_DB}
\title{Run a demo app}
\usage{
runDemoApp_DB()
}
\value{
An object that represents the app. Printing the object or passing it
to \code{\link[shiny:runApp]{runApp()}} will run the app.
}
\description{
Run a demo app
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl_dbi.R
\name{whereSQL}
\alias{whereSQL}
\title{Generate where sql}
\usage{
whereSQL(conn, table, column, operator = "in", values = NULL)
}
\arguments{
\item{conn}{database connection object as given by \code{\link[DBI]{dbConnect}}.}

\item{table}{character table name (or alias used in query)}

\item{column}{character column of table}

\item{operator}{character}

\item{values}{character vector of values}
}
\value{
character sql
}
\description{
Generate where sql
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl_dbi.R
\name{rowInsert}
\alias{rowInsert}
\title{Add a row to a table in the database.}
\usage{
rowInsert(conn, table, values)
}
\arguments{
\item{conn}{database connection object as given by \code{\link[DBI]{dbConnect}}.}

\item{table}{character}

\item{values}{named list, row to add. Names are database column names. Unspecified columns will get database defaults.}
}
\value{
integer number of affected rows.
}
\description{
Add a row to a table in the database.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT_app.R
\name{eDT_app_server}
\alias{eDT_app_server}
\title{Server of eDT_app}
\usage{
eDT_app_server(moduleId = "nevergonnagiveyouup", ...)
}
\arguments{
\item{moduleId}{\code{character(1)} id to connect with eDT_app_server}

\item{...}{arguments passed to \link{eDT}}
}
\value{
moduleServer which on application stop returns version of x with made changes
}
\description{
Server of eDT_app
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinyInput.R
\name{inputServer.default}
\alias{inputServer.default}
\title{An input server for a \code{data.frame}}
\usage{
\method{inputServer}{default}(id, data, colnames, notEditable, foreignTbls, ...)
}
\arguments{
\item{id}{\code{character(1)} module id}

\item{data}{single row \code{data.frame}}

\item{colnames}{named \code{character}}

\item{notEditable}{\code{character} columns that should not be edited}

\item{foreignTbls}{list of foreignTbls. See \code{\link{foreignTbl}}}

\item{...}{for compatibility with other methods}
}
\value{
reactive modified version of data
}
\description{
An input server for a \code{data.frame}
}
\details{
Reads all inputs ids that are identical to column names of the data
and updates the data.
}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{standardizeArgument_colnames}
\alias{standardizeArgument_colnames}
\title{Standardize colnames argument to the format of named character vector}
\usage{
standardizeArgument_colnames(colnames, data)
}
\arguments{
\item{colnames}{if missing, the column names of the data; otherwise it can be
an unnamed character vector of names you want to show in the table header
instead of the default data column names; alternatively, you can provide a
\emph{named} numeric or character vector of the form \code{'newName1' = i1,
'newName2' = i2} or \code{c('newName1' = 'oldName1', 'newName2' =
'oldName2', ...)}, where \code{newName} is the new name you want to show in
the table, and \code{i} or \code{oldName} is the index of the current
column name}

\item{data}{\code{tbl}. When using non-standard tibble objects, install the required package implementing specific methods.
(\code{data.table}: dtplyr, spark: sparklyr, generic database: dbplyr,...).
While other tabular objects like \code{data.frame} are supported, explicitly casting to \code{tbl} is recommended.}
}
\value{
named character vector
}
\description{
Standardize colnames argument to the format of named character vector
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{eDT}
\alias{eDT}
\title{Create a modifieable datatable.}
\usage{
eDT(
  data,
  options = list(dom = "Bfrtlip", keys = TRUE, ordering = FALSE, autoFill = list(update =
    FALSE, focus = "focus"), buttons = list("add", "undo", "redo", "save")),
  class = "display",
  callback = NULL,
  rownames = FALSE,
  colnames = NULL,
  container,
  caption = NULL,
  filter = c("none", "bottom", "top"),
  escape = TRUE,
  style = "auto",
  width = NULL,
  height = NULL,
  elementId = NULL,
  fillContainer = getOption("DT.fillContainer", NULL),
  autoHideNavigation = getOption("DT.autoHideNavigation", NULL),
  selection = "none",
  extensions = c("KeyTable", "AutoFill", "Buttons"),
  plugins = NULL,
  editable = list(target = "cell"),
  id,
  keys = NULL,
  in_place = FALSE,
  format = function(x) {
     x
 },
  foreignTbls = list(),
  columnOrder = c(),
  statusColor = c(insert = "#e6e6e6", update = "#32a6d3", delete = "#e52323"),
  inputUI = editbl::inputUI,
  defaults = tibble(),
  cloneDefaults = tibble(),
  env = environment(),
  canEditRow = TRUE,
  canDeleteRow = TRUE,
  canCloneRow = TRUE,
  utilityColumns = NULL
)
}
\arguments{
\item{data}{\code{tbl}. When using non-standard tibble objects, install the required package implementing specific methods.
(\code{data.table}: dtplyr, spark: sparklyr, generic database: dbplyr,...).
While other tabular objects like \code{data.frame} are supported, explicitly casting to \code{tbl} is recommended.}

\item{options}{a list of initialization options (see
\url{https://datatables.net/reference/option/}); the character options
wrapped in \code{\link[htmlwidgets]{JS}()} will be treated as literal
JavaScript code instead of normal character strings; you can also set
options globally via \code{\link{options}(DT.options = list(...))}, and
global options will be merged into this \code{options} argument if set}

\item{class}{the CSS class(es) of the table; see
\url{https://datatables.net/manual/styling/classes}}

\item{callback}{the body of a JavaScript callback function with the argument
\code{table} to be applied to the DataTables instance (i.e. \code{table})}

\item{rownames}{\code{TRUE} (show row names) or \code{FALSE} (hide row names)
or a character vector of row names; by default, the row names are displayed
in the first column of the table if exist (not \code{NULL})}

\item{colnames}{if missing, the column names of the data; otherwise it can be
an unnamed character vector of names you want to show in the table header
instead of the default data column names; alternatively, you can provide a
\emph{named} numeric or character vector of the form \code{'newName1' = i1,
'newName2' = i2} or \code{c('newName1' = 'oldName1', 'newName2' =
'oldName2', ...)}, where \code{newName} is the new name you want to show in
the table, and \code{i} or \code{oldName} is the index of the current
column name}

\item{container}{a sketch of the HTML table to be filled with data cells; by
default, it is generated from \code{htmltools::tags$table()} with a table
header consisting of the column names of the data}

\item{caption}{the table caption; a character vector or a tag object
generated from \code{htmltools::tags$caption()}}

\item{filter}{whether/where to use column filters; \code{none}: no filters;
\code{bottom/top}: put column filters at the bottom/top of the table; range
sliders are used to filter numeric/date/time columns, select lists are used
for factor columns, and text input boxes are used for character columns; if
you want more control over the styles of filters, you can provide a named
list to this argument; see Details for more}

\item{escape}{whether to escape HTML entities in the table: \code{TRUE} means
to escape the whole table, and \code{FALSE} means not to escape it;
alternatively, you can specify numeric column indices or column names to
indicate which columns to escape, e.g. \code{1:5} (the first 5 columns),
\code{c(1, 3, 4)}, or \code{c(-1, -3)} (all columns except the first and
third), or \code{c('Species', 'Sepal.Length')}; since the row names take
the first column to display, you should add the numeric column indices by
one when using \code{rownames}}

\item{style}{either \code{'auto'}, \code{'default'}, \code{'bootstrap'}, or
\code{'bootstrap4'}. If \code{'auto'}, and a **bslib** theme is
currently active, then bootstrap styling is used in a way that "just works"
for the active theme. Otherwise,
\href{https://datatables.net/manual/styling/classes}{DataTables
\code{'default'} styling} is used. If set explicitly to \code{'bootstrap'}
or \code{'bootstrap4'}, one must take care to ensure Bootstrap's HTML
dependencies (as well as Bootswatch themes, if desired) are included on the
page. Note, when set explicitly, it's the user's responsibility to ensure
that only one unique `style` value is used on the same page, if multiple
DT tables exist, as different styling resources may conflict with each other.}

\item{width, height}{Width/Height in pixels (optional, defaults to automatic
sizing)}

\item{elementId}{An id for the widget (a random string by default).}

\item{fillContainer}{\code{TRUE} to configure the table to automatically fill
it's containing element. If the table can't fit fully into it's container
then vertical and/or horizontal scrolling of the table cells will occur.}

\item{autoHideNavigation}{\code{TRUE} to automatically hide navigational UI
(only display the table body) when the number of total records is less
than the page size. Note, it only works on the client-side processing mode
and the `pageLength` option should be provided explicitly.}

\item{selection}{the row/column selection mode (single or multiple selection
or disable selection) when a table widget is rendered in a Shiny app;
alternatively, you can use a list of the form \code{list(mode = 'multiple',
selected = c(1, 3, 8), target = 'row', selectable = c(-2, -3))} to
pre-select rows and control the selectable range; the element
\code{target} in the list can be \code{'column'} to enable column
selection, or \code{'row+column'} to make it possible to select both rows
and columns (click on the footer to select columns), or \code{'cell'} to
select cells. See details section for more info.}

\item{extensions}{a character vector of the names of the DataTables
extensions (\url{https://datatables.net/extensions/index})}

\item{plugins}{a character vector of the names of DataTables plug-ins
(\url{https://rstudio.github.io/DT/plugins.html}).  Note that only those
plugins supported by the \code{DT} package can be used here. You can see
the available plugins by calling \code{DT:::available_plugins()}}

\item{editable}{\code{FALSE} to disable the table editor, or \code{TRUE} (or
\code{"cell"}) to enable editing a single cell. Alternatively, you can set
it to \code{"row"} to be able to edit a row, or \code{"column"} to edit a
column, or \code{"all"} to edit all cells on the current page of the table.
In all modes, start editing by doubleclicking on a cell. This argument can
also be a list of the form \code{list(target = TARGET, disable =
list(columns = INDICES))}, where \code{TARGET} can be \code{"cell"},
\code{"row"}, \code{"column"}, or \code{"all"}, and \code{INDICES} is an
integer vector of column indices. Use the list form if you want to disable
editing certain columns. You can also restrict the editing to accept only
numbers by setting this argument to a list of the form \code{list(target =
TARGET, numeric = INDICES)} where \code{INDICES} can be the vector of the
indices of the columns for which you want to restrict the editing to
numbers or \code{"all"} to restrict the editing to numbers for all columns.
If you don't set \code{numeric}, then the editing is restricted to numbers
for all numeric columns; set \code{numeric = "none"} to disable this
behavior. It is also possible to edit the cells in text areas, which are
useful for large contents. For that, set the \code{editable} argument to a
list of the form \code{list(target = TARGET, area = INDICES)} where
\code{INDICES} can be the vector of the indices of the columns for which
you want the text areas, or \code{"all"} if you want the text areas for
all columns. Of course, you can request the numeric editing for some
columns and the text areas for some other columns by setting
\code{editable} to a list of the form \code{list(target = TARGET, numeric
= INDICES1, area = INDICES2)}. Finally, you can edit date cells with a
calendar with \code{list(target = TARGET, date = INDICES)}; the target
columns must have the \code{Date} type. If you don't set \code{date} in
the \code{editable} list, the editing with the calendar is automatically
set for all \code{Date} columns.}

\item{id}{\code{character(1)} module id}

\item{keys}{\code{character}. Defaults to all columns under the assumption that at least every row is unique.}

\item{in_place}{\code{logical}. Whether to modify the data object in place or to return a modified copy.}

\item{format}{function accepting and returning a \code{\link[DT]{datatable}}}

\item{foreignTbls}{\code{list}. List of objects created by \code{\link{foreignTbl}}}

\item{columnOrder}{\code{vector}. Order of columns, original data and foreignTbls combined. Defaults to no order specified.}

\item{statusColor}{named \code{character}. Colors to indicate status of the row.}

\item{inputUI}{\code{function}. UI function of a shiny module with at least arguments \code{id} \code{data} and \code{...}.
#'   elements with inputIds identical to one of the column names are used to update the data.}

\item{defaults}{expression that evaluates to a \code{tibble} with (a subset of) columns of the data.
It will be evaluated for each new row in the environment defined by 'env'.
This allows for defaults like Sys.time() or uuid::UUIDgenerate() as well as dynamic inputs.}

\item{cloneDefaults}{expression that evaluates to a \code{tibble} with (a subset of) columns of the data.
It will be evaluated for each cloned row in the environment defined by 'env'.}

\item{env}{\code{environment} in which the server function is running. Should normally not be modified.}

\item{canEditRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}

\item{canDeleteRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}

\item{canCloneRow}{can be either of the following:
\itemize{
\item \code{logical}, e.g. TRUE or FALSE. In case of \code{FALSE}, buttons will not be visible except for new rows.
\item \code{function}. Needs as input an argument \code{row} which accepts a single row \code{tibble} and as output TRUE/FALSE.
}}

\item{utilityColumns}{named character vector. Defines names for (hidden) utility columns
used by \code{eDT} to keep track of modifications. Should normally only be adjusted in rare case of name clashes with data.

\if{html}{\out{<div class="sourceCode">}}\preformatted{c(
  status = '_editbl_status',
  buttons = '_editbl_buttons',
  identity = '_editbl_identity',
  deleted = '_editbl_deleted'
 )
}\if{html}{\out{</div>}}}
}
\value{
list
\itemize{
\item result \code{reactive} modified version of \code{data} (saved)
\item state \code{reactive} current state of the \code{data} (unsaved)
\item selected \code{reactive} selected rows of the \code{data} (unsaved)
\item edited \code{reactive} edited rows (saved)
\item inserted \code{reactive} inserted rows (saved)
\item deleted \code{reactive} deleted rows (saved)
}
}
\description{
Create a modifieable datatable.
}
\details{
Works the same as \code{\link[DT]{datatable}}.
This function is however a shiny module and comes with additional arguments and different defaults.
Instead of having \code{output$id = renderDT(DT::datatable(iris))}, \code{eDT(id = 'id', data = iris)} should be used on the server side.
On the UI side \code{\link{eDTOutput}} should be used instead of \code{\link[DT]{DTOutput}}.

Can also be used as standalone app when not ran in reactive context.

All arguments except 'id' and 'env' can be normal objects or reactive objects.
}
\examples{
## Only run this example in interactive R sessions
if(interactive()){
  # tibble support
  modifiedData <- editbl::eDT(tibble::as_tibble(mtcars))

  # data.table support
  modifiedData <- editbl::eDT(dtplyr::lazy_dt(data.table::data.table(mtcars)))

  # database support
  tmpFile <- tempfile(fileext = ".sqlite")
  file.copy(system.file("extdata", "chinook.sqlite", package = 'editbl'), tmpFile)

  conn <- editbl::connectDB(dbname = tmpFile)
  modifiedData <- editbl::eDT(dplyr::tbl(conn, "Artist"), in_place = TRUE)
  DBI::dbDisconnect(conn)

  unlink(tmpFile)

  # Within shiny
  library(shiny)
  library(editbl)
  shinyApp(
    ui = fluidPage(fluidRow(column(12, eDTOutput('tbl')))),
    server = function(input, output) {
      eDT('tbl',iris,)
    }
  )

  # Custom inputUI
  editbl::eDT(mtcars, inputUI = function(id, data){
    ns <- NS(id)
    textInput(
    ns("mpg"),
    label = "mpg",
    value = data$mpg)})

  # Do not allow delete
  editbl::eDT(mtcars, canDeleteRow = FALSE)
}

}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{castForDisplay}
\alias{castForDisplay}
\title{Cast columns in \code{data.frame} to editable types in datatable}
\usage{
castForDisplay(data, cols = colnames(data))
}
\arguments{
\item{data}{\code{data.frame}}

\item{cols}{\code{character} columns to perform casting on.}
}
\value{
\code{data.frame} with some columns cast to another type
}
\description{
Cast columns in \code{data.frame} to editable types in datatable
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreignTbl.R
\name{fillDeductedColumns}
\alias{fillDeductedColumns}
\title{Fill data columns based on foreignTbls}
\usage{
fillDeductedColumns(tbl, foreignTbls)
}
\arguments{
\item{tbl}{\code{tbl}}

\item{foreignTbls}{list of foreign tbls as created by \code{\link{foreignTbl}}}
}
\value{
tbl
}
\description{
Fill data columns based on foreignTbls
}
\details{
When a combination of columns is not found in the foreignTbl,
fill the deductedColumns with NA.

on foreignTbls suggesting conflicting data,
an arbitrary choice is made. It is best to afterwards check with
checkForeignTbls to see if a valid result is obtained.
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinyInput.R
\name{inputUI.default}
\alias{inputUI.default}
\title{UI part for modal with input fields for editing}
\usage{
\method{inputUI}{default}(id, ...)
}
\arguments{
\item{id}{character module id}

\item{...}{for compatibility with method}
}
\value{
HTML. A set of input fields corresponding to the given row.
}
\description{
UI part for modal with input fields for editing
}
\details{
The UI elements that have an id identical to a column name are used for updating the data.
}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreignTbl.R
\name{castToFactor}
\alias{castToFactor}
\title{Cast all columns that exist in a foreignTbl to factor}
\usage{
castToFactor(data, foreignTbls)
}
\arguments{
\item{data}{\code{data.frame}}

\item{foreignTbls}{list of foreign tbls as created by \code{\link{foreignTbl}}}
}
\value{
data.frame
}
\description{
Cast all columns that exist in a foreignTbl to factor
}
\details{
Can be used to fixate possible options when editing.
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fixInteger64}
\alias{fixInteger64}
\title{Replace instances of integer64 with actual NA values instead of weird default 9218868437227407266}
\usage{
fixInteger64(x)
}
\arguments{
\item{x}{\code{data.frame}}
}
\value{
x with \code{integer64} columns set to \code{bit64::as.integer64(NA)}
}
\description{
Replace instances of integer64 with actual NA values instead of weird default 9218868437227407266
}
\details{
\href{https://github.com/Rdatatable/data.table/issues/4561}{github issue}
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eDT.R
\name{createCloneButtonHTML_shiny}
\alias{createCloneButtonHTML_shiny}
\title{Helper function to write HTML}
\usage{
createCloneButtonHTML_shiny(ns = "\%1$s", suffix = "\%2$s", disabled = FALSE)
}
\arguments{
\item{ns}{\code{character(1)} namespace}

\item{suffix}{\code{character(1)} id of the row}

\item{disabled}{\code{logical(1)} wether or not the button has to be disabled}
}
\description{
Helper function to write HTML
}
\details{
only to be used interactively. sprintf() implementation
is faster.
}
\seealso{
createCloneButtonHTML
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foreignTbl.R
\name{joinForeignTbl}
\alias{joinForeignTbl}
\title{Merge a tbl with it a foreignTbl}
\usage{
joinForeignTbl(
  tbl,
  foreignTbl,
  keepNA = TRUE,
  by = foreignTbl$by,
  copy = TRUE,
  type = c("inner", "left")[1]
)
}
\arguments{
\item{tbl}{\code{tbl}}

\item{foreignTbl}{\code{list} as created by \code{\link{foreignTbl}}}

\item{keepNA}{\code{logical} keep rows from tbl with NA keys.}

\item{by}{named \code{character}, columns to join on.}

\item{copy}{\code{logical}, whether or not to copy the \code{foreignTbl} to the source of argument \code{tbl} for joining.}

\item{type}{\code{character(1)}, type of joint to perform. Can be 'inner' or 'left'.}
}
\value{
\code{tbl}, containing both columns from argument \code{tbl} and argument \code{foreignTbl}.
}
\description{
Merge a tbl with it a foreignTbl
}
\details{
see also \code{dplyr} join functions, for example \code{dplyr::left_join}.
}
\author{
Jasper Schelfhout
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demoApp.R
\name{demoServer_mtcars}
\alias{demoServer_mtcars}
\title{Server of the mtcars demo app}
\usage{
demoServer_mtcars(id)
}
\arguments{
\item{id}{\code{character(1)}}
}
\value{
NULL, just executes the module server.
}
\description{
Server of the mtcars demo app
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{standardizeArgument_editable}
\alias{standardizeArgument_editable}
\title{Standardized editable argument to be in the form of a list}
\usage{
standardizeArgument_editable(editable, data)
}
\arguments{
\item{editable}{\code{FALSE} to disable the table editor, or \code{TRUE} (or
\code{"cell"}) to enable editing a single cell. Alternatively, you can set
it to \code{"row"} to be able to edit a row, or \code{"column"} to edit a
column, or \code{"all"} to edit all cells on the current page of the table.
In all modes, start editing by doubleclicking on a cell. This argument can
also be a list of the form \code{list(target = TARGET, disable =
list(columns = INDICES))}, where \code{TARGET} can be \code{"cell"},
\code{"row"}, \code{"column"}, or \code{"all"}, and \code{INDICES} is an
integer vector of column indices. Use the list form if you want to disable
editing certain columns. You can also restrict the editing to accept only
numbers by setting this argument to a list of the form \code{list(target =
TARGET, numeric = INDICES)} where \code{INDICES} can be the vector of the
indices of the columns for which you want to restrict the editing to
numbers or \code{"all"} to restrict the editing to numbers for all columns.
If you don't set \code{numeric}, then the editing is restricted to numbers
for all numeric columns; set \code{numeric = "none"} to disable this
behavior. It is also possible to edit the cells in text areas, which are
useful for large contents. For that, set the \code{editable} argument to a
list of the form \code{list(target = TARGET, area = INDICES)} where
\code{INDICES} can be the vector of the indices of the columns for which
you want the text areas, or \code{"all"} if you want the text areas for
all columns. Of course, you can request the numeric editing for some
columns and the text areas for some other columns by setting
\code{editable} to a list of the form \code{list(target = TARGET, numeric
= INDICES1, area = INDICES2)}. Finally, you can edit date cells with a
calendar with \code{list(target = TARGET, date = INDICES)}; the target
columns must have the \code{Date} type. If you don't set \code{date} in
the \code{editable} list, the editing with the calendar is automatically
set for all \code{Date} columns.}

\item{data}{\code{tbl}. When using non-standard tibble objects, install the required package implementing specific methods.
(\code{data.table}: dtplyr, spark: sparklyr, generic database: dbplyr,...).
While other tabular objects like \code{data.frame} are supported, explicitly casting to \code{tbl} is recommended.}
}
\value{
list of the form \code{list(target = foo, ...)}
}
\description{
Standardized editable argument to be in the form of a list
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{castFromTbl}
\alias{castFromTbl}
\title{Cast tbl to class of template}
\usage{
castFromTbl(tbl, template)
}
\arguments{
\item{tbl}{\code{tbl}}

\item{template}{tabular object like \code{data.frame} or \code{data.table} or \code{tbl}.}
}
\value{
tbl cast to the type of template
}
\description{
Cast tbl to class of template
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getColumnTypeSums}
\alias{getColumnTypeSums}
\title{Get types of columns in a tbl}
\usage{
getColumnTypeSums(tbl)
}
\arguments{
\item{tbl}{\code{tbl}}
}
\value{
named list with types of the colums
}
\description{
Get types of columns in a tbl
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devApp.R
\name{runDevApp}
\alias{runDevApp}
\title{Run a development app}
\usage{
runDevApp()
}
\value{
An object that represents the app. Printing the object or passing it
to \code{\link[shiny:runApp]{runApp()}} will run the app.
}
\description{
Run a development app
}
\details{
This app prints some of the server objects and has a button to interactively browse the code.
This is useful for debugging and experimenting with new features.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl_dbi.R
\name{get_db_table_name}
\alias{get_db_table_name}
\title{Get name of the tbl in the database}
\usage{
get_db_table_name(x)
}
\arguments{
\item{x}{\code{tbl_dbi}}
}
\value{
SQL, the table name as used in the database
}
\description{
Get name of the tbl in the database
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demoApp.R
\name{demoUI_DB}
\alias{demoUI_DB}
\title{UI of the DB demo app}
\usage{
demoUI_DB(id, conn)
}
\arguments{
\item{id}{\code{character(1)}}

\item{conn}{database connection object as given by \code{\link[DBI]{dbConnect}}.}
}
\value{
HTML
}
\description{
UI of the DB demo app
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl_dbi.R
\name{castToSQLSupportedType}
\alias{castToSQLSupportedType}
\title{Cast the data type to something supported by SQL.}
\usage{
castToSQLSupportedType(x)
}
\arguments{
\item{x}{single value or vector of values}
}
\value{
x, possibly cast to different type
}
\description{
Cast the data type to something supported by SQL.
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinyInput.R
\name{shinyInput}
\alias{shinyInput}
\title{Get a shiny input for a column of a \code{tbl}}
\usage{
shinyInput(x, inputId, label, selected)
}
\arguments{
\item{x}{column}

\item{inputId}{shiny input Id}

\item{label}{\code{character(1)}}

\item{selected}{object of class of x}
}
\value{
shiny input
}
\description{
Get a shiny input for a column of a \code{tbl}
}
\author{
Jasper Schelfhout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{connectDB}
\alias{connectDB}
\title{Connect to a database.}
\usage{
connectDB(
  dbname = system.file("extdata", "chinook.sqlite", package = utils::packageName()),
  drv = RSQLite::SQLite(),
  ...
)
}
\arguments{
\item{dbname}{\code{character(0)}}

\item{drv}{database driver}

\item{...}{arguments passed to \code{DBI::dbConnect}}
}
\value{
database connection
}
\description{
Connect to a database.
}
\details{
Connects by default to a test SQLite database originally obtained here:
\href{https://github.com/lerocha/chinook-database/blob/master/ChinookDatabase/DataSources/Chinook_Sqlite.sqlite}{chinook_git}
}
\examples{

conn <- connectDB()
DBI::dbDisconnect(conn)

}
